unit Unit2;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  VML_Log, CommPort, VML_Port, VML_Profile, VML_Modem, VML_NoModem,
  VML_IS101, VML_USRobotics;

const

  { Notify message }

  WM_MODEMNOTIFY   = WM_USER + 1;

  { Modem operation state }

  MS_IDLE      = 0; { idle state }
  MS_STARTING  = 1; { initialize modem }
  MS_PLAYING   = 2; { playing voice }
  MS_RECORDING = 3; { recording voice }
  MS_ANSWERING = 4; { waiting for incoming ring }
  MS_STOPPING  = 5; { stopping current action }
  MS_ERROR     = 6; { error }

type

{ TModemThread }

  TModemOperation = (opPlay, opRecord, opAnswer);

  TModemThread = class(TThread)
  private
    FOwner: TObject;
    FPlayFileName: string;
    FRecordFileName: string;
    FOperation: TModemOperation;
    FModemEvents: TList;
    FErrorMsg: string;
    procedure DoAnswer;
    procedure DoErrorMsg;
    procedure DoModemEvent(Sender: TObject; Event: Integer; Data: TEventData; var Result: Integer);
    procedure DoNotify(ModemState: Integer);
    procedure DoPlay;
    procedure DoRecord;
  protected
    procedure Call(Method: TThreadMethod);
    function Check(ModemResult: Integer; const ErrorMsg: string): Integer;
    procedure DoTerminate; override;
    procedure Execute; override;
  public
    constructor Create(AOwner: TObject);
    destructor Destroy; override;
    procedure Stop;
  end;

{ TModemModule }

  TModemModule = class(TDataModule)
    VoiceModem: TVoiceModem;
    ModemProfile: TModemProfile;
    ModemPort: TModemPort;
    ModemLogFile: TModemLogFile;
    ModemTraceFile: TModemLogFile;
    USRobotics: TUSRobotics;
    procedure ModemModuleCreate(Sender: TObject);
  private
    FThread: TModemThread;
    procedure CreateThread(Owner: TObject);
    procedure ThreadDone(Sender: TObject);
  public
    procedure PlayFile(Owner: TObject; const PlayFileName: string);
    procedure RecordFile(Owner: TObject; const RecordFileName: string);
    procedure Stop(Wait: Boolean);
    procedure WaitRing(Owner: TObject; const PlayFileName, RecordFileName: string);
  end;

var
  ModemModule: TModemModule;

implementation

uses
  Unit1, VML_Consts;

{$R *.DFM}

{ TModemThread }

constructor TModemThread.Create(AOwner: TObject);
begin
  inherited Create(True);
  FOwner := AOwner;
  FModemEvents := TList.Create;
end;

destructor TModemThread.Destroy;
begin
  FModemEvents.Free;
  inherited Destroy;
end;

procedure TModemThread.DoAnswer;
const
  Error1 = 'An error occured while setting telephone line device (Error %d).';
  Error2 = 'An error occured while waiting incoming ring (Error %d)';
  Error21 = 'An error occured while answering call (Error %d)';
  Error3 = 'An error occured while playing voice file (Error %d).';
  Error4 = 'An error occured while recording voice file (Error %d).';
  Error5 = 'An error occured while trying to reset modem device (Error %d)';
  Error6 = 'An error occured while trying hangs up (Error %d)';
  Error7 = 'An error occured while restoring voice mode (Error %d)';
  Message1 = 'You have new message.';
type
  TCallState = (csWaiting, csAnswering, csPlayingGreeting, csBeeping, csRecordingAnswer);
var
  CallState: TCallState;
  CallComplete: Boolean;

  procedure HandleModemEvents;
  const
    Event1 = 'Abonent does not want to save message.';
  var
    I, Event: Integer;
  begin
    for I := FModemEvents.Count - 1 downto 0 do
    begin
      Event := Integer(FModemEvents[I]);
      { Check calling state }
      case CallState of
        csPlayingGreeting, csBeeping:
          { Playing greeting in progress }
          case Event of
            BUSY_TONE:
              { Busy tone detected }
              begin
                CallComplete := True;
                FErrorMsg := Event1;
                Call(DoErrorMsg);
                Exit;
              end;
          end;
      end;
    end;
  end;

  function ProcessCallState: Boolean;
  begin
    with ModemModule.VoiceModem do
    begin
      FModemEvents.Clear;
      { Process call step }
      case CallState of
        csWaiting:
          while True do
            { Wait RING }
            case Check(WaitAnswer('', 0), Error2) of
              INTERRUPTED:
                { Wait interrupted }
                begin
                  CallComplete := True;
                  Break;
                end;
              VMA_RING:
                { Incoming ring detected }
                begin
                  Break;
                end;
            end;
        csAnswering: Check(AnswerPhone, Error21);
        csPlayingGreeting: if Check(PlayFile(FPlayFileName), Error3) = INTERRUPTED then CallComplete := True;
        csBeeping: Check(Beep(Profile.BeepFrequency, Profile.BeepLength), Error3);
        csRecordingAnswer: Check(RecordFile(FRecordFileName), Error4);
      end;
      { Process modem events }
      HandleModemEvents;
      { Check step result }
      if CallComplete then
      begin
        { Hangup, restore voice mode }
        HangUp;
        RestoreVoiceMode;
        { Exit; }
        Result := False;
        Exit;
      end;
      { Sucess }
      Result := True;
    end;
  end;

begin
  with ModemModule.VoiceModem do
  begin
    CallComplete := False;
    { Hook modem events handler }
    OnEvent := DoModemEvent;
    try
      DoNotify(MS_ANSWERING);
      { Set line device }
      Check(SetDevice(DIALUP_LINE), Error1);
      { Wait incoming ring }
      CallState := csWaiting;
      if not ProcessCallState then Exit;
      { Answer call }
      CallState := csAnswering;
      if not ProcessCallState then Exit;
      { Play greeting }
      CallState := csPlayingGreeting;
      if not ProcessCallState then Exit;
      { Beep signal before recording }
      CallState := csBeeping;
      if not ProcessCallState then Exit;
      { Record answer }
      CallState := csRecordingAnswer;
      if not ProcessCallState then Exit;
      { Reset device }
      Check(SetDevice(NO_DEVICE), Error5);
      { Hangup, restore voice mode  }
      Check(HangUp, Error6);
      Check(RestoreVoiceMode, Error7);
      { New message received }
      FErrorMsg := Message1;
      Call(DoErrorMsg);
    finally
      OnEvent := nil;
    end;
  end;
end;

procedure TModemThread.DoErrorMsg;
const
  Flags = MB_OK or MB_ICONERROR;
begin
  with Application do MessageBox(PChar(FErrorMsg), PChar(Title), Flags);
end;

procedure TModemThread.DoModemEvent(Sender: TObject; Event: Integer; Data: TEventData; var Result: Integer);
begin
  FModemEvents.Add(Pointer(Event));
end;

procedure TModemThread.DoNotify(ModemState: Integer);
begin
  PostMessage(AnswerForm.Handle, WM_MODEMNOTIFY, ModemState, Longint(FOwner));
end;

procedure TModemThread.DoPlay;
const
  Error1 = 'An error occured while setting play device (Error %d).';
  Error2 = 'An error occured while playing voice file (Error %d).';
  Error3 = 'An error occured while trying to reset play device (Error %d).';
  Error4 = 'An error occured while restoring voice mode (Error %d)';
begin
  with ModemModule.VoiceModem do
  begin
    { Set play device }
    Check(SetDevice(Profile.PlayDevice), Error1);
    { Play file }
    DoNotify(MS_PLAYING);
    Check(PlayFile(FPlayFileName), Error2);
    { Reset device }
    Check(SetDevice(NO_DEVICE), Error3);
    Check(RestoreVoiceMode, Error4);
  end;
end;

procedure TModemThread.DoRecord;
const
  Error1 = 'An error occured while setting record device (Error %d).';
  Error2 = 'An error occured while recording voice file (Error %d).';
  Error3 = 'An error occured while trying to reset record device (Error %d).';
  Error4 = 'An error occured while restoring voice mode (Error %d)';
begin
  with ModemModule.VoiceModem do
  begin
    { Set record device }
    Check(SetDevice(Profile.RecDevice), Error1);
    { Record voice }
    DoNotify(MS_RECORDING);
    Check(RecordFile(FRecordFileName), Error2);
    { Reset device }
    Check(SetDevice(NO_DEVICE), Error3);
    Check(RestoreVoiceMode, Error4);
  end;
end;

procedure TModemThread.Call(Method: TThreadMethod);
begin
  { In main thread call method without synchronization }
  if GetCurrentThreadID = MainThreadID then
  begin
    Method;
    Exit;
  end;
  { Check thread not terminated and synchronize method }
  if not Terminated then Synchronize(Method);
end;

function TModemThread.Check(ModemResult: Integer; const ErrorMsg: string): Integer;
begin
  Result := ModemResult;
  { Check for modem errors }
  if (ModemResult = FAIL) or (ModemResult = VMA_FAIL) then
  begin
    with ModemModule.VoiceModem do
    begin
      { Save modem last error, close modem }
      ReturnValue := LastError;
      HangUp;
      Close;
    end;
    { Raise exception }
    raise Exception.CreateFmt(ErrorMsg, [ReturnValue]);
  end;
end;

procedure TModemThread.DoTerminate;
begin
  { Idle state }
  DoNotify(MS_IDLE);
  { Check thread not terminated and call inherited method }
  if not Terminated then inherited DoTerminate;
end;

procedure TModemThread.Execute;
const
  Error1 = 'Could not access modem. Device is busy.';
  Error2 = 'An error occured while initialize modem (Error %d). Check modem settings and try again.';
begin
  with ModemModule.VoiceModem do
  try
    { Lock modem }
    if not Lock(Self) then
    begin
      ReturnValue := VME_MODEM_LOCKED;
      raise Exception.Create(Error1);
    end;
    try
      { Starting in progress }
      DoNotify(MS_STARTING);
      { Open device }
      Check(Open, Error2);
      { Check modem }
      Check(Command('AT', 'OK'), Error2);
      { Flush modem event queue }
      FlushEvents;
      { Check tread terminated }
      if not Terminated then
      begin
        { Process operation }
        case FOperation of
          opPlay: DoPlay;
          opRecord: DoRecord;
          opAnswer: DoAnswer;
        end;
      end;
    finally
      Unlock;
    end;
  except
    on E: Exception do
    begin
      { Error state }
      DoNotify(MS_ERROR);
      { Show error message }
      FErrorMsg := E.Message;
      Call(DoErrorMsg);
    end;
  end;
end;

procedure TModemThread.Stop;
begin
  { Queue stop event }
  ModemModule.VoiceModem.Stop;
  { Stopping in progress }
  DoNotify(MS_STOPPING);
end;

{ TModemModule }

procedure TModemModule.ThreadDone(Sender: TObject);
begin
  FThread := nil;
end;

procedure TModemModule.CreateThread(Owner: TObject);
const
  Error1 = 'Could not access modem. Device is busy.';
  Flags = MB_OK or MB_ICONERROR;
begin
  { Lock modem }
  if (FThread <> nil) or VoiceModem.Locked then
  begin
    with Application do MessageBox(PChar(Error1), PChar(Title), Flags);
    Abort;
  end;
  { Create operation thread }
  FThread := TModemThread.Create(Owner);
  FThread.FreeOnTerminate := True;
  FThread.OnTerminate := ThreadDone;
end;

procedure TModemModule.PlayFile(Owner: TObject; const PlayFileName: string);
begin
  { Create operation thread }
  CreateThread(Owner);
  { Fill thread params }
  FThread.FPlayFileName := PlayFileName;
  FThread.FOperation := opPlay;
  { Start operation }
  FThread.Resume;
end;

procedure TModemModule.RecordFile(Owner: TObject; const RecordFileName: string);
begin
  { Create operation thread }
  CreateThread(Owner);
  { Fill thread params }
  FThread.FRecordFileName := RecordFileName;
  FThread.FOperation := opRecord;
  { Start operation }
  FThread.Resume;
end;

procedure TModemModule.Stop(Wait: Boolean);
begin
  { Check operation in progress }
  if FThread <> nil then
  begin
    Screen.Cursor := crHourGlass;
    try
      { Stop operation }
      if Wait then
      begin
        VoiceModem.Stop;
        { Wait thread }
        FThread.Terminate;
        FThread.WaitFor;
        { Delete thread link }
        FThread := nil;
      end
      else
        FThread.Stop;
    finally
      Screen.Cursor := crDefault;
    end;
  end;
end;

procedure TModemModule.WaitRing(Owner: TObject; const PlayFileName, RecordFileName: string);
begin
  { Create operation thread }
  CreateThread(Owner);
  { Fill thread params }
  FThread.FPlayFileName := PlayFileName;
  FThread.FRecordFileName := RecordFileName;
  FThread.FOperation := opAnswer;
  { Start operation }
  FThread.Resume;
end;

procedure TModemModule.ModemModuleCreate(Sender: TObject);
begin
  with ModemLogFile do FileName := ExtractFilePath(Application.ExeName) + FileName;
  with ModemTraceFile do FileName := ExtractFilePath(Application.ExeName) + FileName;
end;

end.
