unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls, CommPort, VML_Port, VML_Modem, VML_Profile, VML_Log,
  VML_NoModem, VML_IS101, VML_USRobotics;

const

  { Notify message }

  WM_MODEMNOTIFY   = WM_USER + 1;

  { Modem operation state }

  MS_IDLE      = 0; { idle state }
  MS_STARTING  = 1; { initialize modem }
  MS_PLAYING   = 2; { playing voice }
  MS_RECORDING = 3; { recording voice }
  MS_STOPPING  = 4; { stopping current action }
  MS_ERROR     = 5; { error }

type

{ TModemThread }

  TModemOperation = (opPlay, opRecord);

  TModemThread = class(TThread)
  private
    FFileName: string;
    FOperation: TModemOperation;
    FErrorMsg: string;
    procedure DoErrorMsg;
  protected
    procedure Call(Method: TThreadMethod);
    procedure DoTerminate; override;
    procedure Execute; override;
  public
    constructor Create(const AFileName: string; AOperation: TModemOperation);
    procedure Stop;
  end;

{ TThreadForm }

  TThreadForm = class(TForm)
    FileLabel: TLabel;
    FileEdit: TEdit;
    BrowseButton: TButton;
    Bevel: TBevel;
    PlayButton: TButton;
    StopButton: TButton;
    RecordButton: TButton;
    OpenDialog: TOpenDialog;
    VoiceModem: TVoiceModem;
    ModemProfile: TModemProfile;
    USRobotics: TUSRobotics;
    ModemPort: TModemPort;
    ModemLogFile: TModemLogFile;
    ModemTraceFile: TModemLogFile;
    procedure FormCreate(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure BrowseButtonClick(Sender: TObject);
    procedure PlayButtonClick(Sender: TObject);
    procedure StopButtonClick(Sender: TObject);
    procedure RecordButtonClick(Sender: TObject);
  private
    FThread: TModemThread;
    procedure Start(const FileName: string; Operation: TModemOperation);
    procedure Stop(Wait: Boolean);
    procedure ThreadDone(Sender: TObject);
    procedure WMModemNotify(var Message: TMessage); message WM_MODEMNOTIFY;
  public
    { Public declarations }
  end;

var
  ThreadForm: TThreadForm;

implementation

uses
  VML_Consts;

{$R *.DFM}

{ TModemThread }

constructor TModemThread.Create(const AFileName: string; AOperation: TModemOperation);
begin
  inherited Create(True);
  FFileName := AFileName;
  FOperation := AOperation;
end;

procedure TModemThread.DoErrorMsg;
const
  Flags = MB_OK or MB_ICONERROR;
begin
  with Application do MessageBox(PChar(FErrorMsg), PChar(Title), Flags);
end;

procedure TModemThread.Call(Method: TThreadMethod);
begin
  { In main thread call method without synchronization }
  if GetCurrentThreadID = MainThreadID then
  begin
    Method;
    Exit;
  end;
  { Check thread not terminated and synchronize method }
  if not Terminated then Synchronize(Method);
end;

procedure TModemThread.DoTerminate;
begin
  { Idle state }
  PostMessage(ThreadForm.Handle, WM_MODEMNOTIFY, MS_IDLE, 0);
  { Check thread not terminated and call inherited method }
  if not Terminated then inherited DoTerminate;
end;

procedure TModemThread.Execute;
const
  Error1 = 'Could not access modem. Device is busy.';
  Error2 = 'An error occured while initialize modem (Error %d). Check modem settings and try again.';
  Error3 = 'An error occured while playing voice file (Error %d).';
  Error4 = 'An error occured while recording voice file (Error %d).';
  Error5 = 'An error occured while trying to reset device (Error %d).';
  Error6 = 'An error occured while restoring voice mode (Error %d)';

  procedure Check(Result: Integer; const ErrorMsg: string);
  begin
    if (Result = FAIL) or (Result = VMA_FAIL) then
    begin
      with ThreadForm.VoiceModem do
      begin
        { Save modem last error, close modem }
        ReturnValue := LastError;
        Close;
      end;
      { Raise exception }
      raise Exception.CreateFmt(ErrorMsg, [ReturnValue]);
    end;
  end;

begin
  with ThreadForm.VoiceModem do
  try
    { Lock modem }
    if not Lock(Self) then
    begin
      ReturnValue := VME_MODEM_LOCKED;
      raise Exception.Create(Error1);
    end;
    try
      { Starting in progress }
      PostMessage(ThreadForm.Handle, WM_MODEMNOTIFY, MS_STARTING, 0);
      { Open device }
      Check(Open, Error2);
      { Check modem }
      Check(Command('AT', 'OK'), Error2);
      { Flush modem event queue }
      FlushEvents;
      { Check tread terminated }
      if not Terminated then
      begin
        { Process operation }
        case FOperation of
          opPlay:
            { Play voice file }
            begin
              Check(SetDevice(Profile.PlayDevice), Error2);
              PostMessage(ThreadForm.Handle, WM_MODEMNOTIFY, MS_PLAYING, 0);
              Check(PlayFile(FFileName), Error3);
            end;
          opRecord:
            { Record voice file }
            begin
              Check(SetDevice(Profile.RecDevice), Error2);
              PostMessage(ThreadForm.Handle, WM_MODEMNOTIFY, MS_RECORDING, 0);
              Check(RecordFile(FFileName), Error4);
            end;
        end;
        { Reset device, restore voice mode }
        Check(SetDevice(NO_DEVICE), Error5);
        Check(RestoreVoiceMode, Error6);
      end;
    finally
      Unlock;
    end;
  except
    on E: Exception do
    begin
      { Error state }
      PostMessage(ThreadForm.Handle, WM_MODEMNOTIFY, MS_ERROR, 0);
      { Show error message }
      FErrorMsg := E.Message;
      Call(DoErrorMsg);
    end;
  end;
end;

procedure TModemThread.Stop;
begin
  { Queue stop event }
  ThreadForm.VoiceModem.Stop;
  { Stopping in progress }
  PostMessage(ThreadForm.Handle, WM_MODEMNOTIFY, MS_STOPPING, 0);
end;

{ TThreadForm }

procedure TThreadForm.Start(const FileName: string; Operation: TModemOperation);
const
  Error1 = 'Could not access modem. Device is busy.';
  Flags = MB_OK or MB_ICONERROR;
begin
  { Lock modem }
  if (FThread <> nil) or VoiceModem.Locked then
  begin
    with Application do MessageBox(PChar(Error1), PChar(Title), Flags);
    Abort;
  end;
  { Create operation thread }
  FThread := TModemThread.Create(FileName, Operation);
  FThread.FreeOnTerminate := True;
  FThread.OnTerminate := ThreadDone;
  { Start operation }
  FThread.Resume;
end;

procedure TThreadForm.Stop(Wait: Boolean);
begin
  { Check operation in progress }
  if FThread <> nil then
  begin
    Screen.Cursor := crHourGlass;
    try
      { Stop operation }
      if Wait then
      begin
        VoiceModem.Stop;
        { Wait thread }
        FThread.Terminate;
        FThread.WaitFor;
        { Delete thread link }
        FThread := nil;
      end
      else
        FThread.Stop;
    finally
      Screen.Cursor := crDefault;
    end;
  end;
end;

procedure TThreadForm.ThreadDone(Sender: TObject);
begin
  FThread := nil;
end;

procedure TThreadForm.WMModemNotify(var Message: TMessage);
const
  Cursors: array[Boolean] of TCursor = (crDefault, crAppStart);
begin
  with Message do
  begin
    PlayButton.Enabled := WParam = MS_IDLE;
    StopButton.Enabled := WParam in [MS_PLAYING, MS_RECORDING];
    RecordButton.Enabled := WParam = MS_IDLE;
    Screen.Cursor := Cursors[WParam in [MS_STARTING, MS_STOPPING]];
  end;
end;

procedure TThreadForm.FormCreate(Sender: TObject);
begin
  with ModemLogFile do FileName := ExtractFilePath(Application.ExeName) + FileName;
  with ModemTraceFile do FileName := ExtractFilePath(Application.ExeName) + FileName;
end;

procedure TThreadForm.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
  Stop(True);
end;

procedure TThreadForm.BrowseButtonClick(Sender: TObject);
begin
  with OpenDialog do
  begin
    FileName := FileEdit.Text;
    if Execute then FileEdit.Text := FileName;
  end;
end;

procedure TThreadForm.PlayButtonClick(Sender: TObject);
begin
  Start(FileEdit.Text, opPlay);
end;

procedure TThreadForm.StopButtonClick(Sender: TObject);
begin
  Stop(False);
end;

procedure TThreadForm.RecordButtonClick(Sender: TObject);
begin
  Start(FileEdit.Text, opRecord);
end;

end.
