{
  TCommPort component for Borland Delphi for Win95 & WinNT

  Version 1.02

  Original source code by
    (C) Boris Loboda  2:461/256
        barry@audit.kharkov.com
        
  Bugs fix, source reformatting by
    (C) Roman M. Mochalov
        roman@sar.nnov.ru
}

unit CommPort;

interface

uses
  Windows, Messages, SysUtils, Classes;

type

{ Common types }

  PBuffArray = ^TBuffArray;
  TBuffArray = array[0..32767] of char;

{ Predefine classes }

  TCustomCommPort = class;

{ TStatusThread }

  TStatusThread = class(TThread)
  private
    FComPort: TCustomCommPort;
    FComHandle: THandle;
    FDelay: DWORD;
    FSOL: TOverlapped;
    FEvtMask: DWORD;
    FErrors: DWORD;
    procedure DoBreakSignal;
    procedure DoCTSSignal;
    procedure DoDataAvail;
    procedure DoDSRSignal;
    procedure DoDTRSignal;
    procedure DoErrorSignal;
    procedure DoRingSignal;
    procedure DoRLSDSignal;
    procedure DoTxEmptySignal;
    procedure HandleCommEvent;
  protected
    procedure Execute; override;
  public
    constructor Create(ComPort: TCustomCommPort);
    destructor Destroy; override;
  end;

{ TWriteThread }

  TWriteThread = class(TThread)
  protected
    FComPort: TCustomCommPort;
    FComHandle: THandle;
    FDelay: DWORD;
    FWOL: TOverlapped;
  protected
    procedure Execute; override;
  public
    constructor Create(ComPort: TCustomCommPort);
    destructor Destroy; override;
  end;

{ TReadThread }

  TReadThread = class(TThread)
  protected
    FComPort: TCustomCommPort;
    FComHandle: THandle;
    FDelay: DWORD;
    FROL: TOverlapped;
    procedure DoOnDataAvail;
  protected
    procedure Execute; override;
  public
    constructor Create(ComPort: TCustomCommPort);
    destructor Destroy; override;
  end;

{ TCustomCommPort }

  {
    Comm port class.

    Methods:

    SetBREAKState -   Place the transmission line in a break or nonbreak
                      state.
    SetDTRState -     Sends or clear the DTR (data-terminal-ready) signal.
    SetRTSState -     Sends or clear the RTS (request-to-send) signal.

    CharReady -       Input buffer is not empty.
    ClosePort -       Close COM-port.
    Enabled -         COM-port is open.
    FlushInBuffer -   Flush input data-queue.
    FlushOutBuffer -  Flush output data-queue.
    GetBlock -        Get data from COM-port.
                      Check: Count <= InBuffUsed.
    GetChar -         Get char from input buffer.
                      Check: (CharReady = True) or (InBuffUsed > 0).
    OpenPort -        Open COM-port.
    PutBlock -        Put data in COM-port.
                      Check: Count <= OutBuffFree.
    PutChar -         Put char to output buffer. Check: OutBuffFree > 0.
    PutString -       Put string into output buffer.
                      Check: OutBuffFree >= Length(string).
    SendBreak -       Set Break signal. Ticks - ~1/18 sec.

    Properties:

    BaudRate -        Baud rate at which the communications device operates.
    ComHandle -       Opened COM-port handle.
    ComNumber -       Number of COM-port.
    CTS -             CTS signal state.
    DataBits -        Number of bits in the bytes transmitted and received. 
    RLSD -            RLSD signal state.
    DSR -             DSR signal state.
    DTR -             Set/Clear DTR signal.
    Enabled -         Such as Opened property.
    Error -           Error signal state.
    InBuffFree -      Free size in input buffer.
    InBuffSize -      Input buffer size.
    InBuffUsed -      Used size in input buffer.
    Opened -          Indicates whether the COM-port device has been opened.
    OutBuffFree -     Free size in output buffer.
    OutBuffSize -     Output buffer size.
    OutBuffUsed -     Used size in output buffer.
    OutQueCount -     Internal output queue count (in Win32 API buffer).
    Parity -          Parity scheme.
    ReadDelay -       Read data delay.
    RING -            RING signal detected.
    RTS -             Set/Clear RTS signal.
    RxD -             Not implemented.
    StatusDelay -     Signals polling delay.
    StopBits -        Number of stop bits.
    TxD -             Not implemented.
    WriteDelay -      Write data interval.

    Events:

    OnBreakSignal -   BREAK signal changed.
    OnCTSSignal -     CTS (clear-to-send) signal changed.
    OnDataAvail -     Input data received.
    OnDSRSignal -     DSR (data-set-ready) signal changed.
    OnDTRSignal -     DTR (data-terminal-ready) signal changed.
    OnErrorSignal -   ERROR signal changed.
    OnOpen -          COM-port opened.
    OnRingSignal -    Hardware Ring detected.
    OnRLSDSignal -    RLSD (receive-line-signal-detect) signal changed.
    OnRxDSignal -     Not implemented.
    OnTxEmptySignal - Output buffer is empty.
    OnTxDSignal -     Not implemented.
  }

  TComNumber = (cnCOM1, cnCOM2, cnCOM3, cnCOM4);
  TBaudRate = (br110, br300, br600, br1200, br2400, br4800, br9600, br14400,
    br19200, br38400, br56000, br57600, br115200, br128000, br256000);
  TParity = (paNone, paOdd, paEven, paMark, paSpace);
  TStopBits = (sb1_0, sb1_5, sb2_0);
  TDataBits = (db4, db5, db6, db7, db8);

  TCommDataAvailEvent = procedure(Sender: TObject; Count: Integer) of object;
  TCommErrorEvent = procedure(Sender: TObject; Errors: Integer) of object;

  TCustomCommPort = class(TComponent)
  private
    FComHandle: THandle;
    FComDCB: TDCB;
    FComNumber: TComNumber;
    FBaudRate: TBaudRate;
    FParity: TParity;
    FStopBits: TStopBits;
    FDataBits: TDataBits;
    FStatusThread: TStatusThread;
    FStatusEvent: THandle;
    FStatusDelay: DWORD;
    FWriteThread: TWriteThread;
    FWriteEvent: THandle;
    FWriteDelay: DWORD;
    FReadThread: TReadThread;
    FReadReadyEvent: THandle;
    FReadEvent: THandle;
    FReadDelay: DWORD;
    FKillThreads: Boolean;
    FWriteSection: TRTLCriticalSection;
    FReadSection: TRTLCriticalSection;
    FOutBuffer: PBuffArray;
    FOutBuffSize: DWORD;
    FOutBuffUsed: DWORD;
    FInBuffer: PBuffArray;
    FInBuffSize: DWORD;
    FInBuffUsed: DWORD;
    FCTS: Boolean;
    FDSR: Boolean;
    FDTR: Boolean;
    FRLSD: Boolean;
    FRTS: Boolean;
    FRxD: Boolean;
    FTxD: Boolean;
    FError: Boolean;
    FRing: Boolean;
    FOpened: Boolean;
    FOnOpen: TNotifyEvent;
    FOnClose: TNotifyEvent;
    FOnCTSSignal: TNotifyEvent;
    FOnDSRSignal: TNotifyEvent;
    FOnDTRSignal: TNotifyEvent;
    FOnRLSDSignal: TNotifyEvent;
    FOnTxEmptySignal: TNotifyEvent;
    FOnRxDSignal: TNotifyEvent;
    FOnTxDSignal: TNotifyEvent;
    FOnDataAvail: TCommDataAvailEvent;
    FOnBreakSignal: TNotifyEvent;
    FOnErrorSignal: TCommErrorEvent;
    FOnRingSignal: TNotifyEvent;
    function GetInBuffFree: DWORD;
    function GetInBuffUsed: DWORD;
    function GetOutBuffFree: DWORD;
    function GetOutBuffUsed: DWORD;
    function GetOutQueCount: DWORD;
    procedure SetBaudRate(Value: TBaudRate);
    procedure SetDataBits(Value: TDataBits);
    procedure SetDTR(Value: Boolean);
    procedure SetInBuffSize(Value: DWORD);
    procedure SetOpened(Value: Boolean);
    procedure SetOutBuffSize(Value: DWORD);
    procedure SetParity(Value: TParity);
    procedure SetRTS(Value: Boolean);
    procedure SetStopBits(Value: TStopBits);
  protected
    function SetBREAKState(State: Boolean): Boolean;
    function SetDTRState(State: Boolean): Boolean;
    function SetRTSState(State: Boolean): Boolean;
    property InBuffer: PBuffArray read FInBuffer;
    property OutBuffer: PBuffArray read FOutBuffer;
    property ReadSection: TRTLCriticalSection read FReadSection;
    property WriteSection: TRTLCriticalSection read FWriteSection;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function CharReady: Boolean;
    procedure ClosePort;
    function Enabled: Boolean;
    procedure FlushInBuffer;
    procedure FlushOutBuffer;
    function GetBlock(var Buf; Count: DWORD): DWORD;
    function GetChar: Char;
    procedure OpenPort;
    function PutBlock(const Buf; Count: DWORD): DWORD;
    function PutChar(C: Char): DWORd;
    function PutString(const S: string): DWORD;
    function SendBreak(Ticks: Word; Yield: Boolean): Boolean;
    property BaudRate: TBaudRate read FBaudRate write SetBaudRate default br19200;
    property ComHandle: THandle read FComHandle;
    property ComNumber: TComNumber read FComNumber write FComNumber default cnCOM2;
    property CTS: Boolean read FCTS;
    property DataBits: TDataBits read FDataBits write SetDataBits default db8;
    property DCD: Boolean read FRLSD;
    property DSR: Boolean read FDSR;
    property DTR: Boolean read FDTR write SetDTR;
    property Error: Boolean read FError;
    property InBuffFree: DWORD read GetInBuffFree;
    property InBuffSize: DWORD read FInBuffSize write SetInBuffSize default 4096;
    property InBuffUsed: DWORD read GetInBuffUsed;
    property Opened: Boolean read FOpened write SetOpened;
    property OutBuffFree: DWORD read GetOutBuffFree;
    property OutBuffSize: DWORD read FOutBuffSize write SetOutBuffSize default 4096;
    property OutBuffUsed: DWORD read GetOutBuffUsed;
    property OutQueCount: DWORD read GetOutQueCount;
    property Parity: TParity read FParity write SetParity default paNone;
    property ReadDelay: DWORD read FReadDelay write FReadDelay;
    property Ring: Boolean read FRing;
    property RLSD: Boolean read FRLSD;
    property RTS: Boolean read FRTS write SetRTS;
    property RxD: Boolean read FRxD;
    property StatusDelay: DWORD read FStatusDelay write FStatusDelay;
    property StopBits: TStopBits read FStopBits write SetStopBits default sb1_0;
    property TxD: Boolean read FTxD;
    property WriteDelay: DWORD read FWriteDelay write FWriteDelay;
    property OnBreakSignal: TNotifyEvent read FOnBreakSignal write FOnBreakSignal;
    property OnClose: TNotifyEvent read FOnClose write FOnClose;
    property OnCTSSignal: TNotifyEvent read FOnCTSSignal write FOnCTSSignal;
    property OnDataAvail: TCommDataAvailEvent read FOnDataAvail write FOnDataAvail;
    property OnDSRSignal: TNotifyEvent read FOnDSRSignal write FOnDSRSignal;
    property OnDTRSignal: TNotifyEvent read FOnDTRSignal write FOnDTRSignal;
    property OnErrorSignal: TCommErrorEvent read FOnErrorSignal write FOnErrorSignal;
    property OnOpen: TNotifyEvent read FOnOpen write FOnOpen;
    property OnRingSignal: TNotifyEvent read FOnRingSignal write FOnRingSignal;
    property OnRLSDSignal: TNotifyEvent read FOnRLSDSignal write FOnRLSDSignal;
    property OnRxDSignal: TNotifyEvent read FOnRxDSignal write FOnRxDSignal;
    property OnTxDSignal: TNotifyEvent read FOnTxDSignal write FOnTxDSignal;
    property OnTxEmptySignal: TNotifyEvent read FOnTxEmptySignal write FOnTxEmptySignal;
  end;

  TCommPort = class(TCustomCommPort)
  published
    property BaudRate;
    property ComNumber;
    property DataBits;
    property InBuffSize;
    property Parity;
    property StopBits;
    property OutBuffSize;
    property OnBreakSignal;
    property OnClose;
    property OnCTSSignal;
    property OnDSRSignal;
    property OnDTRSignal;
    property OnErrorSignal;
    property OnOpen;
    property OnRingSignal;
    property OnRLSDSignal;
    property OnRxDSignal;
    property OnDataAvail;
    property OnTxDSignal;
    property OnTxEmptySignal;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Samples', [TCommPort]);
end;

{ TStatusThread }

constructor TStatusThread.Create(ComPort: TCustomCommPort);
begin
  inherited Create(True);
  FComPort := ComPort;
  FComHandle := ComPort.ComHandle;
  FDelay := ComPort.StatusDelay;
  FillChar(FSOL, Sizeof(FSOL), 0);
  FSOL.hEvent := CreateEvent(nil, True, False, nil);
end;

destructor TStatusThread.Destroy;
begin
  CloseHandle(FSOL.hEvent);
  inherited Destroy;
end;

procedure TStatusThread.DoBreakSignal;
begin
  FComPort.FOnBreakSignal(FComPort);
end;

procedure TStatusThread.DoCTSSignal;
begin
  FComPort.FOnCTSSignal(FComPort);
end;

procedure TStatusThread.DoDataAvail;
begin
  FComPort.OnDataAvail(FComPort, FComPort.FInBuffUsed);
end;

procedure TStatusThread.DoDSRSignal;
begin
  FComPort.OnDSRSignal(FComPort);
end;

procedure TStatusThread.DoDTRSignal;
begin
  FComPort.OnDTRSignal(FComPort);
end;

procedure TStatusThread.DoErrorSignal;
begin
  FComPort.FOnErrorSignal(FComPort, FErrors);
end;

procedure TStatusThread.DoRingSignal;
begin
  FComPort.OnRingSignal(FComPort);
end;

procedure TStatusThread.DoRLSDSignal;
begin
  FComPort.FOnRLSDSignal(FComPort);
end;

procedure TStatusThread.DoTxEmptySignal;
begin
  FComPort.FOnTxEmptySignal(FComPort);
end;

{
  Errorflags for OnErrorSignal
  
  CE_BREAK -     The hardware detected a break condition.
  CE_DNS -       Windows 95 only: A parallel device is not selected.
  CE_FRAME -     The hardware detected a framing error.
  CE_IOE -       An I/O error occurred during communications with the
                 device.
  CE_MODE -      The requested mode is not supported, or the hFile parameter
                 is invalid. If this value is specified, it is the only
                 valid error.
  CE_OOP -       Windows 95 only: A parallel device signaled that it is out
                 of paper.
  CE_OVERRUN -   A character-buffer overrun has occurred. The next character
                 is lost.
  CE_PTO -       Windows 95 only: A time-out occurred on a parallel device.
  CE_RXOVER -    An input buffer overflow has occurred. There is either no
                 room in the input buffer, or a character was received after
                 the EOF (end-of-file) character.
  CE_RXPARITY -  The hardware detected a parity error.
  CE_TXFULL -    The application tried to transmit a character, but the
                 output buffer was full.
}
procedure TStatusThread.HandleCommEvent;
var
  ComStat: TComStat;
begin
  FErrors := 0;
  if FComPort.Enabled then
    ClearCommError(FComHandle, FErrors, @ComStat);
  if FEvtMask and EV_BREAK > 0 then
    if Assigned(FComPort.FOnBreakSignal) then Synchronize(DoBreakSignal);
  if FEvtMask and EV_CTS > 0 then
    if Assigned(FComPort.FOnCTSSignal) then Synchronize(DoCTSSignal);
  if FEvtMask and EV_DSR > 0 then
    if Assigned(FComPort.FOnDSRSignal) then Synchronize(DoDSRSignal);
  if FEvtMask and EV_ERR > 0 then
    if Assigned(FComPort.FOnErrorSignal) then Synchronize(DoErrorSignal);
  if FEvtMask and EV_RING > 0 then
    if Assigned(FComPort.FOnRingSignal) then Synchronize(DoRingSignal);
  if FEvtMask and EV_RLSD > 0 then
    if Assigned(FComPort.FOnRLSDSignal) then Synchronize(DoRLSDSignal);
  if FEvtMask and EV_TXEMPTY > 0 then
    if Assigned(FComPort.FOnTxEmptySignal) then Synchronize(DoTxEmptySignal);
end;

procedure TStatusThread.Execute;
var
  OK: Boolean;
  ModemStat: DWORD;
  BytesTransferred: DWORD; // Dummy, not valid for WaitCommEvent
begin
  { Get comm signals initial state }
  if FComPort.Enabled and GetCommModemStatus(FComHandle, ModemStat) then
  begin
    FComPort.FCTS := ModemStat and MS_CTS_ON <> 0;
    FComPort.FDSR := ModemStat and MS_DSR_ON <> 0;
    FComPort.FRing := ModemStat and MS_RING_ON <> 0;
    FComPort.FRLSD := ModemStat and MS_RLSD_ON <> 0;
    FEvtMask := EV_CTS or EV_DSR or EV_RING or EV_RLSD;
    HandleCommEvent;
  end;
  { Status thread started }
  SetEvent(FComPort.FStatusEvent);
  { Thread circle }
  repeat
    if FcomPort.FKillThreads then Break;
    OK := WaitCommEvent(FComHandle, FEvtMask, @FSOL);
    if not OK then Sleep(FDelay);
    if FComPort.FKillThreads then Break;
    if (not OK) and (GetLastError = ERROR_IO_PENDING) then
    begin
      OK := GetOverlappedResult(FComHandle, FSOL, BytesTransferred, True);
      if OK then ResetEvent(FSOL.hEvent);
    end;
    if FComPort.FKillThreads then Break;
    if OK and ((FEvtMask or EV_RXCHAR) > 0) then
    begin
      if FComPort.Enabled then SetEvent(FComPort.FReadReadyEvent);
      if FComPort.FInBuffUsed > 0 then
        if Assigned(FComPort.FOnDataAvail) then Synchronize(DoDataAvail);
    end;
    if OK and (FEvtMask and (EV_CTS or EV_DSR or EV_RING or EV_RLSD or EV_ERR) <> 0) and GetCommModemStatus(FComHandle, ModemStat) then
    begin
      FComPort.FCTS := ModemStat and MS_CTS_ON <> 0;
      FComPort.FDSR := ModemStat and MS_DSR_ON <> 0;
      FComPort.FRing := ModemStat and MS_RING_ON <> 0;
      FComPort.FRLSD := ModemStat and MS_RLSD_ON <> 0;
    end;
    HandleCommEvent;
  until FComPort.FKillThreads or Terminated;
  SetEvent(FComPort.FStatusEvent);
end;

{ TWriteThread }

constructor TWriteThread.Create(ComPort: TCustomCommPort);
begin
  inherited Create(True);
  FComPort := ComPort;
  FComHandle := FComPort.ComHandle;
  FDelay := ComPort.WriteDelay;
  FillChar(FWOL, SizeOf(FWOL), 0);
  FWOL.hEvent := CreateEvent(nil, True, False, nil);
end;

destructor TWriteThread.Destroy;
begin
  CloseHandle(FWOL.hEvent);
  inherited;
end;

procedure TWriteThread.Execute;
var
  BytesToWrite, BytesWritten: DWORD;
  lpErrors: DWORD;
  lpStat: TComStat;

  procedure WriteSuccess;
  begin
    if BytesWritten < FComPort.FOutBuffUsed then
      Move(FComPort.FOutBuffer[BytesWritten], FComPort.FOutBuffer[0], FComPort.FOutBuffUsed - BytesWritten);
    if BytesWritten >= FComPort.FOutBuffUsed then
      FComPort.FOutBuffUsed := 0
    else
      Dec(FComPort.FOutBuffUsed, BytesWritten);
  end;

begin
  SetEvent(FComPort.FWriteEvent);
  repeat
    if FComPort.FKillThreads then Break;
    if (FComPort.FOutBuffUsed > 0) then
    begin
      EnterCriticalSection(FComPort.FWriteSection);
      ClearCommError(FComPort.ComHandle, lpErrors, @lpStat);
      BytesToWrite := FComPort.FOutBuffUsed;
      if BytesToWrite > (FComPort.FOutBuffSize - lpStat.cbOutQue) then
        BytesToWrite := (FComPort.FOutBuffSize - lpStat.cbOutQue);
      if FComPort.Enabled and (BytesToWrite > 0) and (not FComPort.FKillThreads) and FComPort.CTS then
      begin
        if WriteFile(FComHandle, FComPort.FOutBuffer^[0], BytesToWrite, BytesWritten, @FWOL) then
          WriteSuccess
        else
          if GetLastError = ERROR_IO_PENDING then
            if GetOverlappedResult(FComHandle, FWOL, BytesWritten, True) then
            begin
              ResetEvent(FWOL.hEvent);
              WriteSuccess;
            end;
      end;
      LeaveCriticalSection(FComPort.FWriteSection);
    end;
    Sleep(FDelay);
  until FComPort.FKillThreads or Terminated;
  SetEvent(FComPort.FWriteEvent);
end;

{ TReadThread }

constructor TReadThread.Create(ComPort: TCustomCommPort);
begin
  inherited Create(True);
  FComPort := ComPort;
  FComHandle := FComPort.ComHandle;
  FDelay := ComPort.ReadDelay;
  FillChar(FROL, SizeOf(FROL), 0);
  FROL.hEvent := CreateEvent(nil, True, False, nil);
end;

destructor TReadThread.Destroy;
begin
  CloseHandle(FROL.hEvent);
  inherited;
end;

procedure TReadThread.Execute;
var
  BytesToRead, BytesReaded: DWORD;
  lpErrors: DWORD;
  lpStat: TComStat;

  procedure ReadSuccess;
  begin
    Inc(FComPort.FInBuffUsed, BytesReaded);
    if FComPort.FInBuffUsed > FComPort.FInBuffSize then
      FComPort.FInBuffUsed := FComPort.FInBuffSize;
  end;

begin
  SetEvent(FComPort.FReadEvent);
  repeat
    if WaitForSingleObject(FcomPort.FReadReadyEvent, FDelay) = WAIT_OBJECT_0 then
    begin
      if FComPort.FKillThreads then Break;
      if FComPort.Enabled then
        ClearCommError(FComHandle, lpErrors, @lpStat)
      else
      begin
        lpErrors := 0;
        FillChar(lpStat, SizeOf(LpStat), 0);
      end;
      EnterCriticalSection(FComPort.FReadSection);
//      BytesToRead := FComPort.FInBuffSize-FComPort.FInBuffUsed;
      BytesToRead := lpStat.cbInQue;
      if BytesToRead > FComPort.FInBuffSize - FComPort.FInBuffUsed then
        BytesToRead := FComPort.FInBuffSize - FComPort.FInBuffUsed;
      BytesReaded := 0;
      if (BytesToRead > 0) and FComPort.Enabled then
      begin
        if ReadFile(FComHandle, FComPort.FInBuffer^[FComPort.FInBuffUsed], BytesToRead, BytesReaded, @FROL) then
          ReadSuccess
        else
          if GetLastError = ERROR_IO_PENDING then
            if GetOverlappedResult(FComHandle, FROL, BytesReaded, True) then
            begin
              ResetEvent(FROL.hEvent);
              ReadSuccess;
            end;
      end;
      LeaveCriticalSection(FComPort.FReadSection);
      if FComPort.FInBuffUsed > 0 then
        if Assigned(FComPort.FOnDataAvail) then  Synchronize(DoOnDataAvail);
    end
    else
      SetEvent(FComPort.FReadReadyEvent);
  until FComPort.FKillThreads or Terminated;
  SetEvent(FComPort.FReadEvent);
end;

procedure TReadThread.DoOnDataAvail;
begin
  FComPort.OnDataAvail(FComPort, FComPort.FInBuffUsed);
end;

{ TCustomCommPort }

constructor TCustomCommPort.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FComHandle := INVALID_HANDLE_VALUE;
  FComNumber := cnCOM2;
  FBaudRate := br19200;
  FParity := paNone;
  FStopBits := sb1_0;
  FDataBits := db8;
  FInBuffSize := 4096;
  FOutBuffSize := 4096;
  InitializeCriticalSection(FWriteSection);
  InitializeCriticalSection(FReadSection);
  FReadReadyEvent := CreateEvent(nil, False, False, nil);
  FStatusEvent := CreateEvent(nil, False, False, nil);
  FWriteEvent := CreateEvent(nil, False, False, nil);
  FReadEvent := CreateEvent(nil, False, False, nil);
end;

destructor TCustomCommPort.Destroy;
begin
  ClosePort;
  CloseHandle(FStatusEvent);
  CloseHandle(FWriteEvent);
  CloseHandle(FReadEvent);
  CloseHandle(FReadReadyEvent);
  DeleteCriticalSection(FReadSection);
  DeleteCriticalSection(FWriteSection);
  inherited Destroy;
end;

function TCustomCommPort.GetInBuffFree: DWORD;
begin
  EnterCriticalSection(FReadSection);
  Result := FInBuffSize - FInBuffUsed;
  LeaveCriticalSection(FReadSection);
end;

function TCustomCommPort.GetInBuffUsed: DWORD;
begin
  Result := FInBuffUsed;
end;

function TCustomCommPort.GetOutBuffFree: DWORD;
begin
  Result := FOutBuffSize - FOutBuffUsed;
end;

function TCustomCommPort.GetOutBuffUsed: DWORD;
begin
  Result := FOutBuffUsed;
end;

function TCustomCommPort.GetOutQueCount: DWORD;
var
  lpErrors: DWORD;
  lpStat: TComStat;
begin
  Result := 0;
  if Enabled then
  begin
    ClearCommError(FComHandle, lpErrors, @lpStat);
    Result := lpStat.cbOutQue;
  end;
end;

procedure TCustomCommPort.SetBaudRate(Value: TBaudRate);
const
  CBR: array[TBaudRate] of Integer = (
    CBR_110,
    CBR_300,
    CBR_600,
    CBR_1200,
    CBR_2400,
    CBR_4800,
    CBR_9600,
    CBR_14400,
    CBR_19200,
    CBR_38400,
    CBR_56000,
    CBR_57600,
    CBR_115200,
    CBR_128000,
    CBR_256000);
const
  DCBF_BINARY              = $00000001;
  DCBF_OUTXCTSFLOW         = $00000004;
  DCBF_OUTXDSRFLOW         = $00000008;
  DCBF_DTRCONTROLHANDSHAKE = $00000020;
  DCBF_RTSCONTROLHANDSHAKE = $00002000;
  DCBF_INHDWFLOW           = DCBF_OUTXDSRFLOW OR DCBF_DTRCONTROLHANDSHAKE;
  DCBF_OUTHDWFLOW          = DCBF_OUTXCTSFLOW OR DCBF_RTSCONTROLHANDSHAKE;
  DCBF_HDWFLOW             = DCBF_INHDWFLOW or DCBF_OUTHDWFLOW;
begin
  FBaudRate := Value;
  if Enabled then
  begin
    GetCommState(FComHandle, FComDCB);
    FComDCB.BaudRate := CBR[FBaudRate];
    FComDCB.Flags := DCBF_BINARY or DCBF_HDWFLOW;
    SetCommState(FComHandle, FComDCB);
  end;
end;

procedure TCustomCommPort.SetDataBits(Value: TDataBits);
const
  DAB: array[TDataBits] of byte = (4, 5, 6, 7, 8);
begin
  FDataBits := Value;
  if Enabled then
  begin
    GetCommState(FComHandle, FComDCB);
    FComDCB.Bytesize := DAB[Value];
    SetCommState(FComHandle, FComDCB);
  end;
end;

procedure TCustomCommPort.SetDTR(Value: Boolean);
begin
  SetDTRState(Value);
end;

procedure TCustomCommPort.SetInBuffSize(Value: DWORD);
begin
  if FInBuffer = nil then FInBuffSize := Value;
end;

procedure TCustomCommPort.SetOpened(Value: Boolean);
begin
  if Value then OpenPort else ClosePort;
end;

procedure TCustomCommPort.SetOutBuffSize(Value: DWORD);
begin
  if FOutBuffer = nil then FOutBuffSize := Value;
end;

procedure TCustomCommPort.SetParity(Value: TParity);
const
  PAR: array[TParity] of byte = (
    NOPARITY,
    ODDPARITY,
    EVENPARITY,
    MARKPARITY,
    SPACEPARITY);
begin
  FParity := Value;
  if Enabled then
  begin
    GetCommState(FComHandle, FComDCB);
    FComDCB.Parity := PAR[FParity];
    SetCommState(FComHandle, FComDCB);
  end;
end;

procedure TCustomCommPort.SetRTS(Value: Boolean);
begin
  SetRTSState(Value);
end;

procedure TCustomCommPort.SetStopBits(Value: TStopBits);
const
  STB: array[TStopBits] of byte = (
    ONESTOPBIT,
    ONE5STOPBITS,
    TWOSTOPBITS);
begin
  FStopBits := Value;
  if Enabled then
  begin
    GetCommState(FComHandle, FComDCB);
    FComDCB.StopBits := STB[Value];
    SetCommState(FComHandle, FComDCB);
  end;
end;

function TCustomCommPort.SetBREAKState(State: Boolean): Boolean;
const
  BREAK: array[Boolean] of Integer = (CLRBREAK, SETBREAK);
begin
  Result := False;
  if FOpened then
    Result := EscapeCommFunction(FComHandle, BREAK[State]);
end;

function TCustomCommPort.SetDTRState(State: Boolean): Boolean;
const
  DTR: array[Boolean] of Integer = (CLRDTR, Windows.SETDTR);
begin
  Result := False;
  FDTR := State;
  if FOpened then
    Result := EscapeCommFunction(FComHandle, DTR[State]);
  if Assigned(FOnDTRSignal) then
    if Assigned(FStatusThread) then FStatusThread.Synchronize(FStatusThread.DoDTRSignal);
end;

function TCustomCommPort.SetRTSState(State: Boolean): Boolean;
const
  RTS: array[Boolean] of Integer = (CLRRTS, Windows.SETRTS);
begin
  Result := False;
  FRTS := State;
  if FOpened then
    Result := EscapeCommFunction(FComHandle, RTS[State]);
end;

function TCustomCommPort.CharReady: Boolean;
begin
  Result := FInBuffUsed > 0;
end;

procedure TCustomCommPort.ClosePort;
begin
  FKillThreads := True;
  { Check device already closed }
  if FOpened then
  begin
    { Clear signal flags }
    FTxD := False;
    FRxD := False;
    FRing := False;
    FRLSD := False;
//    FlushOutBuffer;
//    FlushInBuffer;
    { Terminate write thread }
    FWriteThread.Terminate;
    FWriteThread.WaitFor;
    FWriteThread.Free;
    FWriteThread := nil;
    { Terminate read thread }
    SetEvent(FReadReadyEvent);
    FReadThread.Terminate;
    FReadThread.WaitFor;
    FReadThread.Free;
    FReadThread := nil;
    { Clear set of monitored events }
    SetCommMask(FComHandle, 0);
    { Terminate status thread }
    FStatusThread.Terminate;
    FStatusThread.WaitFor;
    FStatusThread.Free;
    FStatusThread := nil;
    { Close device }
    CloseHandle(FComHandle);
    FComHandle := INVALID_HANDLE_VALUE;
    FDTR := False;
    { Free buffers }
    FreeMem(FOutBuffer, FOutBuffSize);
    FreeMem(FInBuffer, FInBuffSize);
    FInBuffer := nil;
    FOutBuffer := nil;
    FInBuffUsed := 0;
    FOutBuffUsed := 0;
    { Success }
    FOpened := False;
    { User events }
    if Assigned(FOnClose) then FOnClose(Self);
  end;
end;

function TCustomCommPort.Enabled: Boolean;
begin
  Result := FComHandle <> INVALID_HANDLE_VALUE;
end;

procedure TCustomCommPort.FlushInBuffer;
begin
  { Purge comm }
  if Enabled then PurgeComm(FComHandle, PURGE_RXABORT or PURGE_RXCLEAR);
  { Reset in buffer used size }
  EnterCriticalSection(FReadSection);
  FInBuffUsed := 0;
  LeaveCriticalSection(FReadSection);
end;

procedure TCustomCommPort.FlushOutBuffer;
begin
  { Purge comm }
  if Enabled then PurgeComm(FComHandle, PURGE_TXABORT or PURGE_TXCLEAR);
  { Reset out buffer used size }
  EnterCriticalSection(FWriteSection);
  FOutBuffUsed := 0;
  LeaveCriticalSection(FWriteSection);
end;

function TCustomCommPort.GetBlock(var Buf; Count: DWORD): DWORD;
begin
  Result := 0;
  { Check com handle }
  if Enabled then
  begin
    EnterCriticalSection(FReadSection);
    { Check read count }
    Result := Count;
    if Count > FInBuffUsed then Result := FInBuffUsed;
    { Get data brom buffer }
    Move(FInBuffer[0], Buf, Result);
    Move(FInBuffer[Result], FInBuffer[0], FInBuffUsed - Result);
    { Update buffer size }
    Dec(FInBuffUsed, Result);
    LeaveCriticalSection(FReadSection);
  end;
end;

function TCustomCommPort.GetChar: Char;
begin
  Result := #0;
  if FInBuffUsed > 0 then GetBlock(Result, 1)
end;

procedure TCustomCommPort.OpenPort;
var
  DeviceName: string;
  CommTimeouts: TCommTimeouts;
begin
  if not FOpened then
  begin
    { Allicate buffers }
    GetMem(FOutBuffer, FOutBuffSize);
    GetMem(FInBuffer, FInBuffSize);
    { Clear error }
    SetLastError(0);
    { Open device }
    DeviceName := '\\.\COM' + IntToStr(Ord(FComNumber) + 1);
    FComHandle := CreateFile(PCHAR(DeviceName),
      GENERIC_READ or GENERIC_WRITE,
      0, nil, OPEN_EXISTING,
      FILE_ATTRIBUTE_NORMAL or FILE_FLAG_OVERLAPPED, 0);
    { Check device open }
    if Enabled then
    begin
      { Initializes the communications parameters }
      SetupComm(FComHandle, FInBuffSize, FOutBuffSize);
      { Sets the time-out parameters }
      FillChar(CommTimeOuts, SizeOf(TCommTimeouts), 0);
  //    CommTimeouts.ReadIntervalTimeout := MAXDWORD;
      SetCommTimeOuts(FComHandle, CommTimeOuts);
      { Set comm params }
      SetBaudrate(FBaudrate);
      SetParity(FParity);
      SetStopBits(FStopBits);
      SetDataBits(FDataBits);
      FKillThreads := False;
      { Start and wait status thread }
      FStatusThread := TStatusThread.Create(Self);
      FStatusThread.Resume;
      WaitForSingleObject(FStatusEvent, 1000);
      { Set monitored events }
      SetCommMask(FComHandle, EV_BREAK or EV_CTS or EV_DSR or EV_ERR or
        EV_RING or EV_RLSD or EV_RXCHAR or EV_TXEMPTY);
      { Start and wait read thread }
      FReadThread := TReadThread.Create(Self);
      FReadThread.Resume;
      WaitForSingleObject(FReadEvent, 1000);
      { Start and wait write thread }
      FWriteThread := TWriteThread.Create(Self);
      FWriteThread.Resume;
      WaitForSingleObject(FWriteEvent, 1000);
      { Success }
      FOpened := True;
      { User event }
      if Assigned(FOnOpen) then FOnOpen(Self);
    end;
  end;
end;

function TCustomCommPort.PutBlock(const Buf; Count: DWORD): DWORD;
begin
  Result := 0;
  { Check com handle }
  if Enabled then
  begin
    EnterCriticalSection(FWriteSection);
    { Check write count }
    Result := Count;
    if FOutBuffSize - FOutBuffUsed < Count then Result := FOutBuffSize - FOutBuffUsed;
    { Put data into buffer }
    Move(Buf, FOutBuffer[FOutBuffUsed], Result);
    { Update buffer size }
    Inc(FOutBuffUsed, Result);
    LeaveCriticalSection(FWriteSection);
  end;
end;

function TCustomCommPort.PutChar(C: Char): DWORD;
begin
  Result := PutBlock(C, 1);
end;

function TCustomCommPort.PutString(const S: string): DWORD;
begin
  Result := 0;
  if Length(S) > 0 then Result := PutBlock(S[1], Length(S));
end;

function TCustomCommPort.SendBreak(Ticks: Word; Yield: Boolean): Boolean;
begin
  Result := SetBREAKState(True);
  if Result then
  begin
    Sleep(Ticks * 55);
    Result := SetBREAKState(False);
  end;
end;

end.

