{
  Voice Modem Library v1.0

  This file contains specific hardware stuff for Cirrus Logic CL-MD1414XXX
  family voice modems.

  Based on a original source code by Mitch <Mitch.DSouza@NetComm.IE>.

  Roman M. Mochalov
  E-mail: roman@sar.nnov.ru
}

unit VML_CLMD1414;

interface

uses
  Windows, SysUtils, Classes, VML_Modem, VML_Consts, VML_IS101, VML_Idents;

const

{ CL-MD1414 available compression modes }

  COMPRESSION_CLMD1414_CL1     = 1; {?}
  COMPRESSION_CLMD1414_ADPCM_3 = 3;
  COMPRESSION_CLMD1414_ADPCM_4 = 4;

type

{ TCirrusLogicMD1414 }

  TCirrusLogicMD1414 = class(TIS101)
  public
    constructor Create(AOwner: TComponent); override;
    function AnswerPhone: Integer; override;
    class function GetCompressionMap(var Map: PIdentMap): Integer; override;
    class function GetDeviceMap(var Map: PIdentMap): Integer; override;
    function Init: Integer; override;
    function InitVoice: Integer; override;
    function HandleDLE(Code: Char): Integer; override;
    class function ModemName: string; override;
    class function ModemRMDName: string; override;
    function SetCompression(var Compression, Speed, Bits: Integer): Integer; override;
    function SetDevice(Device: Integer): Integer; override;
  end;

implementation

{$I *.INC}

{ Some internal constants }

const
  CompressionMap: array[0..2] of TIdentMapEntry = (
    (Value: COMPRESSION_CLMD1414_CL1; Name: SCL1),
    (Value: COMPRESSION_CLMD1414_ADPCM_3; Name: SADPCM3),
    (Value: COMPRESSION_CLMD1414_ADPCM_4; Name: SADPCM4));

  DeviceMap: array[0..3] of TIdentMapEntry = (
    (Value: DIALUP_LINE; Name: SDialupLine),
    (Value: EXTERNAL_MICROPHONE; Name: SExternalMicrophone),
    (Value: INTERNAL_SPEAKER; Name: SInternalSpeaker),
    (Value: LOCAL_HANDSET; Name: SLocalHandset));

{ TCirrusLogicMD1414 }

constructor TCirrusLogicMD1414.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  HardFlowCmd := 'AT';
  SoftFlowCmd := 'AT';
  BeepCmd := 'AT#VBP';
  StartPlayCmd := 'AT#VPY';
  StartPlayAnsw := 'OK|CONNECT';
  IntrPlayCmd := DLE + 'A' + DLE + ETX;
  IntrPlayAnsw := 'OK|VCON';
  StopPlayAnsw := 'OK|VCON';
  StartRecCmd := 'AT#VRD';
  StartRecAnsw := 'OK|CONNECT';
  StopRecCmd := CR;
  StopRecAnsw := 'OK|VCON';
  VoiceModeOnCmd := 'AT#VCL=1';
  VoiceModeOffCmd := 'AT#VCL=0';
end;

function TCirrusLogicMD1414.AnswerPhone: Integer;
begin
  { Check modem specified }
  if not ModemAssigned('AnswerPhone', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Set dialup line device }
  if Modem.Command('AT#VLN=1', 'OK') <>  VMA_USER_1 then
  begin
    Result := VMA_ERROR;
    Exit;
  end;
  { Answer call }
  if Modem.Command('AT#VIP=1', 'OK') <> VMA_USER_1 then
  begin
    Result := VMA_ERROR;
    Exit;
  end;
  { Success }
  Result := VMA_OK;
end;

class function TCirrusLogicMD1414.GetCompressionMap(var Map: PIdentMap): Integer;
begin
  Map := @CompressionMap;
  Result := High(CompressionMap) - Low(CompressionMap) + 1;
end;

class function TCirrusLogicMD1414.GetDeviceMap(var Map: PIdentMap): Integer;
begin
  Map := @DeviceMap;
  Result := High(DeviceMap) - Low(DeviceMap) + 1;
end;

function TCirrusLogicMD1414.Init: Integer;
begin
  DebugMsgFmt(L_MESG, SInitModem, [ModemName]);
  { Check modem specified }
  if not ModemAssigned('Init', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Update modem state }
  Modem.State := INITIALIZING;
  try
    { AT\Q3 - Bidirectional hardware flow control }
    if Modem.Command('AT\Q3', 'OK') <> VMA_USER_1 then
    begin
      DebugMsg(L_WARN, SFlowControlError);
      Result := FAIL;
      Exit;
    end;
    { Set extended verbal result messages }
    if Modem.Command('ATX4', 'OK') <> VMA_USER_1 then
    begin
      DebugMsg(L_WARN, SExResultCodesError);
      Result := FAIL;
      Exit;
    end;
  finally
    Modem.State := IDLE;
  end;
  { Init complete }
  Result := OK;
end;

function TCirrusLogicMD1414.InitVoice: Integer;
var
  TransmitGain, ReceiveGain: Integer;
begin
  DebugMsgFmt(L_MESG, SInitVoice, ['Cirrus Logic ' + ModemName]);
  { Check modem specified }
  if not ModemAssigned('InitVoice', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Update modem state }
  Modem.State := INITIALIZING;
  try
    { Set the record volume range }
    ReceiveGain := Profile.ReceiveGain;
    if ReceiveGain = -1 then ReceiveGain := 127;
    if Modem.CommandFmt('AT#VRL=%d', 'OK', [ReceiveGain * (131 - 121) div 100 + 121]) <> VMA_USER_1 then
      DebugMsg(L_WARN, SReceiveGainError);
    { Set the play volume }
    TransmitGain := Profile.TransmitGain;
    if TransmitGain = -1 then TransmitGain := 127;
    if Modem.CommandFmt('AT#VPL=%d', 'OK', [TransmitGain * (131 - 121) div 100 + 121]) <> VMA_USER_1 then
      DebugMsg(L_WARN, STransmitGainError);
    { Set the silence threshold }
    if Modem.CommandFmt('AT#VSL=%d', 'OK', [Profile.RecSilenceSens * (131 - 121) div 100 + 121]) <> VMA_USER_1 then
      DebugMsg(L_WARN, SSilenceThresholdError);
    { Set the silence timers }
    if Modem.CommandFmt('AT#VSQT=%d', 'OK', [Profile.RecSilenceLen]) <> VMA_USER_1 then
      DebugMsg(L_WARN, SSilenceQLenError);
    if Modem.CommandFmt('AT#VSST=%d', 'OK', [Profile.RecSilenceLen]) <> VMA_USER_1 then
      DebugMsg(L_WARN, SSilenceSLenError);
    { Set silence deletion }
    if Modem.CommandFmt('AT#VCSD=%d', 'OK', [Ord(Profile.RecRemoveSilence)]) <> VMA_USER_1 then
      DebugMsg(L_WARN, SSilenceDetectionError);
  finally
    Modem.State := IDLE;
  end;
  { Success }
  Result := OK;
end;

function TCirrusLogicMD1414.HandleDLE(Code: Char): Integer;
begin
  { Check modem specified }
  if not ModemAssigned('HandleDLE', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Handle DLE codes }
  case Code of
    'E':
      { DTMF digit * }
      begin
        Result := Modem.QueueEvent(TModemEvent.CreateDataC(RECEIVED_DTMF, '*'));
        Exit;
      end;
    'F':
      { DTMF digit # }
      begin
        Result := Modem.QueueEvent(TModemEvent.CreateDataC(RECEIVED_DTMF, '#'));
        Exit;
      end;
    'N':
      { 1100-Hz fax CNG tone }
      begin
        Result := Modem.QueueEvent(TModemEvent.Create(FAX_CALLING_TONE));
        Exit;
      end;
    'g':
      { 1300-Hz data calling tone (CL-MD1414XX only) }
      begin
        Result := Modem.QueueEvent(TModemEvent.Create(DATA_CALLING_TONE));
        Exit;
      end;
    'l', 'L':
      { Remote hang-up }
      begin
        Result := Modem.QueueEvent(TModemEvent.Create(REMOTE_HANGUP));
        Exit;
      end;
  end;
  { Standard handling }
  Result := inherited HandleDLE(Code);
end;

class function TCirrusLogicMD1414.ModemName: string;
begin
  Result := 'Cirrus Logic CL-MD1414';
end;

class function TCirrusLogicMD1414.ModemRMDName: string;
begin
  Result := 'CL-MD1414';
end;

function TCirrusLogicMD1414.SetCompression(var Compression, Speed, Bits: Integer): Integer;
begin
  { Check modem specified }
  if not ModemAssigned('SetCompression', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { The default mode for CL-MD1414 modems is CL1 }
  if Compression = 0 then Compression := COMPRESSION_CLMD1414_CL1;
  { 9600 Hz is currently the only recording freq supported by CL-MD1414 }
  if Speed = 0 then Speed := 9600;
  if Speed <> 9600 then
  begin
    DebugMsgFmt(L_WARN, SIllegalSampleRate, [Speed]);
    Result := FAIL;
    Exit;
  end;
  { Setup sampling rate }
  if Modem.Command('AT#VSR=9600', 'OK') <> VMA_USER_1 then
  begin
    Result := FAIL;
    Exit;
  end;
  { Setup compression }
  case Compression of
    COMPRESSION_CLMD1414_ADPCM_3:
      { 3 bit ADPCM }
      begin
        Bits := 3;
        if Modem.Command('AT#VSM=AD3', 'OK') <> VMA_USER_1 then
        begin
          Result := VMA_FAIL;
          Exit;
        end;
      end;
    COMPRESSION_CLMD1414_ADPCM_4:
      { 4 bit ADPCM }
      begin
        Bits := 4;
        if Modem.Command('AT#VSM=AD4', 'OK') <> VMA_USER_1 then
        begin
          Result := VMA_FAIL;
          Exit;
        end;
      end;
    COMPRESSION_CLMD1414_CL1:
      { ? }
      begin
        Bits := 5;
        if Modem.Command('AT#VSM=CL1', 'OK') <> VMA_USER_1 then
        begin
          Result := VMA_FAIL;
          Exit;
        end;
      end;
    else
      begin
        DebugMsgFmt(L_WARN, SIllegalCompressionMethod, [Compression]);
        Modem.LastError := VME_ILLEGAL_COMPRESSION;
        Result := FAIL;
        Exit;
      end;
  end;
  Result := OK;
  Exit;
end;

function TCirrusLogicMD1414.SetDevice(Device: Integer): Integer;
begin
  { Check modem specified }
  if not ModemAssigned('SetDevice', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Set device }
  case Device of
    NO_DEVICE:
      begin
        Modem.Command('AT#VLN=0', 'OK');
        Result := OK;
        Exit;
      end;
    DIALUP_LINE:
      begin
        Modem.Command('AT#VLN=1', 'OK');
        Result := OK;
        Exit;
      end;
    EXTERNAL_MICROPHONE:
      begin
        Modem.Command('AT#VLN=32', 'OK');
        Result := OK;
        Exit;
      end;
    INTERNAL_SPEAKER:
      begin
        Modem.Command('AT#VLN=16', 'OK');
        Result := OK;
        Exit;
      end;
    LOCAL_HANDSET:
      begin
        Modem.Command('AT#VLN=2', 'OK');
        Result := OK;
        Exit;
      end;
  end;
  { Given device unknown }
  DebugMsgFmt(L_WARN, SUnknownDevice, [Device]);
  Modem.LastError := VME_UNKNOWN_DEVICE;
  Result := FAIL;
end;

initialization
  RegisterModemDriver(TCirrusLogicMD1414.ModemName, TCirrusLogicMD1414);

end.
