{
  Voice Modem Library v1.0

  This file contains generic hardware driver functions for modems that
  follow the IS-101 interim standard for voice modems. Since the commands
  are set in the modem structure, it should be quite generic.

  Roman M. Mochalov
  E-mail: roman@sar.nnov.ru      
}

unit VML_IS101;

interface

uses
  Windows, SysUtils, Classes, VML_Modem, VML_Consts, VML_NoModem;

type

{ TIS101 }

  TIS101 = class(TNoModem)
  protected
    FStopDialing: Boolean;
    FStopPlaying: Boolean;
    FStopRecording: Boolean;
    FStopWaiting: Boolean;
  public
    constructor Create(AOwner: TComponent); override;
    function AnswerPhone: Integer; override;
    function Beep(Frequency, Length: Integer): Integer; override;
    function Dial(const Number: string): Integer; override;
    class function GetVoiceLen(Compression, Speed, Bits: Integer; DataSize: Longint): Integer; override;
    function HandleDLE(Code: Char): Integer; override;
    function HandleEvent(Event: Integer; Data: TEventData): Integer; override;
    function HangUp: Integer; override;
    class function ModemName: string; override;
    class function ModemRMDName: string; override;
    function PlayDTMF(const Number: string): Integer; override;
    function PlayStream(Stream: TStream; BPS: Integer): Integer; override;
    function RecordStream(Stream: TStream; BPS: Integer): Integer; override;
    function RestoreVoiceMode: Integer; override;
    function SetCompression(var Compression, Speed, Bits: Integer): Integer; override;
    function SetDevice(Device: Integer): Integer; override;
    function StartPlayStream: Integer; override;
    function StartRecordStream: Integer; override;
    function StopDialing: Integer; override;
    function StopPlayStream: Integer; override;
    function StopPlaying: Integer; override;
    function StopRecordStream: Integer; override;
    function StopRecording: Integer; override;
    function StopWaiting: Integer; override;
    function SwitchToDataFax(const Mode: string): Integer; override;
    function VoiceModeOff: Integer; override;
    function VoiceModeOn: Integer; override;
    function Wait(TimeOut: DWORD): Integer; override;
    function WaitAnswer(const ExpectedAnswers: string; TimeOut: DWORD): Integer; override;
  end;

implementation

{$I *.INC}

{ TIS101 }

constructor TIS101.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  HangUpCmd := 'ATH0';
  HangUpAnsw := 'OK';
  PickPhoneCmd := 'AT+VLS=2A';
  PickPhoneAnsw := 'OK';
  BeepCmd := 'AT+VTS=[%d,0,%d]';
  BeepAnsw := 'OK';
  BeepTimeUnit := 10;
  HardFlowCmd := 'AT+FLO=2';
  HardFlowAnsw := 'OK';
  SoftFlowCmd := 'AT+FLO=1';
  SoftFlowAnsw := 'OK';
  StartPlayCmd := 'AT+VTX';
  StartPlayAnsw := 'CONNECT';
  IntrPlayCmd := DLE + CAN + DLE + ETX;
  IntrPlayAnsw := 'OK';
  StopPlayCmd := DLE + ETX;
  StopPlayAnsw := 'OK';
  StartRecCmd := 'AT+VRX';
  StartRecAnsw := 'CONNECT';
  StopRecCmd := DLE + '!';
  StopRecAnsw := 'OK';
  SwitchModeCmd := 'AT+FCLASS=%s';
  SwitchModeAnsw := 'OK';
  VoiceModeOnCmd := 'AT+FCLASS=8';
  VoiceModeOnAnsw := 'OK';
  VoiceModeOffCmd := 'AT+FCLASS=0';
  VoiceModeOffAnsw := 'OK';
  PlayDTMFCmd := 'AT+VTS=%s';
  PlayDTMFExtra := ',%s';
  PlayDTMFAnsw := 'OK';
end;

function TIS101.AnswerPhone: Integer;
begin
  { Check modem specified }
  if not ModemAssigned('AnswerPhone', True) then
  begin
    Result := VMA_ERROR;
    Exit;
  end;
  { Answer Phone }
  if Modem.Command(PickPhoneCmd, PickPhoneAnsw) <> VMA_USER_1 then
  begin
    Result := VMA_ERROR;
    Exit;
  end;
  Result := VMA_OK;
end;

function TIS101.Beep(Frequency, Length: Integer): Integer;
var
  TimeOut, StartTime: DWORD;
begin
  { Check modem specified }
  if not ModemAssigned('Beep', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Write beep command }
  if Modem.CommandFmt(BeepCmd, '', [Frequency, Length div BeepTimeUnit]) <> OK then
  begin
    Result := FAIL;
    Exit;
  end;
  TimeOut := Length;
  StartTime := GetTickCount;
  { Wait modem answer }
  while (TimeOut = 0) or (DWORD(Abs(GetTickCount - StartTime)) < TimeOut) do
  begin
    if Modem.Port.CharReady then Break;
    SleepEx(Profile.PortReadDelay, False);
  end;
  { Read answer }
  if Modem.Command('', BeepAnsw) <> VMA_USER_1 then
  begin
    Result := FAIL;
    Exit;
  end;
  { Success }
  Result := OK;
end;

function TIS101.Dial(const Number: string): Integer;
var
  Command, Answer: string;
  TimeOut, StartTime: DWORD;
  DialComplete: Boolean;
begin
  { Check modem specified }
  if not ModemAssigned('Dial', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Update modem state }
  Modem.State := DIALING;
  try
    FStopDialing := False;
    { Format dial command }
    Command := Format('ATD%s%s', [Profile.DialMode, Number]);
    { Dial number }
    if Modem.Command(Command, '') <> OK then
    begin
      Result := FAIL;
      Exit;
    end;
    TimeOut := Profile.DialTimeOut * 1000;
    StartTime := GetTickCount;
    DialComplete := False;
    Result := FAIL;
    { Wait modem answer }
    while (not FStopDialing) and (not DialComplete) do
    begin
      { Check for modem char }
      if Modem.Port.CharReady then
      begin
        { Read modem answer string }
        if Modem.ReadStr(Answer) <> OK then
        begin
          Result := FAIL;
          Exit;
        end;
        { Convert answer string }
        Result := Modem.Analyze(Answer, Command);
        { Analyze answer }
        case Result of
          VMA_BUSY:
            begin
              DialComplete := True;
              Result := OK;
              Modem.QueueEvent(TModemEvent.Create(BUSY_TONE));
            end;
          VMA_EMPTY:
            begin
            end;
          VMA_NO_ANSWER:
            begin
              DialComplete := True;
              Result := OK;
              Modem.QueueEvent(TModemEvent.Create(NO_ANSWER));
            end;
          VMA_NO_DIAL_TONE:
            begin
              DialComplete := True;
              Result := OK;
              Modem.QueueEvent(TModemEvent.Create(NO_DIAL_TONE));
            end;
          VMA_RINGING, VMA_USER_1:
            begin
            end;
          VMA_OK, VMA_VCON:
            begin
              DialComplete := True;
              Result := OK;
            end;
          else
            begin
              Modem.LastError := VME_INVALID_MODEM_ANSWER;
              DialComplete := True;
              Result := FAIL;
            end;
        end;
      end
      else
        { Wait for data }
        SleepEx(Profile.PortReadDelay, False);
      { Yield callback }
      Yield;
      { Check for new events }
      Modem.CheckEvents;
      { Check timeout }
      if (TimeOut > 0) and (DWORD(Abs(GetTickCount - StartTime)) >= TimeOut) and (not FStopDialing) then
      begin
        Modem.QueueEvent(TModemEvent.Create(OPERATION_TIMEOUT));
        Modem.StopDialing;
      end;
    end;
    { Check for dialing interrupted }
    if FStopDialing then Result := INTERRUPTED;
  finally
     Modem.State := IDLE;
  end;
  { Process events queue }
  Modem.CheckEvents;
end;

class function TIS101.GetVoiceLen(Compression, Speed, Bits: Integer; DataSize: Longint): Integer;
var
  BPS: Integer;
begin
  BPS := (Speed * Bits) div 8;
  { Check BPS and return duration of voice message }
  if BPS > 0 then
  begin
    Result := Round(DataSize / BPS * 1000);
    Exit;
  end;
  { Error }
  Result := 0;
end;

function TIS101.HandleDLE(Code: Char): Integer;
begin
  { Check modem specified }
  if not ModemAssigned('HandleDLE', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Handle DLE code }
  case Code of
    DLE:
      { Shielded <DLE> code }
      begin
        DebugMsg(L_WARN, SShieldedDLEReceived);
        Result := OK;
        Exit;
      end;
    SUB:
      { Shielded <DLE> <DLE> code }
      begin
        DebugMsg(L_WARN, SShieldedDLEDLEReceived);
        Result := OK;
        Exit;
      end;
    ETX:
      { <ETX> code }
      begin
        DebugMsg(L_WARN, SDLEETXReceived);
        Result := OK;
        Exit;
      end;
    '$':
      { Bong tone detected }
      begin
        Result := Modem.QueueEvent(TModemEvent.Create(BONG_TONE));
        Exit;
      end;
    '/':
      { Start of DTMF shielding }
      begin
        DebugMsg(L_WARN, SStartDTMFShielding);
        Result := OK;
        Exit;
      end;
    '~':
      { End of DTMF shielding }
      begin
        DebugMsg(L_WARN, SEndDTMFShielding);
        Result := OK;
        Exit;
      end;
    '0',
    '1',
    '2',
    '3',
    '4',
    '5',
    '6',
    '7',
    '8',
    '9',
    '*',
    '#',
    'A',
    'B',
    'C',
    'D':
      { DTMF tone detected }
      begin
        Result := Modem.QueueEvent(TModemEvent.CreateDataC(RECEIVED_DTMF, Code));
        Exit;
      end;
    'a':
      { Data or fax answer detected (V.25/T.30 2100 Hz) }
      begin
        Result := Modem.QueueEvent(TModemEvent.Create(DATA_OR_FAX_DETECTED));
        Exit;
      end;
    'b', 'K':
      { Busy tone detected }
      begin
        Result := Modem.QueueEvent(TModemEvent.Create(BUSY_TONE));
        Exit;
      end;
    'c', 'm':
      { Fax calling tone detected (T.30 1100 Hz) }
      begin
        Result := Modem.QueueEvent(TModemEvent.Create(FAX_CALLING_TONE));
        Exit;
      end;
    'd', 'i':
      { Dial tone detected }
      begin
        Result := Modem.QueueEvent(TModemEvent.Create(DIAL_TONE));
        Exit;                                     
      end;
    'e':
      { European Data Modem Calling Tone (V.25 1300 Hz) }
      begin
        Result := Modem.QueueEvent(TModemEvent.Create(DATA_CALLING_TONE));
        Exit;
      end;
    'f':
      { Bell Answer Tone (2225 Hz) }
      begin
        Result := Modem.QueueEvent(TModemEvent.Create(DATA_CALLING_TONE));
        Exit;
      end;
    'E':
      { Invalid voice format detected }
      begin
        DebugMsg(L_WARN, SInvalidVoiceFormat);
        Result := OK;
        Exit;
      end;
    'h', 'p':
      { Local handset goes on hook }
      begin
        Result := Modem.QueueEvent(TModemEvent.Create(HANDSET_ON_HOOK));
        Exit;
      end;
    'H', 'P':
      { Local handset goes off hook }
      begin
        Result := Modem.QueueEvent(TModemEvent.Create(HANDSET_OFF_HOOK));
        Exit;
      end;
    'J':
      { SIT tone detected }
      begin
        Result := Modem.QueueEvent(TModemEvent.Create(SIT_TONE));
        Exit;
      end;
    'l':
      { Loop current break }
      begin
        Result := Modem.QueueEvent(TModemEvent.Create(LOOP_BREAK));
        Exit;
      end;
    'L':
      { Loop current polarity reversal }
      begin
        Result := Modem.QueueEvent(TModemEvent.Create(LOOP_POLARITY_CHANGE));
        Exit;
      end;
    'o':
      { Buffer overrun (oveflow) }
      begin
        DebugMsg(L_WARN, 'Buffer overrun');
        Result := OK;
        Exit;
      end;
    'q':
      { Modem detected silence }
      begin
        Result := Modem.QueueEvent(TmodemEvent.Create(SILENCE_DETECTED));
        Exit;
      end;
    'Q':
      { XON received }
      begin
        DebugMsg(L_WARN, SXONReceived);
        Result := OK;
        Exit;
      end;
    'r':
      { Ringback detected }
      begin
        Result := Modem.QueueEvent(TModemEvent.Create(RINGBACK_DETECTED));
        Exit;
      end;
    'R':
      { Incoming ring detected }
      begin
        Result := Modem.QueueEvent(TModemEvent.Create(RING_DETECTED));
        Exit;
      end;
    's':
      { Modem could not detect voice energy on the line }
      begin
        Result := Modem.QueueEvent(TModemEvent.Create(NO_VOICE_ENERGY));
        Exit;
      end;
    'S':
      { XOFF received }
      begin
        DebugMsg(L_WARN, SXOFFReceived);
        Result := OK;
        Exit;
      end;
    't':
      { TDD detected }
      begin
        Result := Modem.QueueEvent(TModemEvent.Create(TDD_DETECTED));
        Exit;
      end;
    'T':
      { Timing mark will be ignored }
      begin
        Result := OK;
        Exit;
      end;
    'u':
      { Buffer underrun (becomes empty) }
      begin
        DebugMsg(L_WARN, SBufferUnderrun);
        Result := OK;
        Exit;
      end;
    'v',
    'V':
      { Voice detected }
      begin
        Result := Modem.QueueEvent(TModemEvent.Create(VOICE_DETECTED));
        Exit;
      end;
    'w':
      { Call waiting, beep interrupt }
      begin
        Result := Modem.QueueEvent(TModemEvent.Create(CALL_WAITING));
        Exit;
      end;
    'Y':
      { Lost data detected event }
      begin
        DebugMsg(L_WARN, SLostData);
        Result := OK;
        Exit;
      end;
  end;
  { Unknown DLE code }
  DebugMsgFmt(L_WARN, SUnknownDLECode, [Code]);
  Result := FAIL;
end;

function TIS101.HandleEvent(Event: Integer; Data: TEventData): Integer;
begin
  Result := UNKNOWN_EVENT;
end;

function TIS101.HangUp: Integer;
begin
  { Check modem specified }
  if not ModemAssigned('HangUp', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Hang up }
  if Modem.Command(HangUpCmd, HangUpAnsw) <> VMA_USER_1 then
  begin
    Result := FAIL;
    Exit;
  end;
  Result := OK;
end;

class function TIS101.ModemName: string;
begin
  Result := 'IS-101 compatible modem';
end;

class function TIS101.ModemRMDName: string;
begin
  Result := 'IS-101';
end;

function TIS101.PlayDTMF(const Number: string): Integer;
var
  I, L: Integer;
  Command: string;
begin
  { Check modem specified }
  if not ModemAssigned('PlayDTMF', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  L := Length(Number);
  { Check number }
  if L = 0 then
  begin
    Result := FAIL;
    Exit;
  end;
  { Format command }
  try
    Command := Format(PlayDTMFCmd, [Number[1]]);
    for I := 2 to L do Command := Format(Command + PlayDTMFExtra, [Number[I]]);
  except
    DebugMsg(L_ERROR, SInvalidDTMFCommandFormat);
    Modem.LastError := VME_INVALID_COMMAND_FORMAT;
    Result := FAIL;
    Exit;
  end;
  { Send command }
  if Modem.Command(Command, '') <> OK then
  begin
    Result := FAIL;
    Exit;
  end;
  { Delay }
  SleepEx(L * 500, False);
  { Read modem answer }
  if (Modem.Command('', PlayDTMFAnsw) and VMA_USER_1) <> VMA_USER_1 then
  begin
    Result := FAIL;
    Exit;
  end;
  { Success }
  Result := OK;
end;

function TIS101.PlayStream(Stream: TStream; BPS: Integer): Integer;
var
  OutputBuffer: array[0..VOICE_BUF_LEN + 1] of Char;
  BytesMax, BytesOut: Integer;
  DataByte, ModemByte: Char;
  PlayComplete: Boolean;
  TimeOut, StartTime: DWORD;
begin
  { Check modem specified }
  if not ModemAssigned('PlayStream', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Update modem state }
  Modem.State := PLAYING;
  try
    BytesMax := (BPS div 8) div 10;
    if BytesMax > VOICE_BUF_LEN then BytesMax := VOICE_BUF_LEN;
    PlayComplete := False;
    if Profile.PlayPreview then
      TimeOut := Profile.PlayPreviewLen * 1000
    else
      TimeOut := Profile.PlayMaxLen * 1000;
    StartTime := GetTickCount;
    { <START PLAYING> }
    DebugMsg(L_JUNK, SStartPlaying);
    { Playing voice data }
    while (not FStopPlaying) and (not PlayComplete) do
    begin
      { Read voice data from stream and prepare to output }
      BytesOut := 0;
      while BytesOut < BytesMax do
      begin
        { Red byte from stream }
        if Stream.Read(DataByte, 1) = 0 then
        begin
          PlayComplete := True;
          Break;
        end;
        { Add byte to output buffer }
        OutputBuffer[BytesOut] := DataByte;
        Inc(BytesOut);
        { Check for shielded <DLE> codes }
        if DataByte = DLE then
        begin
          OutputBuffer[BytesOut] := DLE;
          Inc(BytesOut);
        end;
        { Yield callback }
        Yield;
      end;
      { Send buffer to modem }
      if BytesOut > 0 then
      begin
        DebugMsgFmt(L_JUNK, SVoiceData, [BytesOut]);
        { Send data }
        if Modem.Write(@OutputBuffer, BytesOut) = FAIL then
        begin
          Result := FAIL;
          Exit;
        end;
        { Progress callback }
        Progress(Round(Stream.Position / Stream.Size * 100));
      end;
      { Check for modem char }
      while Modem.Port.CharReady do
      begin
        if Modem.ReadChar(ModemByte) = FAIL then
        begin
          Result := FAIL;
          Exit;
        end;
        { Check for <DLE> codes }
        if ModemByte = DLE then
        begin
          if Modem.ReadChar(ModemByte) = FAIL then
          begin
            Result := FAIL;
            Exit;
          end;
          DebugMsgFmt(L_JUNK, SReadDLE, [ModemByte]);
          Modem.HandleDLE(ModemByte);
        end
        else
          DebugMsgFmt(L_WARN, SUnexpectedByte, [ModemByte]);
      end;
      { Check for new events }
      Modem.CheckEvents;
      { Check timeout }
      if ((TimeOut > 0) and (DWORD(Abs(GetTickCount - StartTime)) >= TimeOut)) and (not FStopPlaying) then
      begin
        Modem.QueueEvent(TModemEvent.Create(OPERATION_TIMEOUT));
        Modem.StopPlaying;
      end;
    end;
  finally
    Modem.State := IDLE;
  end;
  { Check for playing stopped }
  if FStopPlaying then
  begin
    Result := INTERRUPTED;
    Exit;
  end;
  { Process events queue }
  Modem.CheckEvents;
  { Playing complete }
  Result := OK;
end;

function TIS101.RecordStream(Stream: TStream; BPS: Integer): Integer;
var
  InputBuffer: array[0..VOICE_BUF_LEN + 1] of Char;
  InputByte: Char;
  BytesMax, BytesIn: Integer;
  GotDLEETX, WasDLE: Boolean;
  TimeOut, StartTime: DWORD;
begin
  { Check modem specified }
  if not ModemAssigned('RecordStream', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Update modem state }
  Modem.State := RECORDING;
  try
    GotDLEETX := False;
    WasDLE := False;
    BytesMax := (BPS div 8) div 10;
    BytesIn := 0;
    TimeOut := Profile.RecMaxLen * 1000;
    StartTime := GetTickCount;
    { Recording voice data }
    DebugMsg(L_JUNK, SStartRecording);
    while not GotDLEETX do
    begin
      { Check timeout }
      if (TimeOut > 0) and (DWORD(Abs(GetTickCount - StartTime)) >= TimeOut) and (not FStopRecording) then
      begin
        Modem.QueueEvent(TModemEvent.Create(OPERATION_TIMEOUT));
        Modem.StopRecording;
      end;
      { Receive data (check error if modem has no data during timeout) }
      if Modem.Read(@InputByte, 1) <> 1 then
      begin
        Result := FAIL;
        Exit;
      end;
      { Check about DLE code }
      if WasDLE then
      begin
        WasDLE := False;
        { Analyse <DLE> char }
        case InputByte of
          DLE:
            { Shielded DLE code }
            begin
              InputBuffer[BytesIn] := DLE;
              Inc(BytesIn);
            end;
          ETX:
            { End of Stream }
            begin
              { Flush buffer }
              Stream.Write(InputBuffer, BytesIn);
              { Record complete }
              GotDLEETX := True;
              { Output debug message }
              DebugMsgFmt(L_JUNK, SVoiceData, [BytesIn]);
              DebugMsg(L_JUNK, SDLEETX);
              { Progress callback }
              Progress(Stream.Size);
            end;
          SUB:
            { ? }
            begin
              InputBuffer[BytesIn] := DLE;
              Inc(BytesIn);
              InputBuffer[BytesIn] := DLE;
              Inc(BytesIn);
            end;
          else
            { DLE code }
            begin
              DebugMsgFmt(L_JUNK, SReadDLE, [InputByte]);
              Modem.HandleDLE(InputByte);
            end;
        end;
      end
      else
        { Check input byte }
        if InputByte = DLE then
          { DLE byte received }
          WasDLE := True
        else
        begin
          { Data byte received }
          InputBuffer[BytesIn] := InputByte;
          Inc(BytesIn);
        end;
      { Write buffer to stream }
      if (not GotDLEETX) and (BytesIn >= BytesMax) then
      begin
        Stream.Write(InputBuffer, BytesIn);
        DebugMsgFmt(L_JUNK, SVoiceData, [BytesIn]);
        BytesIn := 0;
        { Progress callback }
        Progress(Stream.Size);
      end;
      { Yield callback }
      Yield;
      { Check for new events }
      Modem.CheckEvents;
    end;
  finally
    Modem.State := IDLE;
  end;
  { Process events queue }
  Modem.CheckEvents;
  { Recording complete (always interrupted) }
  Result := INTERRUPTED;
end;

function TIS101.RestoreVoiceMode: Integer;
begin
  Result := VoiceModeOn;
end;

function TIS101.SetCompression(var Compression, Speed, Bits: Integer): Integer;
begin
  { Use default compression method }
  Result := OK;
end;

function TIS101.SetDevice(Device: Integer): Integer;
begin
  { Use default device }
  Result := OK;
end;

function TIS101.StartPlayStream: Integer;
begin
  FStopPlaying := False;
  { Check modem specified }
  if not ModemAssigned('StartPlayStream', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Set hardware flow control }
  if (Modem.Command(HardFlowCmd, HardFlowAnsw) and VMA_USER) <> VMA_USER then
  begin
    Result := FAIL;
    Exit;
  end;
  { Start playing }
  if (Modem.Command(StartPlayCmd, StartPlayAnsw) and VMA_USER) <> VMA_USER then
  begin
    Result := FAIL;
    Exit;
  end;
  Result := OK;
end;

function TIS101.StartRecordStream: Integer;
begin
  FStopRecording := False;
  { Check modem specified }
  if not ModemAssigned('StartRecordStream', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Set hardware flow control }
  if (Modem.Command(HardFlowCmd, HardFlowAnsw) and VMA_USER) <> VMA_USER then
  begin
    Result := FAIL;
    Exit;
  end;
  { Tells the modem ready to receive }
  if Modem.Command(StartRecCmd, StartRecAnsw) <> VMA_USER_1 then
  begin
    Result := FAIL;
    Exit;
  end;
  Result := OK;
end;

function TIS101.StopDialing: Integer;
begin
  FStopDialing := True;
  Result := OK;
end;

function TIS101.StopPlayStream: Integer;
const
  NULByte: Char = NUL;
begin
  { Check modem specified }
  if not ModemAssigned('StopPlayStream', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Check for play already interrupted }
  if FStopPlaying then
  begin
    { Flush output buffer }
    Modem.Port.FlushOutBuffer;
    { Write null char }
    if Modem.Write(@NULByte, 1) <> OK then
    begin
      Result := FAIL;
      Exit;
    end;
    { Write interrupt recording command }
    if Modem.Write(PChar(IntrPlayCmd), Length(IntrPlayCmd)) <> OK then
    begin
      Result := FAIL;
      Exit;
    end;
    DebugMsg(L_JUNK, SInterruptPlaying);
    { Wait modem answer it is done }
    if (Modem.Command('', IntrPlayAnsw) and VMA_USER) <> VMA_USER then
    begin
      Result := FAIL;
      Exit;
    end;
  end
  else
  begin
    { Write stop recording command }
    if Modem.Write(PChar(StopPlayCmd), Length(StopPlayCmd)) <> OK then
    begin
      Result := FAIL;
      Exit;
    end;
    DebugMsg(L_JUNK, SStopPlaying);
    { Wait modem answer it is done }
    if (Modem.Command('', StopPlayAnsw) and VMA_USER) <> VMA_USER then
    begin
      Result := FAIL;
      Exit;
    end;
  end;
  { Playing complete }
  Result := OK;
end;

function TIS101.StopPlaying: Integer;
begin
  FStopPlaying := True;
  Result := OK;
end;

function TIS101.StopRecordStream: Integer;
begin
  { Check modem specified }
  if not ModemAssigned('StopRecordStream', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Wait modem answer it is done }
  if (Modem.Command('', StopRecAnsw) and VMA_USER) <> VMA_USER then
  begin
    Result := FAIL;
    Exit;
  end;
  { Recording complete }
  Result := OK;
end;

function TIS101.StopRecording: Integer;
begin
  FStopRecording := True;
  { Check modem specified }
  if not ModemAssigned('StopRecording', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Write stop recording sequence }
  if Modem.Write(PChar(StopRecCmd), Length(StopRecCmd)) <> OK then
  begin
    Result := FAIL;
    Exit;
  end;
  DebugMsg(L_JUNK, SStopRecording);
  { Success }
  Result := OK;
end;

function TIS101.StopWaiting: Integer;
begin
  FStopWaiting := True;
  Result := OK;
end;

function TIS101.SwitchToDataFax(const Mode: string): Integer;
begin
  { Check modem specified }
  if not ModemAssigned('SwitchToDataFax', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Write switch command }
  if (Modem.CommandFmt(SwitchModeCmd, SwitchModeAnsw, [Mode]) and VMA_USER) <> VMA_USER then
  begin
    Result := FAIL;
    Exit;
  end;
  { Check modem }
  if Modem.Command('AT', 'OK') <> VMA_USER_1 then
  begin
    Result := FAIL;
    Exit;
  end;
  Result := OK;
end;

function TIS101.VoiceModeOff: Integer;
begin
  { Check modem specified }
  if not ModemAssigned('VoiceModeOff', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Write voice mode off command }
  if (Modem.Command(VoiceModeOffCmd, VoiceModeOffAnsw) and VMA_USER) <> VMA_USER then
  begin
    Result := FAIL;
    Exit;
  end;
  Result := OK;
end;

function TIS101.VoiceModeOn: Integer;
begin
  { Check modem specified }
  if not ModemAssigned('VoiceModeOn', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Write voice mode on command }
  if (Modem.Command(VoiceModeOnCmd, VoiceModeOnAnsw) and VMA_USER) <> VMA_USER then
  begin
    Result := FAIL;
    Exit;
  end;
  Result := OK;
end;

function TIS101.Wait(TimeOut: DWORD): Integer;
var
  CharRead: Char;
  StartTime: DWORD;
begin
  { Check modem specified }
  if not ModemAssigned('Wait', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Update modem state }
  Modem.State := WAITING;
  try
    FStopWaiting := False;
    StartTime := GetTickCount;
    { Wait modem answer }
    while not FStopWaiting do
    begin
      { Process modem data }
      while Modem.Port.CharReady do
      begin
        { Read one byte }
        if Modem.ReadChar(CharRead) = FAIL then
        begin
          Result := FAIL;
          Exit;
        end;
        { Check for <DLE> codes }
        if CharRead = DLE then
        begin
          { Read DLE code }
          if Modem.ReadChar(CharRead) = FAIL then
          begin
            Result := FAIL;
            Exit;
          end;
          DebugMsgFmt(L_JUNK, SReadDLE, [CharRead]);
          { Handle DLE }
          Modem.HandleDLE(CharRead);
        end
        else
          DebugMsgFmt(L_WARN, SUnexpectedByte, [CharRead]);
      end;
      { Yield callback }
      Yield;
      { Check for new events }
      Modem.CheckEvents;
      { Check timeout }
      if (TimeOut > 0) and (DWORD(Abs(GetTickCount - StartTime)) >= TimeOut) and (not FStopWaiting) then
      begin
        Modem.QueueEvent(TModemEvent.Create(OPERATION_TIMEOUT));
        Modem.StopWaiting;
      end;
      { Wait for data }
      SleepEx(Profile.PortReadDelay, False);
    end;
  finally
     Modem.State := IDLE;
  end;
  { Process events queue }
  Modem.CheckEvents;
  { Check for waiting interrupted }
  if FStopWaiting then
  begin
    Result := INTERRUPTED;
    Exit;
  end;
  { Waiting success }
  Result := OK;
end;

function TIS101.WaitAnswer(const ExpectedAnswers: string; TimeOut: DWORD): Integer;
var
  CharRead: Char;
  Buffer: string;
  FirstChar: Boolean;
  StartTime: DWORD;
begin
  { Check modem specified }
  if not ModemAssigned('WaitAnswer', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Update modem state }
  Modem.State := WAITING;
  try
    FStopWaiting := False;
    Buffer := '';
    FirstChar := True;
    StartTime := GetTickCount;
    { Wait modem answer }
    while not FStopWaiting do
    begin
      { Check for modem char }
      if Modem.Port.CharReady then
      begin
        { Read one byte }
        if Modem.ReadChar(CharRead) = FAIL then
        begin
          Result := FAIL;
          Exit;
        end;
        { Check for <DLE> codes }
        if CharRead = DLE then
        begin
          { Read DLE code }
          if Modem.ReadChar(CharRead) = FAIL then
          begin
            Result := FAIL;
            Exit;
          end;
          DebugMsgFmt(L_JUNK, SReadDLE, [CharRead]);
          { Handle DLE }
          Modem.HandleDLE(CharRead);
        end
        { Check for valid chars }
        else if not (CharRead in [LF, CR, XON, XOFF]) then
        begin
          if FirstChar then
          begin
            Buffer := CharRead;
            FirstChar := False;
          end
          else
            Buffer := Buffer + CharRead
        end
        { Check for unexpected XON/XOFF codes }
        else if not (CharRead in [LF, CR]) then
          DebugMsgFmt(L_WARN, SUnexpectedByte, [CharRead])
        { Check for end of line }
        else if CharRead = LF then
        begin
          DebugMsgFmt(L_JUNK, SWaitAnswer, [Buffer]);
          { Analyze modem answer }
          if (not FirstChar) and (Length(Buffer) > 0) then
          begin
            Result := Modem.Analyze(Buffer, ExpectedAnswers);
            if Result <> VMA_FAIL then Exit;
          end;
          Buffer := '';
          FirstChar := True;
        end;
      end
      else
        { Wait for data }
        SleepEx(Profile.PortReadDelay, False);
      { Yield callback }
      Yield;
      { Check for new events }
      Modem.CheckEvents;
      { Check timeout }
      if (TimeOut > 0) and (DWORD(Abs(GetTickCount - StartTime)) >= TimeOut) and (not FStopWaiting) then
      begin
        Modem.QueueEvent(TModemEvent.Create(OPERATION_TIMEOUT));
        Modem.StopWaiting;
      end;
    end;
  finally
     Modem.State := IDLE;
  end;
  { Process events queue }
  Modem.CheckEvents;
  { Check for waiting interrupted }
  if FStopWaiting then
  begin
    Result := INTERRUPTED;
    Exit;
  end;
  { Waiting success }
  Result := OK;
end;

end.
