{
  Voice Modem Library v1.0

  This file contain definition of the simple log file.

  Roman M. Mochalov
  E-mail: roman@sar.nnov.ru
}

unit VML_Log;

interface

uses
  Windows, SysUtils, Classes;

type

{ TModemLogFile }

  {
    Simple log file. Open exists log (text) file or creates new file, move to
    the file end and write log title. Then ready to add log lines.

    Procedures:

    Add            - Add line to log file.

    Properties:

    AddDateTime    - Add date/time string to the each log line except log
                     title line.
    CreateNew      - Create new log file, erase old file.
    FileName       - Log file name.
    AddDateTime    - Add date and time to the each log line.
    DateTimeFormat - Date-and-time value format string (see FormatDateTime
                     function help for more details).
    Title          - Log title.

    Events:

    OnGetTitle     - Get log title. This event calls every time right after
                     log file opening.
  }

  TGetLogTitleEvent = procedure(Sender: TObject; var Value: string) of object;

  TModemLogFile = class(TComponent)
  private
    FFileName: string;
    FTitle: string;
    FAddDateTime: Boolean;
    FDateTimeFormat: string;
    FLogFile: TFileStream;
    FCreateNew: Boolean;
    FOnGetTitle: TGetLogTitleEvent;
    function IsDateTimeformatStored: Boolean;
    procedure SetFileName(Value: string);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function Add(const S: string): Boolean;
    function Open: Boolean;
    function Close: Boolean;
  published
    property AddDateTime: Boolean read FAddDateTime write FAddDateTime default False;
    property CretateNew: Boolean read FCreateNew write FCreateNew default False;
    property DateTimeFormat: string read FDateTimeFormat write FDateTimeFormat stored IsDateTimeformatStored;
    property FileName: string read FFileName write SetFileName;
    property Title: string read FTitle write FTitle;
    property OnGetTitle: TGetLogTitleEvent read FOnGetTitle write FOnGetTitle;
  end;

implementation

{ TModemLogFile }

constructor TModemLogFile.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FDateTimeFormat := 'MM-DD-YYYY hh:mm:ss';
end;

destructor TModemLogFile.Destroy;
begin
  Close;
  inherited Destroy;
end;

function TModemLogFile.IsDateTimeformatStored: Boolean;
begin
  Result := FAddDateTime;
end;

procedure TModemLogFile.SetFileName(Value: string);
begin
  if FFileName <> Value then
  begin
    FFileName := Value;
    { Close log file }
    FLogFile.Free;
    FLogFile := nil;
  end;
end;

function TModemLogFile.Add(const S: string): Boolean;
var
  L: string;
begin
  try
    { Open log file }
    if not Open then
    begin
      Result := False;
      Exit;
    end;
    { Create log line }
    L := '';
    if FAddDateTime then L := FormatDateTime(FDateTimeFormat, Now) + ' - ';
    L := L + S;
    { Write line }
    FLogFile.Write(Pointer(L)^, Length(L));
    FLogFile.Write(#13#10, 2);
    { Success }
    Result := True;
  except
    Result := False;
  end;
end;

function TModemLogFile.Open: Boolean;
var
  S: string;
begin
  { Check log already opened }
  if FLogFile <> nil then
  begin
    Result := True;
    Exit;
  end;
  { Open file }
  try
    if not FCreateNew then
    begin
      { Check for new file }
      if not FileExists(FFileName) then
      begin
        FLogFile := TFileStream.Create(FFileName, fmCreate);
        FLogFile.Free;
      end;
      { Open log file, allow other users read log }
      FLogFile := TFileStream.Create(FFileName, fmOpenWrite or fmShareDenyWrite);
      { Move to end of file end write empty line }
      if FLogFile.Size > 0 then
      begin
        FLogFile.Position := FLogFile.Size;
        FLogFile.Write(#13#10, 2);
      end;
    end
    else
    begin
      { Create new file }
      FLogFile := TFileStream.Create(FFileName, fmCreate);
      FLogFile.Free;
      { Open log file, allow other users read log }
      FLogFile := TFileStream.Create(FFileName, fmOpenWrite or fmShareDenyWrite);
    end;
    { Get log title }
    S := FTitle;
    if Assigned(FOnGetTitle) then FOnGetTitle(Self, S);
    { Write title string }
    if Length(S) > 0 then
    begin
      FLogFile.Write(Pointer(S)^, Length(S));
      FLogFile.Write(#13#10, 2);
      FLogFile.Write(#13#10, 2);
    end;
    { Success }
    Result := True;
  except
    Result := False;
  end;
end;

function TModemLogFile.Close: Boolean;
begin
  try
    { Close file }
    FLogFile.Free;
    FLogFile := nil;
    { Success }
    Result := True;
  except
    Result := False;
  end;
end;

end.

