{
  Voice Modem Library v1.0

  This file contain definition of abstract hadrware driver
  class and voice modem class.

  Roman M. Mochalov
  E-mail: roman@sar.nnov.ru
}

unit VML_Modem;

interface

uses
  Windows, Messages, SysUtils, Classes, VML_Consts, VML_Profile,
  VML_Port, VML_Log, VML_Idents, Dialogs;

type

{ Preefine classes }

  TVoiceModem = class;
  TModemDriver = class;
  TModemEvent = class;

{ Modeme event data structures }

  TBeepData = packed record
    Frequency: Integer;
    Length: Integer;
  end;

  TEventData = packed record
    case Integer of
      0: (C: Char);
      1: (I: Integer);
      2: (P: Pointer);
      3: (Beep: TBeepData);
  end;

{ Event queue structures }

  TQueueItem = packed record
    WriteLock: Integer;
    ReadLock: Integer;
    Event: TModemEvent;
  end;

  TEventQueue = array[0..MAX_EVENTS-1] of TQueueItem;

{ Modem event class }

  TModemEvent = class
  private
    FEvent: Integer;
    FData: TEventData;
  public
    constructor Create(AEvent: Integer);
    constructor CreateDataC(AEvent: Integer; C: Char);
    constructor CreateDataI(AEvent: Integer; I: Integer);
    constructor CreateDataP(AEvent: Integer; P: Pointer);
    property Event: Integer read FEvent write FEvent;
    property Data: TEventData read FData write FData;
  end;

{ TModemDriver }

  {
    Hardware driver abstract class.

    Procedures:

    AnswerPhone         - Answering in voice.
    Beep                - Play one or more DTMF or other tones in voice mode.
    Dial                - Dial number and wait answer.
    GetCompression      - Get compression ID for given name. Tries to convert
                          given string into number if compression with given
                          method name not supported.
    GetCompressionMap   - Get compression idents map. Override this method
                          to specify list of allowed compression methods.
    GetCompressionName  - Get compression name for given ID. Return string
                          with given number if given compression method not
                          supported.
    GetCompressionsList - Get supported compression name list. Return number
                          of supported methods.
    GetDevice           - Get device ID for given name. Tries to convert
                          given string into number if device with given name
                          not supported.
    GetDeviceMap        - Get device idents map. Override this method to
                          specify list of allowed devices.
    GetDeviceName       - Get device name for given ID. Return string with
                          given number if given device not supported.
    GetDevicesList      - Get supported device name list. Return number of
                          supported devices.
    GetPlayDevicesList  - Get subset of supported devices for playing voice
                          data.
    GetRecDevicesList   - Get subset of supported devices for recording
                          voice data.
    GetVoiceLen         - Calculate duration of voice for given modem data.
    HandleDLE           - Handle DLE code.
    HandleEvent         - Handle given event.
    HangUp              - Hang up (goes on hook).
    Init                - Hardware specific modem init. WARNING: do nothing
                          to change local echo and result codes settings.
    InitVoice           - Init modem voice params (called in voice mode).
    ModemAssigned       - Check modem property. Generate debug error message
                          if property is not assigned.
    ModemName           - Modem name.
    ModemRMDName        - Modem short name for RMD file header.
    PlayDTMF            - Play one of more DTMF tones.
    PlayStream          - Play data from stream.
    RecordStream        - Record data to stream.
    RestoreVoiceMode    - Switch modem to voice mode, reset voice
                          preferences, which are forgotten after leaving
                          voice mode.
    SetCompression      - Set selected compression. Return valid bits per
                          sample value.
    SetDevice           - Set selected device.
    StartPlayStream     - Write to modem start playing command.
    StartRecordStream   - Write to modem start recording command.
    StopDialing         - Stop dialing operation.
    StopPlayStream      - Write to modem stop play voice command.
    StopPlaying         - Stop playing voice data.
    StopRecordStream    - Write to modem stop voice recording sequence.
    StopRecording       - Stop recording voice data.
    StopWaiting         - Stop waintg operation.
    SwitchToDataFax     - Switch modem to selected mode (data, fax or voice).
    VoiceModeOff        - Switch modem to voice mode.
    VoiceModeOn         - Return to previous mode
    Wait                - Wait and handle modem <DLE> sequence during given
                          timeout in msec.
    WaitAnswer          - Wait modem standard or given answer, detect <DLE>
                          sequences during given timeout in msec.

    Properties:

    PickPhoneCmd        - Manual answer: goes off hook in answer mode and
                          answer.
    PickPhoneAnsw
    HangUpCmd           - Hangs up (goes on hook) command and answer.
    HangUpAnsw                  Format
    BeepCmd             - Play tone command and answer.
    BeepAnsw
    BeepTimeUnit        - Divisor for calculation of value of beep duration.
                          By default, duration of beep signal (Length
                          parameter in Beep() function) is measured at
                          milliseconds. But some modems required beep
                          duration parameter in other units (for example in
                          0.01 second). The true length of beep signal will
                          be calculated as Length / BeepTimeUnit.
    HardFlowCmd         - Turn on handware flow control command and answer
    HardFlowAnsw
    SoftFlowCmd         - Turn on software flow control command and answer
    SoftFlowAnsw
    StartPlayCmd        - Start, stop or interrupt playing or recording
    StartPlayAnsw         voice data commands and answers.
    IntrPlayCmd
    IntrPlayAnsw
    StopPlayCmd
    StopPlayAnsw
    StartRecCmd
    StartRecAnsw
    StopRecCmd
    StopRecAnsw
    SwitchModeCmd       - Select data or fax mode command and answer.
    SwitchModeAnsw
    VoiceModeOnCmd      - Entering voice mode command and answer.
    VoiceModeOnAnsw
    VoiceModeOffCmd     - Leaving voice mode command and answer.
    VoiceModeOffAnsw
    PlayDTMFCmd         - Play DTMF, separate DTMF commands and answer.
    PlayDTMFExtra
    PlayDTMFAnsw
    Modem               - Modem component link. For properly working a driver
                          this property must be always assigned.
    Profile             - Modem profile. If property not specified at design
                          time it will be taken from modem at runtime.

    Events:

    OnDebugMsg          - Debug message event.
    OnProgress          - Progress callback for voice playing and recording.
    OnYield             - Yield event. Use this event for process Windows
                          messages.
  }

  TModemDriverClass = class of TModemDriver;

  TModemDriver = class(TComponent)
  private
    FPickPhoneCmd: string;
    FPickPhoneAnsw: string;
    FHangUpCmd: string;
    FHangUpAnsw: string;
    FBeepCmd: string;
    FBeepAnsw: string;
    FBeepTimeUnit: Integer;
    FHardFlowCmd: string;
    FHardFlowAnsw: string;
    FSoftFlowCmd: string;
    FSoftFlowAnsw: string;
    FStartPlayCmd: string;
    FStartPlayAnsw: string;
    FIntrPlayCmd: string;
    FIntrPlayAnsw: string;
    FStopPlayCmd: string;
    FStopPlayAnsw: string;
    FStartRecCmd: string;
    FStartRecAnsw: string;
    FStopRecCmd: string;
    FStopRecAnsw: string;
    FSwitchModeCmd: string;
    FSwitchModeAnsw: string;
    FVoiceModeOnCmd: string;
    FVoiceModeOnAnsw: string;
    FVoiceModeOffCmd: string;
    FVoiceModeOffAnsw: string;
    FPlayDTMFCmd: string;
    FPlayDTMFExtra: string;
    FPlayDTMFAnsw: string;
    FModem: TVoiceModem;
    FProfile: TModemProfile;
    function GetProfile: TModemProfile;
    procedure SetModem(Value: TVoiceModem);
  protected
    procedure DebugMsg(Msg: Integer; const S: string); virtual;
    procedure DebugMsgFmt(Msg: Integer; const S: string; const Args: array of const);
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure Progress(Done: Longint); virtual;
    procedure Yield; virtual;
  public
    function AnswerPhone: Integer; virtual; abstract;
    function Beep(Frequency, Length: Integer): Integer; virtual; abstract;
    function Dial(const Number: string): Integer; virtual; abstract;
    class function GetCompression(const Name: string): Integer; virtual; 
    class function GetCompressionMap(var Map: PIdentMap): Integer; virtual;
    class function GetCompressionName(Compression: Integer): string; virtual;
    class function GetCompressionsList(Proc: TGetStrProc): Integer; virtual;
    class function GetCompressionsStringList(Strings: TStrings): Integer; virtual;
    class function GetDevice(const Name: string): Integer; virtual; 
    class function GetDeviceMap(var Map: PIdentMap): Integer; virtual;
    class function GetDeviceName(Device: Integer): string; virtual;
    class function GetDevicesList(Proc: TGetStrProc): Integer; virtual;
    class function GetDevicesStringList(Strings: TStrings): Integer; virtual;
    class function GetPlayDevicesList(Proc: TGetStrProc): Integer; virtual;
    class function GetPlayDevicesStringList(Strings: TStrings): Integer; virtual;
    class function GetRecDevicesList(Proc: TGetStrProc): Integer; virtual;
    class function GetRecDevicesStringList(Strings: TStrings): Integer; virtual;
    class function GetVoiceLen(Compression, Speed, Bits: Integer; DataSize: Longint): Integer; virtual;
    function HandleDLE(Code: Char): Integer; virtual; abstract;
    function HandleEvent(Event: Integer; Data: TEventData): Integer; virtual; abstract;
    function HangUp: Integer; virtual; abstract;
    function Init: Integer; virtual; abstract;
    function InitVoice: Integer; virtual; abstract;
    function ModemAssigned(const CheckFunc: string; DebugError: Boolean): Boolean;
    class function ModemName: string; virtual;
    class function ModemRMDName: string; virtual;
    function PlayDTMF(const Number: string): Integer; virtual; abstract;
    function PlayStream(Stream: TStream; BPS: Integer): Integer; virtual; abstract;
    function RecordStream(Stream: TStream; BPS: Integer): Integer; virtual; abstract;
    function RestoreVoiceMode: Integer; virtual; abstract;
    function SetCompression(var Compression, Speed, Bits: Integer): Integer; virtual; abstract;
    function SetDevice(Device: Integer): Integer; virtual; abstract;
    function StartPlayStream: Integer; virtual; abstract;
    function StartRecordStream: Integer; virtual; abstract;
    function StopDialing: Integer; virtual; abstract;
    function StopPlayStream: Integer; virtual; abstract;
    function StopPlaying: Integer; virtual; abstract;
    function StopRecordStream: Integer; virtual; abstract;
    function StopRecording: Integer; virtual; abstract;
    function StopWaiting: Integer; virtual; abstract;
    function SwitchToDataFax(const Mode: string): Integer; virtual; abstract;
    function VoiceModeOff: Integer; virtual; abstract;
    function VoiceModeOn: Integer; virtual; abstract;
    function Wait(TimeOut: DWORD): Integer; virtual; abstract;
    function WaitAnswer(const ExpectedAnswers: string; TimeOut: DWORD): Integer; virtual; abstract;
  published
    property PickPhoneCmd: string read FPickPhoneCmd write FPickPhoneCmd;
    property PickPhoneAnsw: string read FPickPhoneAnsw write FPickPhoneAnsw;
    property HangUpCmd: string read FHangUpCmd write FHangUpCmd;
    property HangUpAnsw: string read FHangUpAnsw write FHangUpAnsw;
    property BeepCmd: string read FBeepCmd write FBeepCmd;
    property BeepAnsw: string read FBeepAnsw write FBeepAnsw;
    property BeepTimeUnit: Integer read FBeepTimeUnit write FBeepTimeUnit;
    property HardFlowCmd: string read FHardFlowCmd write FHardFlowCmd;
    property HardFlowAnsw: string read FHardFlowAnsw write FHardFlowAnsw;
    property SoftFlowCmd: string read FSoftFlowCmd write FSoftFlowCmd;
    property SoftFlowAnsw: string read FSoftFlowAnsw write FSoftFlowAnsw;
    property StartPlayCmd: string read FStartPlayCmd write FStartPlayCmd;
    property StartPlayAnsw: string read FStartPlayAnsw write FStartPlayAnsw;
    property IntrPlayCmd: string read FIntrPlayCmd write FIntrPlayCmd;
    property IntrPlayAnsw: string read FIntrPlayAnsw write FIntrPlayAnsw;
    property StopPlayCmd: string read FStopPlayCmd write FStopPlayCmd;
    property StopPlayAnsw: string read FStopPlayAnsw write FStopPlayAnsw;
    property StartRecCmd: string read FStartRecCmd write FStartRecCmd;
    property StartRecAnsw: string read FStartRecAnsw write FStartRecAnsw;
    property StopRecCmd: string read FStopRecCmd write FStopRecCmd;
    property StopRecAnsw: string read FStopRecAnsw write FStopRecAnsw;
    property SwitchModeCmd: string read FSwitchModeCmd write FSwitchModeCmd;
    property SwitchModeAnsw: string read FSwitchModeAnsw write FSwitchModeAnsw;
    property VoiceModeOnCmd: string read FVoiceModeOnCmd write FVoiceModeOnCmd;
    property VoiceModeOnAnsw: string read FVoiceModeOnAnsw write FVoiceModeOnAnsw;
    property VoiceModeOffCmd: string read FVoiceModeOffCmd write FVoiceModeOffCmd;
    property VoiceModeOffAnsw: string read FVoiceModeOffAnsw write FVoiceModeOffAnsw;
    property PlayDTMFCmd: string read FPlayDTMFCmd write FPlayDTMFCmd;
    property PlayDTMFExtra: string read FPlayDTMFExtra write FPlayDTMFExtra;
    property PlayDTMFAnsw: string read FPlayDTMFAnsw write FPlayDTMFAnsw;
    property Modem: TVoiceModem read FModem write SetModem;
    property Profile: TModemProfile read GetProfile write FProfile;
  end;

{ TVoiceModem }

  {
    Voice Modem component.

    Procedures:

    Analyze             - Analyze the string in buffer for the appearance of
                          one of the possibilities given in expected_answers
                          in the form "OK|BUSY". If this fails, try to detect
                          some standard voice modem answers.
    AnswerPhone         - Set voice modem to answer mode.
    Beep                - Play default Beep signal.
    CheckEvents         - Process event queue.
    Close               - Hang up modem and close voice device.
    CloseDevice         - Close the voice device.
    Command             - Execute the given command and wait for an answer.
                          Returned OK if command output success and no
                          expected answers, VMA_USER_XXX if answer is one of
                          expected answer, VMA_FAIL if error.
    CommandFmt          - Format given command and execute it.
    DetectModemType     - Autodetect the modem type we are connected to. Not
                          implemented in version 1.0.
    Dial                - Dial given number.
    DriverAssigned      - Check driver. Generate debug error message if
                          driver does not exists (assigned).
    Flush               - Read input from the voice modem device.
    FlushEvents         - Flush event queue.
    FlushEx             - Read input from the voice modem device during
                          given timeout in msec.
    HandleDLE           - Handle given DLE code.
    HandleEvent         - Handle given event.
    HangUp              - Hang up (go on hook).
    Init                - Detect the type of voice modem connected (not
                          implemented in version 1.0) and initialize the
                          modem.
    InitVoice           - Set modem voice params (must be called only in
                          voice mode).
    Lock                - Lock modem for private usage. Use this function
                          for multiplie (multithreaded) using one
                          TVoiceModem instance or for prevent start new
                          modem operations
                          before any other operation still active.
    ModemName           - Return modem name for current modem driver.
    ModemRMDName        - Return modem RMD name for current modem driver.
                          This name is stored in RMF file header.
    Open                - Open voice device, automatically initialize modem
                          and switch to voice mode.
    OpenDevice          - Open voice device.
    PlayDTMF            - Play one of more DTMF tones.
    PlayFile            - Play the given file.
    PlayStream          - Play given stream.
    QueueEvent          - Post given event into event queue.
    Read                - Read data from the voice modem device. Returns
                          amount read byte or -1 when error.
    ReadChar            - Read char from the voice modem device.
    ReadStr             - Read string from the voice modem device.
    RecordFile          - Records voice data from the voice modem and save
                          them in the given file.
    RecordStream        - Record voice data into given stream.
    RestoreVoiceMode    - Switch modem to voice mode, reset voice
                          preferences, which are forgotten after leaving
                          voice mode.
    SetCompression      - Set compression mode. Return valid bits per
                          sample value.
    SetDevice           - Set device for voice playback and recording.
    Stop                - Queue stop operation event. 
    StopCurrentAction   - Check for current modem action and stop it.
    StopDialing         - Stop dialing.
    StopPlaying         - Stop playing.
    StopRecording       - Stop recording.
    StopWaiting         - Stop waiting.
    SwitchToDataFax     - Turn on data/fax mode.
    UnqueueEvent        - Get fist event and remove it from event queue.
    Unlock              - Unlock modem private usage. Return previous locked state.
    VoiceModeOn         - Entering voice mode.
    VoiceModeOff        - Leaving voice mode.
    Wait                - Wait modem char and handle DLE events during given
                          timeout in msec.
    WaitAnswer          - Wait modem standard or given answer, detect <DLE>
                          sequences during given timeout in msec.
    Write               - Write data to voice modem device.
    WriteChar           - Write char to voice modem device.
    WriteStr            - Write string to voice modem device (with CR).
    WriteStrFmt         - Format give string and write it to voice modem
                          device.

    Properties:

    Driver              - Hardware driver. If there are not value assigned at
                          design time the driver component instance will be
                          created at runtime with the specified ModemName
                          property of profile settings. If your need to use
                          own driver component (descended from TModemDriver
                          class) place it on form and set this property.
    LogFile             - Log file. All debug messages may be writed into
                          specified log file.
    Port                - Modem port. If there are not value assigned at
                          design time the standard TModemPort component
                          instance will be created at runtime. If your need
                          to use own port component (descended from
                          TModemPort class) place it on form and set this
                          property.
    Profile             - Modem profile. If property not specified at design
                          time a new instance of standard TModemProfile will
                          be created automatically with default settings at
                          runtime.
    TraceFile           - Trace file. Modem text input/output may be writed
                          into specified file.

    Events:

    OnChangeState       - Modem state changed.
    OnEvent             - Modem event queue handler.
    OnFlushChar         - Flush modem char event. Before calling this event
                          you get OnRead event.
    OnFlushString       - Flush modem string event. Before calling this event
                          you get OnFlushChar event. Event string does not
                          inclided chars such as CR, LF.
    OnDebugMsg          - Debug message event. WARNING: the Sender parameter
                          of event procedure not always is Self.
    OnProgress          - Progress callback for voice playing and recording.
    OnRead              - Read data from modem.
    OnReadChar          - Read single char from modem. Before calling this
                          event you get OnRead event.
    OnReadString        - Read string (command) from modem. Before calling
                          this event you get OnReadChar event for each char
                          readed from modem while reading string. Event
                          string does not inclided chars such as CR, LF, XON,
                          XOFF, DLE, ETX.   
    OnTrace             - Got one more modem input/output trace line. 
    OnWrite             - Write data to modem.
    OnWriteChar         - Write single char to modem. Before calling this
                          event you get OnWrite event.
    OnWriteString       - Write string (command) to modem. Before calling
                          this event you get OnWrite event. Event string
                          does not inclided CR char.
    OnYield             - Yield event. Use this event for process Windows
                          messages.
  }

  TModemEventHandler = procedure(Sender: TObject; Event: Integer; Data: TEventData; var Result: Integer) of object;

  TModemDataEvent = procedure(Sender: TObject; Buffer: Pointer; Count: Integer) of object;
  TModemCharEvent = procedure(Sender: TObject; C: Char) of object;
  TModemStringEvent = procedure(Sender: TObject; const S: string) of object;
  TModemTraceEvent = procedure(Sender: TObject; const Action, Line: string) of object;

  TVoiceModem = class(TComponent)
  private
    FPort: TModemPort;
    FDriver: TModemDriver;
    FProfile: TModemProfile;
    FLogFile: TModemLogFile;
    FTraceFile: TModemLogFile;
    FTraceWasCR: Boolean;
    FTraceLine: string;
    FTraceAction: string;
    FEventCount: Integer;
    FFirstEvent: Integer;
    FLastEvent: Integer;
    FEventQueue: TEventQueue;
    FState: Integer;
    FStateSection: TRTLCriticalSection;
    FLocked: Boolean;
    FLockOwner: TObject;
    FLockSection: TRTLCriticalSection;
    FLastError: Integer;
    FOnChangeState: TNotifyEvent;
    FOnEvent: TModemEventHandler;
    FOnDebugMsg: TMessageEvent;
    FOnRead: TModemDataEvent;
    FOnReadChar: TModemCharEvent;
    FOnReadString: TModemStringEvent;
    FOnWrite: TModemDataEvent;
    FOnWriteChar: TModemCharEvent;
    FOnWriteString: TModemStringEvent;
    FOnFlushChar: TModemCharEvent;
    FOnFlushString: TModemStringEvent;
    FOnTrace: TModemTraceEvent;
    FOnProgress: TProgressEvent;
    FOnYield: TNotifyEvent;
    function GetDriver: TModemDriver;
    function GetLocked: Boolean;
    function GetLockOwner: TObject;
    function GetOpened: Boolean;
    function GetPort: TModemPort;
    function GetProfile: TModemProfile;
    function GetState: Integer;
    procedure SetDriver(Value: TModemDriver);
    procedure SetLogFile(Value: TModemLogFile);
    procedure SetPort(Value: TModemPort);
    procedure SetProfile(Value: TModemProfile);
    procedure SetState(Value: Integer);
    procedure SetTraceFile(Value: TModemLogFile);
  protected
    procedure ChangeState; virtual;
    procedure DebugMsg(Msg: Integer; const S: string); virtual;
    procedure DebugMsgFmt(Msg: Integer; const S: string; const Args: array of const); virtual;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure PortError(Sender: TObject; Errors: Integer); virtual;
    procedure Progress(Done: Longint); virtual;
    procedure TraceChar(const Action: string; C: Char); virtual;
    procedure Yield; virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function Analyze(const Buffer, ExpectedAnswers: string): Integer;
    function AnswerPhone: Integer;
    function Beep(Frequency, Length: Integer): Integer;
    function CheckEvents: Integer;
    function Close: Integer;
    function CloseDevice: Integer;
    function Command(const Command, ExpectedAnswers: string): Integer;
    function CommandFmt(const Command, ExpectedAnswers: string; const Args: array of const): Integer;
    function DetectModemType: Integer;
    function Dial(const Number: string): Integer;
    function DriverAssigned(const CheckFunc: string; DebugError: Boolean): Boolean;
    function Flush: Integer;
    function FlushEvents: Integer;
    function FlushEx(TimeOut: DWORD): Integer;
    function HandleDLE(Code: Char): Integer; 
    function HandleEvent(Event: Integer; Data: TEventData): Integer; virtual;
    function HangUp: Integer;
    function Init: Integer;
    function InitVoice: Integer;
    function Lock(Owner: TObject): Boolean;
    function ModemName: string;
    function ModemRMDName: string;
    function Open: Integer;
    function OpenDevice: Integer;
    function PlayDTMF(const Number: string): Integer;
    function PlayFile(const FileName: string): Integer;
    function PlayStream(Stream: TStream): Integer;
    function QueueEvent(Event: TModemEvent): Integer;
    function Read(Buffer: Pointer; Count: Integer): Integer;
    function ReadChar(var CharIn: Char): Integer;
    function ReadStr(var S: string): Integer;
    function RecordFile(const FileName: string): Integer;
    function RecordStream(Stream: TStream): Integer;
    function RestoreVoiceMode: Integer;
    function SetCompression(var Compression, Speed, Bits: Integer): Integer;
    function SetDevice(Device: Integer): Integer;
    function Stop: Integer;
    function StopCurrentAction: Integer;
    function StopDialing: Integer;
    function StopPlaying: Integer;
    function StopRecording: Integer;
    function StopWaiting: Integer;
    function SwitchToDataFax(Mode: string): Integer;
    function UnqueueEvent(var Event: TModemEvent): Integer;
    function Unlock: Boolean;
    function VoiceModeOff: Integer;
    function VoiceModeOn: Integer;
    function Wait(TimeOut: DWORD): Integer;
    function WaitAnswer(const ExpectedAnswers: string; TimeOut: DWORD): Integer;
    function Write(Buffer: Pointer; Count: Integer): Integer;
    function WriteChar(CharOut: Char): Integer;
    function WriteStr(const S: string): Integer;
    function WriteStrFmt(const Format: string; const Args: array of const): Integer;
    property LastError: Integer read FLastError write FLastError;
    property Locked: Boolean read GetLocked;
    property LockOwner: TObject read GetLockOwner;
    property Opened: Boolean read GetOpened;
    property State: Integer read GetState write SetState;
  published
    property Driver: TModemDriver read GetDriver write SetDriver;
    property LogFile: TModemLogFile read FLogFile write SetLogFile;
    property Port: TModemPort read GetPort write SetPort;
    property Profile: TModemProfile read GetProfile write SetProfile;
    property TraceFile: TModemLogFile read FTraceFile write SetTraceFile;
    property OnChangeState: TNotifyEvent read FOnChangeState write FOnChangeState;
    property OnEvent: TModemEventHandler read FOnEvent write FOnEvent;
    property OnFlushChar: TModemCharEvent read FOnFlushChar write FOnFlushChar;
    property OnFlushString: TModemStringEvent read FOnFlushString write FOnFlushString;
    property OnDebugMsg: TMessageEvent read FOnDebugMsg write FOnDebugMsg;
    property OnProgress: TProgressEvent read FOnProgress write FOnProgress;
    property OnRead: TModemDataEvent read FOnRead write FOnRead;
    property OnReadChar: TModemCharEvent read FOnReadChar write FOnReadChar;
    property OnReadString: TModemStringEvent read FOnReadString write FOnReadString;
    property OnTrace: TModemTraceEvent read FOnTrace write FOnTrace;
    property OnWrite: TModemDataEvent read FOnWrite write FOnWrite;
    property OnWriteChar: TModemCharEvent read FOnWriteChar write FOnWriteChar;
    property OnWriteString: TModemStringEvent read FOnWriteString write FOnWriteString;
    property OnYield: TNotifyEvent read FOnYield write FOnYield;
  end;

{ Modem drivers registration routines }

procedure RegisterModemDriver(const ModemName: string; DriverClass: TModemDriverClass);
procedure UnRegisterModemDriver(const ModemName: string);

procedure ReplaceModemDriver(const ModemName: string; NewDriverClass: TModemDriverClass);
procedure ReplaceModemDriverEx(const ModemName: string; NewDriverClass: TModemDriverClass; var OldDriverClass: TModemDriverClass);

function GetModemDriver(const ModemName: string; var DriverClass: TModemDriverClass): Boolean;
function GetModemDriversList(Proc: TGetStrProc): Integer;
function GetModemDriversStringList(Strings: TStrings): Integer;

implementation

{$I *.INC}

var
  FDriversList: TStringList;

function DriversList: TStringList;
begin
  if FDriversList = nil then FDriversList := TStringList.Create;
  Result := FDriversList;
end;

procedure RegisterModemDriver(const ModemName: string; DriverClass: TModemDriverClass);
begin
  DriversList.AddObject(ModemName, TObject(DriverClass));
end;

procedure UnRegisterModemDriver(const ModemName: string);
var
  Index: Integer;
begin
  Index := DriversList.IndexOf(ModemName);
  { Unregister driver }
  if Index <> -1 then DriversList.Delete(Index);
end;

procedure ReplaceModemDriver(const ModemName: string; NewDriverClass: TModemDriverClass);
var 
  OldDriverClass: TModemDriverClass;
begin
  ReplaceModemDriverEx(ModemName, NewDriverClass, OldDriverClass);
end;

procedure ReplaceModemDriverEx(const ModemName: string; NewDriverClass: TModemDriverClass; var OldDriverClass: TModemDriverClass);
var
  Index: Integer;
begin
  { Search registered driver }
  Index := DriversList.IndexOf(ModemName);
  if (Index <> -1) then 
  begin
    { Unregister old driver }
    OldDriverClass := TModemDriverClass(DriversList.Objects[Index]);
    DriversList.Delete(Index);
  end
  else
    { No driver registered for this modem }
    OldDriverClass := nil;
  { Register new driver }
  DriversList.AddObject(ModemName, TObject(NewDriverClass));
end;

function GetModemDriver(const ModemName: string; var DriverClass: TModemDriverClass): Boolean;
var
  Index: Integer;
begin
  Index := DriversList.IndexOf(ModemName);
  { Driver class not found }
  if Index = -1 then
  begin
    DriverClass := nil;
    Result := False;
    Exit;
  end;
  { Success }
  DriverClass := TModemDriverClass(DriversList.Objects[Index]);
  Result := True;
end;

function GetModemDriversList(Proc: TGetStrProc): Integer;
var
  I: Integer;
begin
  Result := 0;
  for I := 0 to DriversList.Count - 1 do
  begin
    Proc(DriversList[I]);
    Inc(Result);
  end;
end;

function GetModemDriversStringList(Strings: TStrings): Integer;
type
  TGetStrFunc = function(const S: string): Integer of object;
var
  Proc: TGetStrFunc;
begin
  Proc := Strings.Add;
  Result := GetModemDriversList(TGetStrProc(Proc));
end;

{ TModemDriver }

function TModemDriver.GetProfile: TModemProfile;
begin
  if FProfile <> nil then
  begin
    Result := FProfile;
    Exit;
  end;
  if not (csDesigning in ComponentState) and (Modem <> nil) then
  begin
    Result := Modem.Profile;
    Exit;
  end;
  Result := nil;
end;

procedure TModemDriver.SetModem(Value: TVoiceModem);
var
  SModem: TVoiceModem;
begin
  if FModem <> Value then
  begin
    SModem := FModem;
    FModem := Value;
    if SModem <> nil then SModem.Driver := nil;
    if FModem <> nil then FModem.Driver := Self;
  end;
end;

procedure TModemDriver.DebugMsg(Msg: Integer; const S: string);
begin
  { Redirect message to modem }
  if FModem <> nil then FModem.DebugMsg(Msg, S);
end;

procedure TModemDriver.DebugMsgFmt(Msg: Integer; const S: string; const Args: array of const);
begin
  try
    DebugMsg(Msg, Format(S, Args));
  except
    { Ignore format exceptions }
  end;
end;

procedure TModemDriver.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
  begin
    if AComponent = FModem then FModem := nil;
    if AComponent = FProfile then FProfile := nil;
  end;
end;

procedure TModemDriver.Progress(Done: Longint);
begin
  { Redirect event to modem }
  if FModem <> nil then FModem.Progress(Done);
end;

procedure TModemDriver.Yield;
begin
  { Redirect event to modem }
  if FModem <> nil then FModem.Yield;
end;

class function TModemDriver.GetCompression(const Name: string): Integer;
var
  Map: PIdentMap;
  Count: Integer;
begin
  Result := StrToIntDef(Name, -1);
  { Convert compression name }
  Count := GetCompressionMap(Map);
  if Count > 0 then Result := IdentToIntDef(Name, Result, Slice(Map^, Count));
end;

class function TModemDriver.GetCompressionMap(var Map: PIdentMap): Integer;
begin
  Map := nil;
  Result := 0;
end;

class function TModemDriver.GetCompressionName(Compression: Integer): string;
var
  Map: PIdentMap;
  Count: Integer;
begin
  Result := IntToStr(Compression);
  { Convert compression number }
  Count := GetCompressionMap(Map);
  if Count > 0 then Result := IntToIdentDef(Compression, Result, Slice(Map^, Count));
end;

class function TModemDriver.GetCompressionsList(Proc: TGetStrProc): Integer;
var
  Map: PIdentMap;
  Count: Integer;
begin
  Result := 0;
  { Fill compression list }
  Count := GetCompressionMap(Map);
  if Count > 0 then Result := GetIdentsList(Slice(Map^, Count), Proc);
end;

class function TModemDriver.GetCompressionsStringList(Strings: TStrings): Integer;
type
  TGetStrFunc = function(const S: string): Integer of object;
var
  Proc: TGetStrFunc;
begin
  Proc := Strings.Add;
  Result := GetCompressionsList(TGetStrProc(Proc));
end;

class function TModemDriver.GetDevice(const Name: string): Integer;
var
  Map: PIdentMap;
  Count: Integer;
begin
  Result := StrToIntDef(Name, -1);
  { Convert device name }
  Count := GetDeviceMap(Map);
  if Count > 0 then Result := IdentToIntDef(Name, Result, Slice(Map^, Count))
end;

class function TModemDriver.GetDeviceMap(var Map: PIdentMap): Integer;
begin
  Map := nil;
  Result := 0;
end;

class function TModemDriver.GetDeviceName(Device: Integer): string;
var
  Map: PIdentMap;
  Count: Integer;
begin
  Result := IntToStr(Device);
  { Convert device number }
  Count := GetDeviceMap(Map);
  if Count > 0 then Result := IntToIdentDef(Device, Result, Slice(Map^, Count));
end;

class function TModemDriver.GetDevicesList(Proc: TGetStrProc): Integer;
var
  Map: PIdentMap;
  Count: Integer;
begin
  Result := 0;
  { Fill device list }
  Count := GetDeviceMap(Map);
  if Count > 0 then Result := GetIdentsList(Slice(Map^, Count), Proc);
end;

class function TModemDriver.GetDevicesStringList(Strings: TStrings): Integer;
type
  TGetStrFunc = function(const S: string): Integer of object;
var
  Proc: TGetStrFunc;
begin
  Proc := Strings.Add;
  Result := GetDevicesList(TGetStrProc(Proc));
end;

class function TModemDriver.GetPlayDevicesList(Proc: TGetStrProc): Integer;
var
  Map, TempMap: PIdentMap;
  Count, TempCount, I, Device: Integer;
begin
  Result := 0;
  { Get device map }
  Count := GetDeviceMap(Map);
  { Check device count }
  if Count > 0 then
  begin
    { Create play device map buffer }
    GetMem(TempMap, Count * SizeOf(TIdentMapEntry));
    try
      FillChar(TempMap^, Count * SizeOf(TIdentMapEntry), 0);
      TempCount := 0;
      { Create play device map }
      for I := 0 to Count - 1 do
      begin
        Device := Map^[I].Value;
        { Ignore record devices }
        if (Device <> EXTERNAL_MICROPHONE) and (Device <> INTERNAL_MICROPHONE) then
        begin
          TempMap^[TempCount].Value := Map^[I].Value;
          TempMap^[TempCount].Name := Map^[I].Name;
          Inc(TempCount);
        end;
      end;
      { Get play device list }
      Result := GetIdentsList(Slice(TempMap^, TempCount), Proc);
    finally
      FreeMem(TempMap, Count * SizeOf(TIdentMapEntry));
    end;
  end;
end;

class function TModemDriver.GetPlayDevicesStringList(Strings: TStrings): Integer;
type
  TGetStrFunc = function(const S: string): Integer of object;
var
  Proc: TGetStrFunc;
begin
  Proc := Strings.Add;
  Result := GetPlayDevicesList(TGetStrProc(Proc));
end;

class function TModemDriver.GetRecDevicesList(Proc: TGetStrProc): Integer;
var
  Map, TempMap: PIdentMap;
  Count, TempCount, I, Device: Integer;
begin
  Result := 0;
  { Get device map }
  Count := GetDeviceMap(Map);
  { Check device count }
  if Count > 0 then
  begin
    { Create record device map buffer }
    GetMem(TempMap, Count * SizeOf(TIdentMapEntry));
    try
      FillChar(TempMap^, Count * SizeOf(TIdentMapEntry), 0);
      TempCount := 0;
      { Create record device map }
      for I := 0 to Count - 1 do
      begin
        Device := Map^[I].Value;
        { Ignore play devices }
        if (Device <> EXTERNAL_SPEAKER) and (Device <> INTERNAL_SPEAKER) then
        begin
          TempMap^[TempCount] := Map^[I];
          Inc(TempCount);
        end;
      end;
      { Get record device list }
      Result := GetIdentsList(Slice(TempMap^, TempCount), Proc);
    finally
      FreeMem(TempMap, Count * SizeOf(TIdentMapEntry));
    end;
  end;
end;

class function TModemDriver.GetVoiceLen(Compression, Speed, Bits: Integer; DataSize: Longint): Integer;
begin
  Result := 0;
end;

class function TModemDriver.GetRecDevicesStringList(Strings: TStrings): Integer;
type
  TGetStrFunc = function(const S: string): Integer of object;
var
  Proc: TGetStrFunc;
begin
  Proc := Strings.Add;
  Result := GetRecDevicesList(TGetStrProc(Proc));
end;

function TModemDriver.ModemAssigned(const CheckFunc: string; DebugError: Boolean): Boolean;
begin
  Result := Modem <> nil;
  { Generate error }
  if (not Result) and DebugError then DebugMsgFmt(L_ERROR, SNoModemError, [CheckFunc]);
end;

class function TModemDriver.ModemName: string;
begin
  Result := '';
end;

class function TModemDriver.ModemRMDName: string;
begin
  Result := '';
end;

{ TModemEvent }

constructor TModemEvent.Create(AEvent: Integer);
begin
  FEvent := AEvent;
end;

constructor TModemEvent.CreateDataC(AEvent: Integer; C: Char);
begin
  FEvent := AEvent;
  FData.C := C;
end;

constructor TModemEvent.CreateDataI(AEvent: Integer; I: Integer);
begin
  FEvent := AEvent;
  FData.I := I;
end;

constructor TModemEvent.CreateDataP(AEvent: Integer; P: Pointer);
begin
  FEvent := AEvent;
  FData.P := P;
end;

{ TVoiceModem }
         
constructor TVoiceModem.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  { Init modem state }
  FState := IDLE;
  { Init critical sections }
  InitializeCriticalSection(FStateSection);
  InitializeCriticalSection(FLockSection);
end;

destructor TVoiceModem.Destroy;
begin
  DeleteCriticalSection(FLockSection);
  DeleteCriticalSection(FStateSection);
  inherited Destroy;
end;

function TVoiceModem.GetDriver: TModemDriver;
var
  DriverClass: TModemDriverClass;
begin
  if not (csDesigning in ComponentState) then
  begin
    if FDriver = nil then
      if GetModemDriver(Profile.ModemName, DriverClass) then
      begin
        FDriver := DriverClass.Create(Self);
        FDriver.Modem := Self;
        FDriver.Profile := Profile;
      end;
  end;
  Result := FDriver;
end;

function TVoiceModem.GetLocked: Boolean;
begin
  EnterCriticalSection(FLockSection);
  Result := FLocked;
  LeaveCriticalSection(FLockSection);
end;

function TVoiceModem.GetLockOwner: TObject;
begin
  EnterCriticalSection(FLockSection);
  Result := FLockOwner;
  LeaveCriticalSection(FLockSection);
end;

function TVoiceModem.GetOpened: Boolean;
begin
  Result := Port.Enabled;
end;

function TVoiceModem.GetPort: TModemPort;
begin
  if not (csDesigning in ComponentState) and (FPort = nil) then
  begin
    FPort := TModemPort.Create(Self);
    FPort.Profile := Profile;
  end;
  Result := FPort;
end;

function TVoiceModem.GetProfile: TModemProfile;
begin
  if not (csDesigning in ComponentState) and (FProfile = nil) then
  begin
    FProfile := TModemProfile.Create(Self);
    if (FPort <> nil) and (FPort.Profile = nil) then FPort.Profile := FProfile;
    if (FDriver <> nil) and (FDriver.Profile = nil) then FDriver.Profile := FProfile;
  end;
  Result := FProfile;
end;

function TVoiceModem.GetState: Integer;
begin
  EnterCriticalSection(FStateSection);
  Result := FState;
  LeaveCriticalSection(FStateSection);
end;

procedure TVoiceModem.SetDriver(Value: TModemDriver);
var
  SDriver: TModemDriver;
begin
  if FDriver <> Value then
  begin
    SDriver := FDriver;
    { Set new profile }
    FDriver := Value;
    if FDriver <> nil then FDriver.FModem := Self;
    { Free default profile component }
    if SDriver <> nil then
    begin
      SDriver.FModem := nil;
      if SDriver.Owner = Self then SDriver.Free;
    end;
  end;
end;

procedure TVoiceModem.SetLogFile(Value: TModemLogFile);
begin
  if FLogFile <> Value then
  begin
    if FLogFile <> nil then FLogFile.Close;
    FLogFile := Value;
  end;
end;

procedure TVoiceModem.SetPort(Value: TModemPort);
var
  SPort: TModemPort;
begin
  if FPort <> Value then
  begin
    SPort := FPort;
    { Set new profile }
    FPort := Value;
    { Free default profile component }
    if (SPort <> nil) and (SPort.Owner = Self) then SPort.Free;
  end;
end;

procedure TVoiceModem.SetProfile(Value: TModemProfile);
var
  SProfile: TModemProfile;
begin
  if FProfile <> Value then
  begin
    SProfile := FProfile;
    { Set new profile }
    FProfile := Value;
    { Free default profile component }
    if (SProfile <> nil) and (SProfile.Owner = Self) then SProfile.Free;
  end;
end;

procedure TVoiceModem.SetState(Value: Integer);
begin
  EnterCriticalSection(FStateSection);
  if FState <> Value then
  begin
    FState := Value;
    ChangeState;
  end;
  LeaveCriticalSection(FStateSection);
end;

procedure TVoiceModem.SetTraceFile(Value: TModemLogFile);
begin
  if FTraceFile <> Value then
  begin
    if FTraceFile <> nil then FTraceFile.Close;
    FTraceFile := Value;
  end;
end;

procedure TVoiceModem.ChangeState;
begin
  if not (csDestroying in ComponentState) and Assigned(FOnChangeState) then
  try
    FOnChangeState(Self);
  except
    on E: Exception do DebugMsg(L_ERROR, 'ChangeState: ' + E.Message);
  end;
  { Yield event }
  Yield;
end;

procedure TVoiceModem.DebugMsg(Msg: Integer; const S: string);
begin
  { Write log message }
  if (FLogFile <> nil) and Profile.ModemLogging then FLogFile.Add(S);
  { Call debug event }
  if not (csDestroying in ComponentState) and Assigned(FOnDebugMsg) then
  try
    FOnDebugMsg(Self, Msg, S);
  except
    { Ignore exceptions }
  end;
  { Yield event }
  Yield;
end;

procedure TVoiceModem.DebugMsgFmt(Msg: Integer; const S: string; const Args: array of const);
begin
  try
    DebugMsg(Msg, Format(S, Args));
  except
    { Ignore format exceptions }
  end;
end;

procedure TVoiceModem.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
  begin
    if AComponent = FPort then FPort := nil;
    if AComponent = FDriver then FDriver := nil;
    if AComponent = FProfile then FProfile := nil;
    if AComponent = FLogFile then FLogFile := nil;
  end;
end;

procedure TVoiceModem.PortError(Sender: TObject; Errors: Integer);
begin
  DebugMsgFmt(L_ERROR, SPortError, [Errors]);
  QueueEvent(TModemEvent.Create(SIGNAL_SIGQUIT));
end;

procedure TVoiceModem.Progress(Done: Longint);
begin
  if Assigned(FOnProgress) then
  try
    FOnProgress(Self, Done);
  except
    on E: Exception do DebugMsg(L_ERROR, 'Progress: ' + E.Message);
  end;
  { Yield event }
  Yield;
end;

procedure TVoiceModem.TraceChar(const Action: string; C: Char);

  function CharToCode(C: Char): string;
  begin
    if C >= ' ' then
      Result := C
    else
      Result := '#' + IntToStr(Ord(C));
  end;

  procedure LineComplete;
  begin
    { Write trace to file }
    if (FTraceFile <> nil) and Profile.ModemLogging then FTraceFile.Add(FTraceLine);
    { Trace message }
    if Assigned(FOnTrace) then
    try
      FOnTrace(Self, FTraceAction, FTraceLine);
    except
      on E: Exception do DebugMsg(L_ERROR, 'Trace: ' + E.Message);
    end;
    { Clear trace line }
    FTraceLine := '';
  end;

begin
  { Check for CR and LF }
  case C of
    CR:
      { CR char }
      begin
        FTraceWasCR := True;
        Exit;
      end;
    LF:
      { LF char }
      if FTraceWasCR then
      begin
        { Check line complete }
        if FTraceAction = Action then
        begin
          FTraceWasCR := False;
          LineComplete;
        end
        else
          { Trace direction changed - ignore char }
          Exit;
      end
      else
        { Add char to trace line }
        FTraceLine := FTraceLine + CharToCode(C);
    else
      { Any char }
      begin
        { Check single CR without LF }
        if FTraceWasCR then
        begin
          FTraceWasCR := False;
          LineComplete;
        end;
        { Add char to trace line }
        FTraceLine := FTraceLine + CharToCode(C);
      end;
  end;
  { Save current action }
  FTraceAction := Action;
end;

procedure TVoiceModem.Yield;
begin
  if Assigned(FOnYield) then
  try
    FOnYield(Self);
  except
    on E: Exception do DebugMsg(L_ERROR, 'Yield: ' + E.Message);
  end;
end;

function TVoiceModem.Analyze(const Buffer, ExpectedAnswers: string): Integer;
var
  UserAnswers: string;
  Current: Integer;
  CurrentAnswer: string;
  DividerPos: Integer;
begin
  { Check for buffer length }
  if Length(Buffer) = 0 then
  begin
    Result := VMA_EMPTY;
    Exit;
  end;
  UserAnswers := ExpectedAnswers;
  Current := 0;
  { Analyze buffer }
  while Length(UserAnswers) > 0 do
  begin
    DividerPos := Pos('|', UserAnswers);
    { Get next expected answer }
    if DividerPos = 0 then
    begin
      CurrentAnswer := UserAnswers;
      UserAnswers := '';
    end
    else
    begin
      CurrentAnswer := Copy(UserAnswers, 1, DividerPos - 1);
      UserAnswers := Copy(UserAnswers, DividerPos + 1, MaxInt);
    end;
    { Check bufferfor current answer }
    if Pos(CurrentAnswer, Buffer) = 1 then
    begin
      Result := VMA_USER + Current;
      Exit;
    end;
    Inc(Current);
  end;
  { Get standard modem answer }
  Result := StrToModemAnswer(Buffer);
end;

function TVoiceModem.AnswerPhone: Integer;
begin
  DebugMsg(L_MESG, SAnsweringCall);
  { Check driver specified }
  if not DriverAssigned('AnswerPhone', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Answer phone }
  Result := Driver.AnswerPhone;
end;

function TVoiceModem.Beep(Frequency, Length: Integer): Integer;
begin
  DebugMsg(L_MESG, SCheckingBeep);
  { Check driver specified }
  if not DriverAssigned('Beep', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Beep }
  Result := Driver.Beep(Frequency, Length);
end;

function TVoiceModem.CheckEvents: Integer;
var
  Event: TModemEvent;
begin
  while UnqueueEvent(Event) = OK do
  begin
    if HandleEvent(Event.Event, Event.Data) = FAIL then
    begin
      DebugMsg(L_ERROR, SHandleEventFailed);
      LastError := VME_EVENT_NOT_HANDLED;
      Result := FAIL;
      Exit; {?}
    end;
    Event.Free;
  end;
  LastError := VME_SUCCESS;
  Result := OK;
end;

function TVoiceModem.Close: Integer;
begin
  { Check modem already closed }
  if not Opened then
  begin
    LastError := VME_SUCCESS;
    Result := OK;
    Exit;
  end;
  { Hang up the modem and close device }
  try
    HangUp;
  finally
    Result := CloseDevice;
  end;
end;

function TVoiceModem.CloseDevice: Integer;
begin
  DebugMsg(L_MESG, SClosingDevice);
  { Check device opened }
  if Opened then
  begin
    { Close device }
    Port.Close;
    { Check port closing result }
    if Port.Enabled then
    begin
      DebugMsg(L_ERROR, SClosingDeviceError);
      LastError := VME_PORT_ERROR;
      Result := FAIL;
      Exit;
    end;
    DebugMsg(L_MESG, SDeviceClosed);
  end
  else
    DebugMsg(L_WARN, SNoOpenDevice);
  { Close log files }
  if FLogFile <> nil then FLogFile.Close;
  if FTraceFile <> nil then FTraceFile.Close;
  { Result }
  LastError := VME_SUCCESS;
  Result := OK;
end;

function TVoiceModem.Command(const Command, ExpectedAnswers: string): Integer;
var
  Buffer: string;
begin
  DebugMsgFmt(L_NOISE, SCommand, [Command, ExpectedAnswers]);
  Result := VMA_FAIL;
  { Command delay }
  if Profile.CommandDelay > 0 then SleepEx(Profile.CommandDelay, False);
  { Process command }
  if Length(Command) <> 0 then
  begin
    { Out command }
    if WriteStr(Command) = FAIL then
    begin
      Flush;
      Result := VMA_FAIL;
      Exit;
    end;
    { Analyze modem command echo }
    if Profile.CommandEcho then
    repeat
      if ReadStr(Buffer) = FAIL then
      begin
        Flush;
        Result := VMA_FAIL;
        Exit;
      end;
      Result := Analyze(Buffer, Command);
      if Result = VMA_FAIL then
      begin
        DebugMsg(L_ERROR, SNoEchoError);
        Flush;
        Exit;
      end;
      if Result = VMA_ERROR then
      begin
        DebugMsg(L_ERROR, SErrorReturned);
        Flush;
        Result := VMA_FAIL;
        Exit;
      end;
    until Result = VMA_USER_1;
    { Command send success }
    Result := OK;
  end;
  { Process modem answer }
  if Length(ExpectedAnswers) <> 0 then
    repeat
      { Read modem answer }
      if ReadStr(Buffer) = FAIL then
      begin
        Flush;
        Result := VMA_FAIL;
        Exit;
      end;
      { Analyze modem answer }
      Result := Analyze(Buffer, ExpectedAnswers);
      { Test result }
      if Result = VMA_FAIL then
      begin
        DebugMsg(L_ERROR, SInvalidAnswer);
        Flush;
        LastError := VME_INVALID_MODEM_ANSWER;
        Exit;
      end;
      if Result = VMA_ERROR then
      begin
        DebugMsg(L_ERROR, SErrorReturned);
        Flush;
        Result := VMA_FAIL;
        Exit;
      end;
    until Result and VMA_USER = VMA_USER;
end;

function TVoiceModem.CommandFmt(const Command, ExpectedAnswers: string; const Args: array of const): Integer;
begin
  try
    Result := Self.Command(Format(Command, Args), ExpectedAnswers);
  except
    DebugMsg(L_ERROR, SInvalidCommandFormat);
    LastError := VME_INVALID_COMMAND_FORMAT;
    Result := VMA_FAIL;
  end;
end;

function TVoiceModem.DetectModemType: Integer;
begin
  LastError := VME_NOT_IMPLEMENTED; 
  Result := FAIL;
end;

function TVoiceModem.Dial(const Number: string): Integer;
begin
  { Check driver specified }
  if not DriverAssigned('Dial', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Dial number }
  Result := Driver.Dial(Number);
end;

function TVoiceModem.DriverAssigned(const CheckFunc: string; DebugError: Boolean): Boolean;
begin
  Result := Driver <> nil;
  { Generate error }
  if (not Result) and DebugError then
    { Check Profile.ModemName property }
    if Length(Profile.ModemName) = 0 then
    begin
      { No modem name specified }
      DebugMsgFmt(L_ERROR, SNoModemError, [CheckFunc]);
      LastError := VME_MODEM_NOT_SPECIFIED;
    end
    else
    begin
      { Driver class is not registered for specified modem name }
      DebugMsgFmt(L_ERROR, SNoDriverError, [CheckFunc]);
      LastError := VME_DRIVER_NOT_SPECIFIED;
    end
end;

function TVoiceModem.Flush: Integer;
begin
  Result := FlushEx(0);
end;

function TVoiceModem.FlushEvents: Integer;
var
  Event: TModemEvent;
begin
  while UnqueueEvent(Event) = OK do
  begin
    DebugMsgFmt(L_WARN, SFlushEvent, [Event.Event, Event.Data.C]);
    Event.Free;
  end;
  LastError := VME_SUCCESS;
  Result := OK;
end;

function TVoiceModem.FlushEx(TimeOut: DWORD): Integer;
var
  StartTime: DWORD;
  FlushChar: Char;
  FlushStr: string;
  FirstChar: Boolean;
begin
  StartTime := GetTickCount;
  FlushStr := '';
  FirstChar := True;
  { Flush data by one char }
  while (TimeOut = 0) or (DWORD(Abs(GetTickCount - StartTime)) < TimeOut) do
  begin
    { Check for more flush data }
    if not Port.CharReady then
    begin
      SleepEx(Profile.PortStatusDelay + Profile.PortReadDelay, False);
      if not Port.CharReady then
        if TimeOut = 0 then Break else Continue;
    end;
    { Read modem char }
    if Read(@FlushChar, 1) <> 1 then
    begin
      Result := FAIL;
      Exit;
    end;
    { Trace input char }
    TraceChar('Recv', FlushChar);
    { Flush char event }
    if Assigned(FOnFlushChar) then
    try
      FOnFlushChar(Self, FlushChar);
    except
      on E: Exception do DebugMsg(L_ERROR, 'FlushChar: ' + E.Message);
    end;
    { Check for line break }
    if FlushChar in [LF, CR] then
    begin
      if not FirstChar then
      begin
        DebugMsgFmt(L_JUNK, SFlushString, [FlushStr]);
        { Flush string event }
        if Assigned(FOnFlushString) then
        try
          FOnFlushString(Self, FlushStr);
        except
          on E: Exception do DebugMsg(L_ERROR, 'FlushString: ' + E.Message);
        end;
      end;
      FlushStr := '';
      FirstChar := True;
    end
    else
      if FirstChar then
      begin
        FlushStr := FlushChar;
        FirstChar := False;
      end
      else
        FlushStr := FlushStr + FlushChar;
  end;
  { Check for lost debug message }
  if Length(FlushStr) > 0 then DebugMsgFmt(L_JUNK, SFlushString, [FlushStr]);
  { Success }
  Result := OK;
end;

function TVoiceModem.HandleDLE(Code: Char): Integer;
begin
  { Check driver specified }
  if not DriverAssigned('HandleDLE', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Handle DLE code }
  Result := Driver.HandleDLE(Code);
end;

function TVoiceModem.HandleEvent(Event: Integer; Data: TEventData): Integer;
begin
  DebugMsgFmt(L_JUNK, SGotEvent, [Event, Data.C]);
  { Driver handling }
  if DriverAssigned('HandleEvent', True) then
  begin
    Result := Driver.HandleEvent(Event, Data);
    if Result <> UNKNOWN_EVENT then Exit;
  end;
  { User handling }
  if Assigned(FOnEvent) then
  begin
    Result := UNKNOWN_EVENT;
    try
      FOnEvent(Self, Event, Data, Result);
    except
      on E: Exception do DebugMsg(L_ERROR, 'HandleEvent: ' + E.Message);
    end;
    if Result <> UNKNOWN_EVENT then Exit;
  end;
  { Default handling }
  case Event of
    SIGNAL_STOP:
      begin
        StopCurrentAction;
        Result := OK;
        Exit;
      end;
  end;
  case State of
    DIALING:
      case Event of
        BUSY_TONE,
        DATA_CALLING_TONE,
        FAX_CALLING_TONE:
          begin
            Result := StopDialing;
            Exit;
          end;
      end;
    PLAYING:
      case Event of
        BUSY_TONE,
        DIAL_TONE,
        DATA_CALLING_TONE,
        FAX_CALLING_TONE:
          begin
            Result := StopPlaying;
            Exit;
          end;
      end;
    RECORDING:
      case Event of
        BUSY_TONE,
        DIAL_TONE,
        DATA_CALLING_TONE,
        FAX_CALLING_TONE,
        NO_VOICE_ENERGY:
          begin
            Result := StopRecording;
            Exit;
          end;
        SILENCE_DETECTED:
          if not Profile.RecIgnoreSilence then
          begin
            Result := StopRecording;
            Exit;
          end
          else
          begin
            Result := OK;
            Exit;
          end;
      end;
    WAITING:
      case Event of
        BUSY_TONE,
        DIAL_TONE,
        DATA_CALLING_TONE,
        FAX_CALLING_TONE,
        RING_DETECTED:
          begin
            Result := StopWaiting;
            Exit;
          end;
      end;
  end;
  DebugMsgFmt(L_WARN, SGotUnknownEvent, [Event, Data.C]);
  Result := UNKNOWN_EVENT;
end;

function TVoiceModem.HangUp: Integer;
begin
  { Check driver specified }
  if not DriverAssigned('HangUp', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Hang up (goes on hook) }
  Result := Driver.HangUp;
end;

function TVoiceModem.Init: Integer;
var
  EchoCmd, EchoAnsw: string;
begin
  { Sel local echo }
  EchoCmd := Format('ATE%d', [Ord(Profile.CommandEcho)]);
  EchoAnsw := EchoCmd + '|OK';
  if WriteStr(EchoCmd) = FAIL then
  begin
    DebugMsg(L_WARN, SlocalEchoError);
    Result := FAIL;
    Exit;
  end;
  if Command('', EchoAnsw) and VMA_USER <> VMA_USER then
  begin
    Result := FAIL;
    Exit;
  end;
  Flush;
  { Set verbal result messages }
  if Command('ATQ0V1', 'OK') <> VMA_USER_1 then
  begin
    DebugMsg(L_WARN, SResultCodesError);
    Result := FAIL;
    Exit;
  end;
  { Common params init complete }
  Result := OK;
  { Driver specific initialization }
  if DriverAssigned('', False) then Result := Driver.Init;
  { Additional initialization command }
  if Length(Profile.ModemInitCmd) > 0 then
    if Command(Profile.ModemInitCmd, Profile.ModemInitAnsw) <> VMA_USER_1 then
      DebugMsg(L_WARN, SAdditionalCmdFailed);
end;

function TVoiceModem.InitVoice: Integer;
begin
  { Check driver specified }
  if not DriverAssigned('InitVoice', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Init voice params }
  Result := Driver.InitVoice;
end;

function TVoiceModem.Lock(Owner: TObject): Boolean;
begin
  EnterCriticalSection(FLockSection);
  { Check for modem already locked }
  Result := not FLocked;
  { Check lock owner }
  if Result then
  begin
    FLocked := True;
    FLockOwner := Owner;
  end;
  LeaveCriticalSection(FLockSection);
end;

function TVoiceModem.ModemName: string;
begin
  { Check driver specified }
  if not DriverAssigned('ModemName', True) then
  begin
    Result := '';
    Exit;
  end;
  { Get modem name }
  Result := Driver.ModemName;
end;

function TVoiceModem.ModemRMDName: string;
begin
  { Check driver specified }
  if not DriverAssigned('ModemRMDName', True) then
  begin
    Result := '';
    Exit;
  end;
  { Get modem name (for RMD file header) }
  Result := Driver.ModemRMDName;
end;

function TVoiceModem.Open: Integer;
begin
  { Check device already opened }
  if not Opened then
  begin
    { Open voice device }
    if OpenDevice = FAIL then
    begin
      Result := FAIL;
      Exit;
    end;
    { Init modem }
    if Init = FAIL then
    begin
      Result := FAIL;
      Exit;
    end;
    { Switch to voice mode }
    if VoiceModeOn = FAIL then
    begin
      Result := FAIL;
      Exit;
    end;
    { Init voice }
    if InitVoice = FAIL then
    begin
      Result := FAIL;
      Exit;
    end;
  end
  else
  begin
    { Send debug messages for full capability with OpenDevice method }
    DebugMsg(L_MESG, SOpeningDevice);
    DebugMsg(L_WARN, SDeviceAlreadyOpen);
  end;
  { Result }
  LastError := VME_SUCCESS;
  Result := OK;
end;

function TVoiceModem.OpenDevice: Integer;
begin
  DebugMsg(L_MESG, SOpeningDevice);
  { Check for device closed }
  if not Opened then
  begin
    { Open port }
    if Port.Open = FAIL then
    begin
      DebugMsg(L_ERROR, SOpenDeviceError);
      LastError := VME_PORT_ERROR;
      Result := FAIL;
      Exit;
    end;
    DebugMsg(L_MESG, SDeviceOpened);
  end
  else
    DebugMsg(L_WARN, SDeviceAlreadyOpen);
  { Result }
  LastError := VME_SUCCESS;
  Result := OK;
end;

function TVoiceModem.PlayDTMF(const Number: string): Integer;
begin
  DebugMsg(L_MESG, SPlayingDTMF);
  { Check driver specified }
  if not DriverAssigned('PlayDTMF', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Play DTMF tone }
  Result := Driver.PlayDTMF(Number);
end;

function TVoiceModem.PlayFile(const FileName: string): Integer;
var
  Handle: Integer;
  Stream: THandleStream;
begin
  { Check file }
  if not FileExists(FileName) then
  begin
    DebugMsgFmt(L_ERROR, SFileNotFound, [FileName]);
    LastError := VME_FILE_NOT_FOUND;
    Result := FAIL;
    Exit;
  end;
  { Open file }
  Handle := FileOpen(FileName, fmOpenRead);
  { Check for file opening }
  if Handle < 0 then
  begin
    DebugMsgFmt(L_ERROR, SFileOpenError, [FileName, GetLastError]);
    { Playing failed }
    LastError := VME_FILE_OPEN_ERROR;
    Result := FAIL;
    Exit;
  end;
  try
    DebugMsgFmt(L_MESG, SPlayingFile, [FileName]);
    { Playing file }
    Stream := THandleStream.Create(Handle);
    try
      Result := PlayStream(Stream);
    finally
      Stream.Free;
    end;
  finally
    FileClose(Handle);
  end;
end;

function TVoiceModem.PlayStream(Stream: TStream): Integer;
var
  Header: TRMDHeader;
  Compression, Speed, Bits: Integer;
begin
  { Check driver specified }
  if not DriverAssigned('PlayStream', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Get RMD header }
  if not Profile.RawData then
  begin
    { Read header }
    if Stream.Read(Header, SizeOf(TRMDHeader)) <> SizeOf(TRMDHeader) then
    begin
      DebugMsg(L_ERROR, SHeaderReadError);
      LastError := VME_STREAM_READ_ERR;
      Result := FAIL;
      Exit;
    end;
    { Check for header magic }
    if StrLComp(Header.Magic, RMD_HEADER_MAGIC, 4) <> 0 then
    begin
      DebugMsg(L_ERROR, SNoHeaderError);
      LastError := VME_BAD_STREAM_FORMAT;
      Result := FAIL;
      Exit;
    end;
    { Check for supported modem name }
    if StrLComp(Header.ModemType, PChar(ModemRMDName), Length(ModemRMDName)) <> 0 then
    begin
      DebugMsg(L_ERROR, SWrongModemError);
      LastError := VME_FORMAT_NOT_SUPPORTED;
      Result := FAIL;
      Exit;
    end;
    { Get data compression }
    Compression := Header.Compression;
    Speed := Header.Speed;
    Bits := Header.Bits;
  end
  else
  begin
    { Get default compression }
    Compression := Profile.RecCompression;
    Speed := Profile.RecSpeed;
    Bits := 8;
  end;
  { Set compression }
  if SetCompression(Compression, Speed, Bits) = FAIL then
  begin
    Result := FAIL;
    Exit;
  end;
  { Start playing }
  if Driver.StartPlayStream = FAIL then
  begin
    DebugMsg(L_ERROR, SStartPlayStreamFailed);
    Result := FAIL;
    Exit;
  end;
  { Play stream }
  Result := Driver.PlayStream(Stream, Speed * Bits);
  { Check playing result }
  if Result = FAIL then
  begin
    DebugMsg(L_ERROR, SPlayStreamFailed);
    Exit;
  end;
  { Stop playing }
  if Driver.StopPlayStream = FAIL then
  begin
    DebugMsg(L_ERROR, SStopPlayStreamFailed);
    Result := FAIL;
    Exit;
  end;
end;

function TVoiceModem.QueueEvent(Event: TModemEvent): Integer;
var
  EventNumber: Integer;
begin
  EventNumber := FLastEvent;
  Inc(FEventQueue[EventNumber].WriteLock);
  Inc(FEventCount);
  if FEventCount >= MAX_EVENTS then
  begin
    DebugMsg(L_ERROR, SEventQueueFull);
    Dec(FEventQueue[EventNumber].WriteLock);
    Dec(FEventCount);
    LastError := VME_EVENT_QUEUE_FULL;
    Result := FAIL;
    Exit;
  end;
  while FEventQueue[EventNumber].WriteLock > 1 do
  begin
    Dec(FEventQueue[EventNumber].WriteLock);
    EventNumber := (EventNumber + 1) mod MAX_EVENTS;
    Inc(FEventQueue[EventNumber].WriteLock);
  end;
  FLastEvent := (FLastEvent + 1) mod MAX_EVENTS;
  FEventQueue[EventNumber].Event := Event;
  Dec(FEventQueue[EventNumber].WriteLock);
  DebugMsgFmt(L_JUNK, SQueuedEvent, [Event.Event, EventNumber]);
  LastError := VME_SUCCESS;
  Result := OK;
end;

function TVoiceModem.Read(Buffer: Pointer; Count: Integer): Integer;
var
  TimeOut, StartTime: DWORD;
begin
  TimeOut := Profile.PortTimeOut * 1000;
  StartTime := GetTickCount;
  { Read data until success or timeout }
  while (TimeOut = 0) or (DWORD(Abs(GetTickCount - StartTime)) < TimeOut) do
  begin
    Result := Port.Read(Buffer, Count);
    { Check reading result }
    if Result < 0 then
    begin
      DebugMsg(L_ERROR, SBufferReadError);
      LastError := VME_PORT_ERROR;
      Exit;
    end
    else if Result > 0 then
    begin
      { Read event }
      if Assigned(FOnRead) then
      try
        FOnRead(Self, Buffer, Result);
      except
        on E: Exception do DebugMsg(L_ERROR, 'Read: ' + E.Message);
      end;
      { Yield callback }
      Yield;
      { Success }
      LastError := VME_SUCCESS;
      Exit;
    end
    else
      SleepEx(Profile.PortReadDelay, False);
    { Yield callback }
    Yield;
  end;
  { Reading timeout }
  DebugMsg(L_ERROR, SBufferReadTimeOut);
  LastError := VME_TIMEOUT;
  Result := -1;
end;

function TVoiceModem.ReadChar(var CharIn: Char): Integer;
begin
  if Read(@CharIn, 1) < 0 then
  begin
    Result := FAIL;
    Exit;
  end;
  { Trace modem input }
  TraceChar('Recv', CharIn);
  { Read char event }
  if Assigned(FOnReadChar) then
  try
    FOnReadChar(Self, CharIn);
  except
    on E: Exception do DebugMsg(L_ERROR, 'ReadChar: ' + E.Message);
  end;
  { Success }
  LastError := VME_SUCCESS;
  Result := OK;
  Exit;
end;

function TVoiceModem.ReadStr(var S: string): Integer;
var
  CharRead: Char;
  CharCount: Integer;
begin
  S := '';
  CharCount := 0;
  { Reading string }
  repeat
    { Read char }
    if ReadChar(CharRead) < 0 then
    begin
      Result := FAIL;
      Exit;
    end;
    { Check for <DLE> codes }
    if CharRead = DLE then
    begin
      { Read <DLE> code }
      if ReadChar(CharRead) = FAIL then
      begin
        Result := FAIL;
        Exit;
      end;
      DebugMsgFmt(L_JUNK, SReadDLE, [CharRead]);
      { Handle <DLE> code }
      HandleDLE(CharRead);
      Result := OK;
      Exit;
    end
    { Check for valid chars }
    else if not (CharRead in [LF, CR, XON, XOFF]) then
    begin
      S := S + CharRead;
      Inc(CharCount);
    end
    { Check for unexpected XON/XOFF codes }
    else if not (CharRead in [LF, CR]) then
      DebugMsgFmt(L_WARN, SUnexpectedByte, [CharRead]);
  until (CharRead = LF) and (CharCount > 0);
  { Read string event }
  if Assigned(FOnReadString) then
  try
    FOnReadString(Self, S);
  except
    on E: Exception do DebugMsg(L_ERROR, 'ReadString: ' + E.Message);
  end;
  { Success }
  DebugMsgFmt(L_JUNK, SReadString, [S]);
  Result := OK;
end;

function TVoiceModem.RecordFile(const FileName: string): Integer;
var
  Handle: Integer;
  Stream: THandleStream;
begin
  { Create output file }
  Handle := FileCreate(FileName);
  { Check file create error }
  if Handle < 0 then
  begin
    DebugMsgFmt(L_ERROR, SFileCreateError, [FileName, GetLastError]);
    { Recording failed }
    LastError := VME_FILE_CREATE_ERROR;
    Result := FAIL;
    Exit;
  end;
  try
    DebugMsgFmt(L_MESG, SRecordingFile, [FileName]);
    { Recording file }
    Stream := THandleStream.Create(Handle);
    try
      Result := RecordStream(Stream);
    finally
      Stream.Free;
    end;
  finally
    FileClose(Handle);
  end;
end;


function TVoiceModem.RecordStream(Stream: TStream): Integer;
var
  Header: TRMDHeader;
  Compression, Speed, Bits: Integer;
begin
  { Check driver specified }
  if not DriverAssigned('RecordStream', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Get default compression }
  Compression := Profile.RecCompression;
  Speed := Profile.RecSpeed;
  Bits := 8;
  { Set compression }
  if SetCompression(Compression, Speed, Bits) = FAIL then
  begin
    Result := FAIL;
    Exit;
  end;
  { Write RMD header }
  if not Profile.RawData then
  begin
    FillChar(Header,  SizeOf(TRMDHeader), 0);
    FormatBuf(Header.Magic, 4, '%s', 2, [RMD_HEADER_MAGIC]);
    FormatBuf(Header.ModemType, 16, '%s', 2, [ModemRMDName]);
    Header.Compression := Compression;
    Header.Speed := Speed;
    Header.Bits := Bits;
    if Stream.Write(Header, SizeOf(TRMDHeader)) <> SizeOf(TRMDHeader) then
    begin
      DebugMsg(L_ERROR, SHeaderWriteError);
      LastError := VME_STREAM_WRITE_ERR;
      Result := FAIL;
      Exit;
    end;
  end;
  { Start recording }
  if Driver.StartRecordStream = FAIL then
  begin
    DebugMsg(L_ERROR, SStartRecordStreamFailed);
    Result := FAIL;
    Exit;
  end;
  { Record stream }
  Result := Driver.RecordStream(Stream, Speed * Bits);
  { Check recording result }
  if Result = FAIL then
  begin
    DebugMsg(L_ERROR, SRecordStreamFailed);
    Exit;
  end;
  { Stop recording }
  if Driver.StopRecordStream = FAIL then
  begin
    DebugMsg(L_ERROR, SStopRecordStreamFailed);
    Result := FAIL;
    Exit;
  end;
end;

function TVoiceModem.RestoreVoiceMode: Integer;
begin
  DebugMsg(L_NOISE, SRestoringVoiceMode);
  { Check driver specified }
  if not DriverAssigned('RestoreVoiceMode', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Restoring voice mode }
  Result := Driver.RestoreVoiceMode;
end;

function TVoiceModem.SetCompression(var Compression, Speed, Bits: Integer): Integer;
begin
  DebugMsg(L_MESG, SSettingCompression);
  { Check driver specified }
  if not DriverAssigned('SetCompression', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Set compression mode }
  Result := Driver.SetCompression(Compression, Speed, Bits);
  { Check result }
  if Result = FAIL then
  begin
    DebugMsgFmt(L_ERROR, SCompressionError, [Compression, Speed]);
    Exit;
  end;
  { Success }
  DebugMsgFmt(L_MESG, SCompression, [Compression, Speed]);
end;

function TVoiceModem.SetDevice(Device: Integer): Integer;
begin
  DebugMsg(L_MESG, SSettingDevice);
  { Check driver specified }
  if not DriverAssigned('SetDevice', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Set device }
  Result := Driver.SetDevice(Device);
end;

function TVoiceModem.Stop: Integer;
begin
  Result := QueueEvent(TModemEvent.Create(SIGNAL_STOP));
end;

function TVoiceModem.StopCurrentAction: Integer;
begin
  case State of
    DIALING: Result := StopDialing;
    PLAYING: Result := StopPlaying;
    RECORDING: Result := StopRecording;
    WAITING: Result := StopWaiting;
    else
      begin
        LastError := VME_SUCCESS;
        Result := OK;
      end;
  end;
end;

function TVoiceModem.StopDialing: Integer;
begin
  { Check driver specified }
  if not DriverAssigned('StopDialing', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Stop dialing }
  Result := Driver.StopDialing;
end;

function TVoiceModem.StopPlaying: Integer;
begin
  { Check driver specified }
  if not DriverAssigned('StopPlaying', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Stop playing }
  Result := Driver.StopPlaying;
end;

function TVoiceModem.StopRecording: Integer;
begin
  { Check driver specified }
  if not DriverAssigned('StopRecording', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Stop recording }
  Result := Driver.StopRecording;
end;

function TVoiceModem.StopWaiting: Integer;
begin
  { Check driver specified }
  if not DriverAssigned('StopWaiting', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Stop waiting }
  Result := Driver.StopWaiting;
end;

function TVoiceModem.SwitchToDataFax(Mode: string): Integer;
begin
  DebugMsg(L_MESG, SSwitchingToDataFax);
  { Check driver specified }
  if not DriverAssigned('SwitchToDataFax', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Set data mode }
  Result := Driver.SwitchToDataFax(Mode);
end;

function TVoiceModem.UnqueueEvent(var Event: TModemEvent): Integer;
var
  EventNumber: Integer;
begin
  EventNumber := FFirstEvent;
  Inc(FEventQueue[EventNumber].ReadLock);
  Dec(FEventCount);
  if FEventCount < 0 then
  begin
    Dec(FEventQueue[EventNumber].ReadLock);
    Inc(FEventCount);
    Event := nil;
    LastError := VME_EVENT_QUEUE_EMPTY;
    Result := FAIL;
    Exit;
  end;
  while FEventQueue[EventNumber].ReadLock > 1 do
  begin
    Dec(FEventQueue[EventNumber].ReadLock);
    EventNumber := (EventNumber + 1) mod MAX_EVENTS;
    Inc(FEventQueue[EventNumber].ReadLock);
  end;
  if FEventQueue[EventNumber].WriteLock > 0 then
  begin
    Dec(FEventQueue[EventNumber].ReadLock);
    Inc(FEventCount);
    Event := nil;
    LastError := VME_EVENT_LOCKED;
    Result := FAIL;
    Exit;
  end;
  FFirstEvent := (FFirstEvent + 1) mod MAX_EVENTS;
  Dec(FEventQueue[EventNumber].ReadLock);
  Event := FEventQueue[EventNumber].Event;
  DebugMsgFmt(L_JUNK, SUnqueuedEvent, [Event.Event, EventNumber]);
  LastError := VME_SUCCESS;
  Result := OK;
end;

function TVoiceModem.Unlock: Boolean;
begin
  EnterCriticalSection(FLockSection);
  { Unlocking the modem }
  Result := FLocked;
  { Check result }
  if Result then
  begin
    FLocked := False;
    FLockOwner := nil;
  end;
  LeaveCriticalSection(FLockSection);
end;

function TVoiceModem.VoiceModeOff: Integer;
begin
  DebugMsg(L_NOISE, SLeavingVoiceMode);
  { Check driver specified }
  if not DriverAssigned('VoiceModeOff', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Leaving voice mode }
  Result := Driver.VoiceModeOff;
end;

function TVoiceModem.VoiceModeOn: Integer;
begin
  DebugMsg(L_NOISE, SEnteringVoiceMode);
  { Check driver specified }
  if not DriverAssigned('VoiceModeOn', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Entering voice mode }
  Result := Driver.VoiceModeOn;
end;

function TVoiceModem.Wait(TimeOut: DWORD): Integer;
begin
  { Check driver specified }
  if not DriverAssigned('Wait', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Wait event }
  Result := Driver.Wait(TimeOut);
end;

function TVoiceModem.WaitAnswer(const ExpectedAnswers: string; TimeOut: DWORD): Integer;
begin
  { Check driver specified }
  if not DriverAssigned('WaitAnswer', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Wait event }
  Result := Driver.WaitAnswer(ExpectedAnswers, TimeOut);
end;

function TVoiceModem.Write(Buffer: Pointer; Count: Integer): Integer;
var
  TimeOut, StartTime: DWORD;
begin
  TimeOut := Profile.PortTimeOut * 1000;
  StartTime := GetTickCount;
  { Write data until success or timeout }
  while ((TimeOut = 0) or (DWORD(Abs(GetTickCount - StartTime)) < TimeOut)) and (Count > 0) do
  begin
    Result := Port.Write(Buffer, Count);
    { Check result }
    if Result < 0 then
    begin
      DebugMsg(L_ERROR, SBufferWriteError);
      LastError := VME_PORT_ERROR;
      Result := FAIL;
      Exit;
    end
    else if Result > 0 then
    begin
      { Write data event }
      if Assigned(FOnWrite) then
      try
        FOnWrite(Self, Buffer, Result);
      except
        on E: Exception do DebugMsg(L_ERROR, 'Write: ' + E.Message);
      end;
      { Shift buffer pointer }
      Buffer := Pointer(Longint(Buffer) + Result);
      Count := Count - Result;
    end
    else
      SleepEx(Profile.PortWriteDelay, False);
    { Yield callback }
    Yield;
  end;
  { Check finally result }
  if Count = 0 then
  begin
    LastError := VME_SUCCESS;
    Result := OK;
    Exit;
  end;
  { Writing timeout }
  DebugMsg(L_ERROR, SBufferWriteTimeOut);
  LastError := VME_TIMEOUT;
  Result := FAIL;
end;

function TVoiceModem.WriteChar(CharOut: Char): Integer;
begin
  { Writing char }
  if Write(@CharOut, 1) = FAIL then
  begin
    Result := FAIL;
    Exit;
  end;
  { Trace modem output }
  TraceChar('Send', CharOut);
  { Write char event }
  if Assigned(FOnWriteChar) then
  try
    FOnWriteChar(Self, CharOut);
  except
    on E: Exception do DebugMsg(L_ERROR, 'WriteChar: ' + E.Message);
  end;
  { Success }
  Result := OK;
end;

function TVoiceModem.WriteStr(const S: string): Integer;
var
  I: Integer;
begin
  DebugMsgFmt(L_JUNK, SWriteString, [S]);
  { Writing string }
  if Write(PChar(S), Length(S)) = FAIL then
  begin
    Result := FAIL;
    Exit;
  end;
  { Trace output }
  for I := 1 to Length(S) do TraceChar('Send', S[I]);
  { Write string event }
  if Assigned(FOnWriteString) then
  try
    FOnWriteString(Self, S);
  except
    on E: Exception do DebugMsg(L_ERROR, 'WriteStr: ' + E.Message);
  end;
  { Writing line break }
  if WriteChar(CR) = FAIL then
  begin
    Result := FAIL;
    Exit;
  end;
  { Success }
  Result := OK;
end;

function TVoiceModem.WriteStrFmt(const Format: string; const Args: array of const): Integer;
begin
  try
    Result := WriteStr(SysUtils.Format(Format, Args));
  except
    DebugMsg(L_ERROR, SInvalidStrFormat);
    LastError := VME_INVALID_STR_FORMAT;
    Result := FAIL;
  end;
end;

initialization
  FDriversList := nil;

finalization
  FDriversList.Free;
  FDriversList := nil;

end.

