{
  Voice Modem Library v1.0

  This file contains the Motorola SM56 specific hardware stuff.

  Roman M. Mochalov
  E-mail: roman@sar.nnov.ru
}

unit VML_MotorolaSM56;

interface

uses
  Windows, SysUtils, Classes, VML_Modem, VML_Consts, VML_IS101, VML_Idents;

const

{ Motorola SM56 available compression modes }

  COMPRESSION_MOTOROLASM56_PCM    = 1;
  COMPRESSION_MOTOROLASM56_ADPCM  = 2;

type

{ TMotorolaSM56 }

  TMotorolaSM56 = class(TIS101)
  public
    constructor Create(AOwner: TComponent); override;
    class function GetCompressionMap(var Map: PIdentMap): Integer; override;
    class function GetDeviceMap(var Map: PIdentMap): Integer; override;
    function Init: Integer; override;
    function InitVoice: Integer; override;
    class function ModemName: string; override;
    class function ModemRMDName: string; override;
    function SetCompression(var Compression, Speed, Bits: Integer): Integer; override;
    function SetDevice(Device: Integer): Integer; override;
  end;

implementation

{$I *.INC}

{ Some internal constants }

const
  CompressionMap: array[0..1] of TIdentMapEntry = (
    (Value: COMPRESSION_MOTOROLASM56_PCM; Name: SPCM),
    (Value: COMPRESSION_MOTOROLASM56_ADPCM; Name: SADPCM));

  DeviceMap: array[0..4] of TIdentMapEntry = (
    (Value: DIALUP_LINE; Name: SDialupLine),
    (Value: EXTERNAL_MICROPHONE; Name: SExternalMicrophone),
    (Value: EXTERNAL_SPEAKER; Name: SExternalSpeaker),
    (Value: INTERNAL_SPEAKER; Name: SInternalSpeaker),
    (Value: SPEAKER_PHONE; Name: SSpeakerphone));

{ TMotorolaSM56 }

constructor TMotorolaSM56.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  PickPhoneCmd := 'ATA';
end;

class function TMotorolaSM56.GetCompressionMap(var Map: PIdentMap): Integer;
begin
  Map := @CompressionMap;
  Result := High(CompressionMap) - Low(CompressionMap) + 1;
end;

class function TMotorolaSM56.GetDeviceMap(var Map: PIdentMap): Integer;
begin
  Map := @DeviceMap;
  Result := High(DeviceMap) - Low(DeviceMap) + 1;
end;

function TMotorolaSM56.Init: Integer;
begin
  DebugMsgFmt(L_MESG, SInitModem, [ModemName]);
  { Check modem specified }
  if not ModemAssigned('Init', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Update modem state }
  Modem.State := INITIALIZING;
  try
    { AT&R1 - Modem assumes CTS always ON
      AT\Q3 - Hardware (RTS/CTS) flow control }
    if Modem.Command('AT&R1\Q3', 'OK') <> VMA_USER_1 then
    begin
      DebugMsg(L_WARN, SFlowControlError);
      Result := FAIL;
      Exit;
    end;
    { Set extended verbal result messages }
    if Modem.Command('ATX4', 'OK') <> VMA_USER_1 then
    begin
      DebugMsg(L_WARN, SExResultCodesError);
      Result := FAIL;
      Exit;
    end;
  finally
    Modem.State := IDLE;
  end;
  { Init complete }
  Result := OK;
end;

function TMotorolaSM56.InitVoice: Integer;
var
  TransmitGain, ReceiveGain: Integer;
begin
  DebugMsgFmt(L_MESG, SInitVoice, [ModemName]);
  { Check modem specified }
  if not ModemAssigned('InitVoice', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Update modem state }
  Modem.State := INITIALIZING;
  try
    { AT+VNH=1 - Disable autohangup }
    if Modem.Command('AT+VNH=1', 'OK') <> VMA_USER_1 then
      DebugMsg(L_WARN, SAutoHangUpError);
    { AT+VSD=x,y - Set silence threshold and duration }
    if Modem.CommandFmt('AT+VSD=%d,%d', 'OK', [Profile.RecSilenceSens * 129 div 100, Profile.RecSilenceLen]) <> VMA_USER_1 then
      DebugMsg(L_WARN, SSilensePreferencesError);
    { AT+VGT - Set the transmit volume for voice samples. }
    TransmitGain := Profile.TransmitGain;
    if TransmitGain = -1 then TransmitGain := 50;
    if Modem.CommandFmt('AT+VGT=%d', 'OK', [TransmitGain * 255 div 100]) <> VMA_USER_1 then
      DebugMsg(L_WARN, STransmitGainError);
    { AT+VGR - Set receive gain for voice samples. }
    ReceiveGain := Profile.ReceiveGain;
    if ReceiveGain = -1 then ReceiveGain := 50;
    if Modem.CommandFmt('AT+VGR=%d', 'OK', [ReceiveGain * 255 div 100]) <> VMA_USER_1 then
      DebugMsg(L_WARN, SReceiveGainError);
    { Delay after ringback or before any ringback before modem assumes
      phone has been answered }
    if Modem.CommandFmt('AT+VRA=%d+VRN=%d', 'OK', [Profile.RingbackGoesAway, Profile.RingbackNeverCame]) <> VMA_USER_1 then
      DebugMsg(L_ERROR, SRingbackTimersError);
  finally
    Modem.State := IDLE;
  end;
  { Success }
  Result := OK;
end;

class function TMotorolaSM56.ModemName: string;
begin
  Result := 'Motorola SM56';
end;

class function TMotorolaSM56.ModemRMDName: string;
begin
  Result := 'Motorola SM56';
end;

function TMotorolaSM56.SetCompression(var Compression, Speed, Bits: Integer): Integer;
begin
  { Check modem specified }
  if not ModemAssigned('SetCompression', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { The default mode for Motorola modems is ADPCM }
  if Compression = 0 then Compression := COMPRESSION_MOTOROLASM56_ADPCM;
  if (Compression <> COMPRESSION_MOTOROLASM56_PCM) and (Compression <> COMPRESSION_MOTOROLASM56_ADPCM) then
  begin
    DebugMsgFmt(L_WARN, SIllegalCompressionMethod, [Compression]);
    Modem.LastError := VME_ILLEGAL_COMPRESSION;
    Result := FAIL;
    Exit;
  end;
  { 8000 Hz is currently the only recording freq supported by Motorola SM56 }
  if Speed = 0 then Speed := 8000;
  if Speed <> 8000 then
  begin
    DebugMsgFmt(L_WARN, SIllegalSampleRate, [Speed]);
    Result := FAIL;
    Exit;
  end;
  { Setup compression }
  Bits := 8;
  if Modem.CommandFmt('AT+VSM=%d,%d,0,0', 'OK', [Compression + 127, Speed]) <> VMA_USER_1 then
  begin
    Result := FAIL;
    Exit;
  end;
  { Success }
  Result := OK;
  Exit;
end;

function TMotorolaSM56.SetDevice(Device: Integer): Integer;
begin
  { Check modem specified }
  if not ModemAssigned('SetDevice', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Set device }
  case Device of
    NO_DEVICE:
      begin
        Modem.Command('AT+VLS=0', 'OK');
        Result := OK;
        Exit;
      end;
    DIALUP_LINE:
      begin
        if Modem.Command('AT+VLS=1', 'OK|VCON') = VMA_USER_2 then
        begin
          Result := VMA_CONNECT;
          Exit;
        end;
        Result := OK;
        Exit;
      end;
    EXTERNAL_MICROPHONE:
      begin
        Modem.Command('AT+VLS=11', 'OK|VCON');
        Result := OK;
        Exit;
      end;
    EXTERNAL_SPEAKER:
      begin
        Modem.Command('AT+VLS=8', 'OK|VCON');
        Result := OK;
        Exit;
      end;
    INTERNAL_SPEAKER:
      begin
        Modem.Command('AT+VLS=9', 'OK|VCON');
        Result := OK;
        Exit;
      end;
    SPEAKER_PHONE:
      begin
        Modem.Command('AT+VLS=13', 'OK|VCON');
        Result := OK;
        Exit;
      end;
  end;
  { Given device unknown }
  DebugMsgFmt(L_WARN, SUnknownDevice, [Device]);
  Modem.LastError := VME_UNKNOWN_DEVICE;
  Result := FAIL;
end;

initialization
  RegisterModemDriver(TMotorolaSM56.ModemName, TMotorolaSM56);

end.
