{
  Voice Modem Library v1.0


  This file contains the ROCKWELL's Voice/Fax/Data chipset based hardware
  stuff.

  Roman M. Mochalov
  E-mail: roman@sar.nnov.ru
}

unit VML_Rockwell;

interface

uses
  Windows, SysUtils, Classes, VML_Modem, VML_Consts, VML_IS101, VML_Idents;

const

{ Rockwell available compression modes }

  COMPRESSION_ROCKWELL_ADPCM_2 = 2;
  COMPRESSION_ROCKWELL_ADPCM_4 = 4;
  COMPRESSION_ROCKWELL_ADPCM_8 = 8;

type

{ TRockwell }

  TRockwell = class(TIS101)
  private
    FCurrentDevice: Integer;
  public
    constructor Create(AOwner: TComponent); override;
    class function GetCompressionMap(var Map: PIdentMap): Integer; override;
    class function GetDeviceMap(var Map: PIdentMap): Integer; override;
    function HandleDLE(Code: Char): Integer; override;
    function Init: Integer; override;
    function InitVoice: Integer; override;
    class function ModemName: string; override;
    class function ModemRMDName: string; override;
    function RestoreVoiceMode: Integer; override;
    function SetCompression(var Compression, Speed, Bits: Integer): Integer; override;
    function SetDevice(Device: Integer): Integer; override;
  end;

implementation

{$I *.INC}

{ Some internal constants }

const
  CompressionMap: array[0..2] of TIdentMapEntry = (
    (Value: COMPRESSION_ROCKWELL_ADPCM_2; Name: SADPCM2),
    (Value: COMPRESSION_ROCKWELL_ADPCM_4; Name: SADPCM4),
    (Value: COMPRESSION_ROCKWELL_ADPCM_8; Name: SADPCM8));

  DeviceMap: array[0..4] of TIdentMapEntry = (
    (Value: DIALUP_LINE; Name: SDialupLine),
    (Value: EXTERNAL_MICROPHONE; Name: SExternalMicrophone),
    (Value: INTERNAL_SPEAKER; Name: SInternalSpeaker),
    (Value: LOCAL_HANDSET; Name: SLocalHandset),
    (Value: SPEAKER_PHONE; Name: SSpeakerPhone));

{ TRockwell }

constructor TRockwell.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FCurrentDevice := -1;
  PickPhoneCmd := 'AT#VLS=0A';
  PickPhoneAnsw := 'VCON';
  BeepCmd := 'AT#VTS=[%d,0,%d]';
  BeepTimeUnit := 100;
  HardFlowCmd := 'AT&K3';
  SoftFlowCmd := 'AT&K4';
  StartPlayCmd := 'AT#VTX';
  IntrPlayAnsw := 'OK|VCON';
  StopPlayAnsw := 'OK|VCON';
  StartRecCmd := 'AT#VRX';
  StopRecCmd := '!';
  StopRecAnsw := 'OK|VCON';
  SwitchModeCmd := 'AT#CLS=%s';
  VoiceModeOnCmd := 'AT#CLS=8';
  VoiceModeOffCmd := 'AT#CLS=0';
  PlayDTMFCmd := 'AT#VTS=%s';
end;

class function TRockwell.GetCompressionMap(var Map: PIdentMap): Integer;
begin
  Map := @CompressionMap;
  Result := High(CompressionMap) - Low(CompressionMap) + 1;
end;

class function TRockwell.GetDeviceMap(var Map: PIdentMap): Integer;
begin
  Map := @DeviceMap;
  Result := High(DeviceMap) - Low(DeviceMap) + 1;
end;

function TRockwell.HandleDLE(Code: Char): Integer;
begin
  { Check modem specified }
  if not ModemAssigned('HandleDLE', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Handle DLE codes }
  case Code of
    't':
      { Local handset goes off hook }
      begin
        Result := Modem.QueueEvent(TModemEvent.Create(HANDSET_OFF_HOOK));
        Exit;
      end;
  end;
  { Standard handling }
  Result := inherited HandleDLE(Code);
end;

function TRockwell.Init: Integer;
begin
  DebugMsgFmt(L_MESG, SInitModem, [ModemName]);
  { Check modem specified }
  if not ModemAssigned('Init', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Update modem state }
  Modem.State := INITIALIZING;
  try
    { AT&K3 Enables RTS/CTS flow control }
    if Modem.Command('AT&K3', 'OK') <> VMA_USER_1 then
    begin
      DebugMsg(L_WARN, SFlowControlError);
      Result := FAIL;
      Exit;
    end;
  finally
    Modem.State := IDLE;
  end;
  { Init complete }
  Result := OK;
end;

function TRockwell.InitVoice: Integer;
begin
  DebugMsgFmt(L_MESG, SInitVoice, [ModemName]);
  { Check modem specified }
  if not ModemAssigned('InitVoice', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Update modem state }
  Modem.State := INITIALIZING;
  try
    { Set silence detection period }
    if Modem.CommandFmt('AT#VSP=%d', 'OK', [Profile.RecSilenceLen]) <> VMA_USER_1 then
      DebugMsg(L_WARN, SSilencePeriodError);
    { Set silence detection sensitive }
    if Modem.CommandFmt('AT#VSS=%d', 'OK', [Profile.RecSilenceSens * 3 div 100]) <> VMA_USER_1 then
      DebugMsg(L_WARN, SSilenceSensitivityError);
    { Set silence deletion }
    if Modem.CommandFmt('AT#VSD=%d', 'OK', [Ord(Profile.RecRemoveSilence)]) <> VMA_USER_1 then
      DebugMsg(L_WARN, SSilenceDetectionError);
    { Delay after ringback or before any ringback before modem assumes
      phone has been answered }
    if Modem.CommandFmt('AT#VRA=%d#VRN=%d', 'OK', [Profile.RingbackGoesAway, Profile.RingbackNeverCame]) <> VMA_USER_1 then
      DebugMsg(L_ERROR, SRingbackTimersError);
    { Select required modem/DTE baud rate }
    if Modem.CommandFmt('AT#BDR=%d', 'OK', [Profile.PortBaudRate div 2400]) <> VMA_USER_1 then
      DebugMsg(L_ERROR, SBaudRateError);
    { Insure that the modem can recover even id DTE dies with the
      baud rate set to a speed which cannot be speed sensed }
    if Modem.Command('ATS30=60', 'OK') <> VMA_USER_1 then
      DebugMsg(L_ERROR, SDeadmanTimerError);
  finally
    Modem.State := IDLE;
  end;
  { Success }
  Result := OK;
end;

class function TRockwell.ModemName: string;
begin
  Result := 'Rockwell';
end;

class function TRockwell.ModemRMDName: string;
begin
  Result := 'Rockwell';
end;

function TRockwell.RestoreVoiceMode: Integer;
begin
  { Switch to voice mode }
  if inherited RestoreVoiceMode = FAIL then
  begin
    Result := FAIL;
    Exit;
  end;
  { Restore required modem/DTE baud rate and DCE/DTE inactivity timeout }
  if Modem.CommandFmt('AT#BDR=%d', 'OK', [Profile.PortBaudRate div 2400]) <> VMA_USER_1 then
    DebugMsg(L_ERROR, SBaudRateError);
  { Restore deadman (inactivity) Timer }
  if Modem.Command('ATS30=60', 'OK') <> VMA_USER_1 then
    DebugMsg(L_ERROR, SDeadmanTimerError);
  { Success }
  Result := OK;
end;

function TRockwell.SetCompression(var Compression, Speed, Bits: Integer): Integer;
begin
  { Check modem specified }
  if not ModemAssigned('SetCompression', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { The default mode for Rockwell modems is ADPCM 2 bits per sample }
  if Compression = 0 then Compression := COMPRESSION_ROCKWELL_ADPCM_2;
  { 7200 Hz is currently the only recording freq supported by Rockwell }
  if Speed = 0 then Speed := 7200;
  if Speed <> 7200 then
  begin
    DebugMsgFmt(L_WARN, SIllegalSampleRate, [Speed]);
    Result := FAIL;
    Exit;
  end;
  { Setup sampling rate }
  if Modem.Command('AT#VSR=7200', 'OK') <> VMA_USER_1 then
  begin
    Result := FAIL;
    Exit;
  end;
  { Setup compression }
  case Compression of
    COMPRESSION_ROCKWELL_ADPCM_2:
      { 2 bit ADPCM }
      begin
        Bits := 2;
        if Modem.Command('AT#VBS=2', 'OK') <> VMA_USER_1 then
        begin
          Result := FAIL;
          Exit;
        end;
      end;
    COMPRESSION_ROCKWELL_ADPCM_4:
      { 4 bit ADPCM }
      begin
        Bits := 4;
        if Modem.Command('AT#VBS=4', 'OK') <> VMA_USER_1 then
        begin
          Result := FAIL;
          Exit;
        end;
      end;
    COMPRESSION_ROCKWELL_ADPCM_8:
      { 8 bit ADPCM }
      begin
        Bits := 8;
        if Modem.Command('AT#VBS=8', 'OK') <> VMA_USER_1 then
        begin
          Result := FAIL;
          Exit;
        end;
      end;
    else
      begin
        DebugMsgFmt(L_WARN, SIllegalCompressionMethod, [Compression]);
        Modem.LastError := VME_ILLEGAL_COMPRESSION;
        Result := FAIL;
        Exit;
      end;
  end;
  Result := OK;
  Exit;
end;

function TRockwell.SetDevice(Device: Integer): Integer;
begin
  { Check modem specified }
  if not ModemAssigned('SetDevice', True) then
  begin
    Result := FAIL;
    Exit;
  end;
(*
  { Check previous device }
  if FCurrentDevice <> Device then
  begin
    if FCurrentDevice >= 0 then Modem.Command('ATH0', 'VCON|OK');
    FCurrentDevice := Device;
  end;
*)
  { Set device }
  case Device of
    NO_DEVICE:
      begin
        Modem.Command('AT#VLS=0', 'OK');
        Result := OK;
        Exit;
      end;
    DIALUP_LINE:
      begin
        Modem.Command('AT#VLS=4', 'OK');
        Result := OK;
        Exit;
      end;
    EXTERNAL_MICROPHONE:
      begin
        Modem.Command('AT#VLS=3', 'VCON');
        Result := OK;
        Exit;
      end;
    INTERNAL_SPEAKER:
      begin
        Modem.Command('AT#VLS=2', 'VCON');
        Result := OK;
        Exit;
      end;
    LOCAL_HANDSET:
      begin
        Modem.Command('AT#VLS=1', 'VCON');
        Result := OK;
        Exit;
      end;
    SPEAKER_PHONE:
      begin
        Modem.Command('AT#VLS=6', 'OK|VCON');
        Result := OK;
        Exit;
      end;
  end;
  { Given device unknown }
  DebugMsgFmt(L_WARN, SUnknownDevice, [Device]);
  Modem.LastError := VME_UNKNOWN_DEVICE;
  Result := FAIL;
end;

initialization
  RegisterModemDriver(TRockwell.ModemName, TRockwell);

end.
