{
  Voice Modem Library v1.0

  This file contains hardware driver functions for
  some USRobotics modems.

  Based on US_Robotics driver, originally from
  Steven Wormley <wormley@step.mother.com>

  Known bugs:
    - My Some USRobotics Sportster Flash doesn't catch dtmf
      codes after a beep. If I send it the "AT#VLS=4A" (answer
      phone) string after an "AT#VTS[%d,0,%d]" (beep) command
      it would resume detecting tones.
    
  Roman M. Mochalov
  E-mail: roman@sar.nnov.ru
}

unit VML_USRobotics;

interface

uses
  Windows, SysUtils, Classes, VML_Modem, VML_Consts, VML_IS101, VML_Idents;

const

{ US Robotics available compression modes }

  COMPRESSION_USROBOTICS_GSM     = 1;
  COMPRESSION_USROBOTICS_ADPCM_2 = 2;
  COMPRESSION_USROBOTICS_ADPCM_3 = 3;
  COMPRESSION_USROBOTICS_ADPCM_4 = 4;

type

{ TUSRobotics }

  TUSRobotics = class(TIS101)
  public
    constructor Create(AOwner: TComponent); override;
    class function GetCompressionMap(var Map: PIdentMap): Integer; override;
    class function GetDeviceMap(var Map: PIdentMap): Integer; override;
    class function GetVoiceLen(Compression, Speed, Bits: Integer; DataSize: Longint): Integer; override;
    function HandleDLE(Code: Char): Integer; override;
    function Init: Integer; override;
    function InitVoice: Integer; override;
    class function ModemName: string; override;
    class function ModemRMDName: string; override;
    function RestoreVoiceMode: Integer; override;
    function SetCompression(var Compression, Speed, Bits: Integer): Integer; override;
    function SetDevice(Device: Integer): Integer; override;
    function StopPlayStream: Integer; override;
  end;

implementation

{$I *.INC}

{ Some internal constants }

const
  CompressionMap: array[0..3] of TIdentMapEntry = (
    (Value: COMPRESSION_USROBOTICS_GSM; Name: SGSM),
    (Value: COMPRESSION_USROBOTICS_ADPCM_2; Name: SADPCM2),
    (Value: COMPRESSION_USROBOTICS_ADPCM_3; Name: SADPCM3),
    (Value: COMPRESSION_USROBOTICS_ADPCM_4; Name: SADPCM4));

  DeviceMap: array[0..5] of TIdentMapEntry = (
    (Value: DIALUP_LINE; Name: SDialupLine),
    (Value: INTERNAL_MICROPHONE; Name: SInternalMicrophone),
    (Value: EXTERNAL_SPEAKER; Name: SExternalSpeaker),
    (Value: INTERNAL_SPEAKER; Name: SInternalSpeaker),
    (Value: LOCAL_HANDSET; Name: SLocalHandset),
    (Value: SPEAKER_PHONE; Name: SSpeakerphone));

{ TUSRobotics }

constructor TUSRobotics.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  PickPhoneCmd := 'AT#VLS=0A';
  PickPhoneAnsw := 'VCON';
  BeepCmd := 'AT#VTS=[%d,0,%d]';
  BeepTimeUnit := 100;
  HardFlowCmd := 'AT';
  SoftFlowCmd := 'AT';
  StartPlayCmd := 'ATE0#VTX';
  IntrPlayCmd := DLE + CAN;
  IntrPlayAnsw := 'OK|VCON';
  StopPlayAnsw := 'OK|VCON';
  StartRecCmd := 'AT#VRX';
  StopRecCmd := DLE;
  StopRecAnsw := 'OK|VCON';
  SwitchModeCmd := 'AT#CLS=%s';
  VoiceModeOnCmd := 'AT#CLS=8';
  VoiceModeOffCmd := 'AT#CLS=0';
  PlayDTMFCmd := 'AT#VTS=%s';
end;

class function TUSRobotics.GetCompressionMap(var Map: PIdentMap): Integer;
begin
  Map := @CompressionMap;
  Result := High(CompressionMap) - Low(CompressionMap) + 1;
end;

class function TUSRobotics.GetDeviceMap(var Map: PIdentMap): Integer;
begin
  Map := @DeviceMap;
  Result := High(DeviceMap) - Low(DeviceMap) + 1;
end;

class function TUSRobotics.GetVoiceLen(Compression, Speed, Bits: Integer; DataSize: Longint): Integer;
var
  BPS: Integer;
begin
  if Compression = COMPRESSION_USROBOTICS_GSM then
  begin
    { This formula correct only for encapsulated GSM frame format }
    BPS := Speed * 38 div 160;
    { Check BPS value and return duration of voice message }
    if BPS > 0 then
    begin
      Result := Round(DataSize / BPS * 1000);
      Exit;
    end;
  end;
  { Use default calculation from IS101 }
  Result := inherited GetVoiceLen(Compression, Speed, Bits, DataSize);
end;

function TUSRobotics.HandleDLE(Code: Char): Integer;
begin
  { Check modem specified }
  if not ModemAssigned('HandleDLE', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Handle DLE codes }
  case Code of
    't':
      { My USRobotics receive <DLE>t event when local handset goes off
        hook but no <DLE>H or <DLE>P. }
      begin
        Result := Modem.QueueEvent(TModemEvent.Create(HANDSET_OFF_HOOK));
        Exit;
      end;
     'd', 'i':
      { In this situation IS_101_handle_dle will create a DAIL_TONE event.
        This DLE is however incorrectly generated by the USRobotics modems
        and should be ignored when we are either PLAYING or RECORDING }
      case Modem.State of
        PLAYING:
          begin
            DebugMsgFmt(L_JUNK, SIgnoreDLE, [Code, 'PLAYING']);
            Result := OK;
            Exit;
          end;
        RECORDING:
          begin
            DebugMsgFmt(L_JUNK, SIgnoreDLE, [Code, 'RECORDING']);
            Result := OK;
            Exit;
          end;
      end;
  end;
  { Standard handling }
  Result := inherited HandleDLE(Code);
end;

function TUSRobotics.Init: Integer;
begin
  DebugMsgFmt(L_MESG, SInitModem, [ModemName]);
  { Check modem specified }
  if not ModemAssigned('Init', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Update modem state }
  Modem.State := INITIALIZING;
  try
    { AT&H1 - Transmit Data (TD) flow control: Clear to Send (CTS).
      AT&R2 - Recive Data flow control: only on Request To Send (RTS).
      AT&I0 - Disable software flow control. }
    if Modem.Command('AT&H1&R2&I0', 'OK') <> VMA_USER_1 then
    begin
      DebugMsg(L_WARN, SFlowControlError);
      Result := FAIL;
      Exit;
    end;
    { Set extended verbal result messages }
    if Modem.Command('ATX4', 'OK') <> VMA_USER_1 then
    begin
      DebugMsg(L_WARN, SExResultCodesError);
      Result := FAIL;
      Exit;
    end;
  finally
    Modem.State := IDLE;
  end;
  { Init complete }
  Result := OK;
end;

function TUSRobotics.InitVoice: Integer;
var
  TransmitGain, ReceiveGain: Integer;
begin
  DebugMsgFmt(L_MESG, SInitVoice, [ModemName]);
  { Check modem specified }
  if not ModemAssigned('InitVoice', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Update modem state }
  Modem.State := INITIALIZING;
  try
    { Setting DLE responses }
    if Modem.Command('AT#VTD=3F,3F,3F', 'OK') <> VMA_USER_1 then
      DebugMsg(L_WARN, SDLEResposesError)
    else
      DebugMsg(L_WARN, SDLEResposesSuccess);
    { Set silence detection period }
    if Modem.CommandFmt('AT#VSP=%d', 'OK', [Profile.RecSilenceLen]) <> VMA_USER_1 then
      DebugMsg(L_WARN, SSilencePeriodError);
    { Set silence detection sensitive }
    if Modem.CommandFmt('AT#VSS=%d', 'OK', [(Profile.RecSilenceSens * 3) div 100]) <> VMA_USER_1 then
      DebugMsg(L_WARN, SSilenceSensitivityError);
    { Set silence deletion }
    if Modem.CommandFmt('AT#VSD=%d', 'OK', [Ord(Profile.RecRemoveSilence)]) <> VMA_USER_1 then
      DebugMsg(L_WARN, SSilenceDetectionError);
    { Set the sensitivity of the microphone in speaker phone mode }
    TransmitGain := Profile.TransmitGain;
    if TransmitGain = -1 then TransmitGain := 75;
    if Modem.CommandFmt('AT#VGT=%d', 'OK', [TransmitGain * 255 div 100]) <> VMA_USER_1 then
      DebugMsg(L_WARN, STransmitGainError);
    { Set the speaker volume in speaker phone mode }
    ReceiveGain := Profile.ReceiveGain;
    if ReceiveGain = -1 then ReceiveGain := 75;
    if Modem.CommandFmt('AT#VGR=%d', 'OK', [ReceiveGain * 255 div 100]) <> VMA_USER_1 then
      DebugMsg(L_WARN, SReceiveGainError);
    { Delay after ringback or before any ringback before modem assumes
      phone has been answered }
    if Modem.CommandFmt('AT#VRA=%d#VRN=%d', 'OK', [Profile.RingbackGoesAway, Profile.RingbackNeverCame]) <> VMA_USER_1 then
      DebugMsg(L_ERROR, SRingbackTimersError);
    { Select required modem/DTE baud rate }
    if Modem.CommandFmt('AT#BDR=%d', 'OK', [Profile.PortBaudRate div 2400]) <> VMA_USER_1 then
      DebugMsg(L_ERROR, SBaudRateError);
    { Insure that the modem can recover even id DTE dies with the
      baud rate set to a speed which cannot be speed sensed }
    if Modem.Command('ATS30=60', 'OK') <> VMA_USER_1 then
      DebugMsg(L_ERROR, SDeadmanTimerError);
  finally
    Modem.State := IDLE;
  end;
  { Success }
  Result := OK;
end;

class function TUSRobotics.ModemName: string;
begin
  Result := 'US Robotics';
end;

class function TUSRobotics.ModemRMDName: string;
begin
  Result := 'US Robotics';
end;

function TUSRobotics.RestoreVoiceMode: Integer;
begin
  { Switch to the voice mode }
  if inherited RestoreVoiceMode = FAIL then
  begin
    Result := FAIL;
    Exit;
  end;
  { Restore required modem/DTE baud rate }
  if Modem.CommandFmt('AT#BDR=%d', 'OK', [Profile.PortBaudRate div 2400]) <> VMA_USER_1 then
    DebugMsg(L_ERROR, SBaudRateError);
  { Restore deadman (inactivity) Timer }
  if Modem.Command('ATS30=60', 'OK') <> VMA_USER_1 then
    DebugMsg(L_ERROR, SDeadmanTimerError);
  { Success }
  Result := OK;
end;

function TUSRobotics.SetCompression(var Compression, Speed, Bits: Integer): Integer;
begin
  { Check modem specified }
  if not ModemAssigned('SetCompression', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { 8000 Hz is currently the only recording freq supported by
    Sportster Vi and Voice }
  if Speed = 0 then Speed := 8000;
  if Speed <> 8000 then
  begin
    DebugMsgFmt(L_WARN, SIllegalSampleRate, [Speed]);
    Result := FAIL;
    Exit;
  end;
  { The default mode for modems not supporting ADPCM is GSM. Note: some of
    these modems seems to have broken GSM playback. The default mode for these
    modems is ADPCM 2 bits/sample. The mode may be configured in Profile. }
  if Compression = 0 then Compression := COMPRESSION_USROBOTICS_GSM;
  { Setup compression }
  case Compression of
    COMPRESSION_USROBOTICS_GSM:
      { This is the GSM mode 8 bits per sample }
      begin
        Bits := 8;
        { Set the mode to GSM }
        if Modem.Command('AT#VSM=128,8000', 'OK') = VMA_FAIL then
        begin
          Result := VMA_FAIL;
          Exit;
        end;
      end;
    COMPRESSION_USROBOTICS_ADPCM_2,
      { This is the ADPCM 2 bits per sample mode }
    COMPRESSION_USROBOTICS_ADPCM_3,
      { This is the ADPCM 3 bits per sample mode }
    COMPRESSION_USROBOTICS_ADPCM_4:
      { This is the ADPCM 4 bits per sample mode }
      begin
        Bits := Compression;
        { Set the mode to ADPCM }
        if Modem.Command('AT#VSM=129,8000', 'OK') = VMA_FAIL then
        begin
          Result := VMA_FAIL;
          Exit;
        end;
        { Note: If the modem does not support ADPCM, the following
          command still gets issued. It has no effect in GSM mode, but
          is included for backwards compatibility with the old driver. }
        if Modem.CommandFmt('AT#VBS=%d', 'OK', [Compression]) = VMA_FAIL then
        begin
          Result := VMA_FAIL;
          Exit;
        end;
      end;
    else
      begin
        DebugMsgFmt(L_WARN, SIllegalCompressionMethod, [Compression]);
        Modem.LastError := VME_ILLEGAL_COMPRESSION;
        Result := FAIL;
        Exit;
      end;
  end;
  Result := OK;
  Exit;
end;

function TUSRobotics.SetDevice(Device: Integer): Integer;
begin
  { Check modem specified }
  if not ModemAssigned('SetDevice', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Set device }
  case Device of
    NO_DEVICE:
      begin
        Modem.Command('AT#VLS=0H0', 'OK|VCON');
        Result := OK;
        Exit;
      end;
    DIALUP_LINE:
      begin
        Modem.Command('AT#VLS=0','OK|VCON');
        Result := OK;
        Exit;
      end;
    INTERNAL_MICROPHONE:
      begin
        Modem.Command('AT#VLS=3','OK|VCON');
        Result := OK;
        Exit;
      end;
    EXTERNAL_SPEAKER:
      begin
        Modem.Command('AT#VLS=2','OK|VCON');
        Result := OK;
        Exit;
      end;
    INTERNAL_SPEAKER:
      begin
        Modem.Command('AT#VLS=4','OK|VCON');
        Result := OK;
        Exit;
      end;
    LOCAL_HANDSET:
      begin
        Modem.Command('AT#VLS=1','OK|VCON');
        Result := OK;
        Exit;
      end;
    SPEAKER_PHONE:
      begin
        Modem.Command('AT#VLS=6','OK|VCON');
        Result := OK;
        Exit;
      end;
  end;
  { output device unknown }
  DebugMsgFmt(L_WARN, SUnknownDevice, [Device]);
  Modem.LastError := VME_UNKNOWN_DEVICE;
  Result := FAIL;
end;

function TUSRobotics.StopPlayStream: Integer;
begin
  { Stop playing }
  if inherited StopPlayStream = FAIL then
  begin
    Result := FAIL;
    Exit;
  end;
  { Restore local echo }
  if Profile.CommandEcho then
  begin
    if Modem.WriteStr('ATE1') = FAIL then
    begin
      Result := FAIL;
      Exit;
    end;
    if Modem.Command('', 'ATE1|OK|VCON') and VMA_USER <> VMA_USER then
    begin
      Result := FAIL;
      Exit;
    end;
  end;
  { Check modem }
  if Modem.Command('AT', 'OK') <> VMA_USER_1 then
  begin
    Result := FAIL;
    Exit;
  end;
  Result := OK;
end;

initialization
  RegisterModemDriver(TUSRobotics.ModemName, TUSRobotics);

end.
