{
  Voice Modem Library v1.0

  This file contain some utils procedures and funtions.

  Roman M. Mochalov
  E-mail: roman@sar.nnov.ru
}

unit VML_Utils;

interface

uses
  Windows, SysUtils, Classes;

type

{ RMD file information }

  TRMDInfo = record
    ModemType: string;        { modem type }
    ModemName: string;        { modem name }
    Compression: SmallInt;    { modem specific compression method }
    CompressionName: string;  { compression name }
    Speed: SmallInt;          { sample speed }
    Bits: Byte;               { bits per Sample }
    DataSize: Longint;        { voice data size in bytes }
    VoiceLen: Longint;        { voice message length in msecs }
  end;

{ Get RMD file information. Returned OK if success, otherwise returned
  error number (see VME_XXX constants for details). }

function GetRMDFileInfo(const FileName: string; var Info: TRMDInfo): Integer;

{ Modem compression and devices routines. }

function GetCompression(const ModemName: string; const Name: string): Integer; 
function GetCompressionName(const ModemName: string; Compression: Integer): string;
function GetCompressionsList(const ModemName: string; Proc: TGetStrProc): Integer;
function GetCompressionsStringList(const ModemName: string; Strings: TStrings): Integer;
function GetDevice(const ModemName: string; const Name: string): Integer;
function GetDeviceName(const ModemName: string; Device: Integer): string;
function GetDevicesList(const ModemName: string; Proc: TGetStrProc): Integer;
function GetDevicesStringList(const ModemName: string; Strings: TStrings): Integer;
function GetPlayDevicesList(const ModemName: string; Proc: TGetStrProc): Integer;
function GetPlayDevicesStringList(const ModemName: string; Strings: TStrings): Integer;
function GetRecDevicesList(const ModemName: string; Proc: TGetStrProc): Integer;
function GetRecDevicesStringList(const ModemName: string; Strings: TStrings): Integer;

implementation

uses
  VML_Modem, VML_Consts;

function GetRMDFileInfo(const FileName: string; var Info: TRMDInfo): Integer;
var
  Handle: Integer;
  Stream: THandleStream;
  Header: TRMDHeader;
  DriverClass: TModemDriverClass;
begin
  { Check file }
  if not FileExists(FileName) then
  begin
    Result := VME_FILE_NOT_FOUND;
    Exit;
  end;
  { Open file }
  Handle := FileOpen(FileName, fmOpenRead);
  { Check for file opening }
  if Handle < 0 then
  begin
    Result := VME_FILE_OPEN_ERROR;
    Exit;
  end;
  try
    Stream := THandleStream.Create(Handle);
    try
      { Read file header }
      if Stream.Read(Header, SizeOf(TRMDHeader)) <> SizeOf(TRMDHeader) then
      begin
        Result := VME_STREAM_READ_ERR;
        Exit;
      end;
      { Check for header magic }
      if StrLComp(Header.Magic, RMD_HEADER_MAGIC, 4) <> 0 then
      begin
        Result := VME_BAD_STREAM_FORMAT;
        Exit;
      end;
      { Clear result }
      FillChar(Info, SizeOf(Info), 0);
      { Get header information }
      with Info do
      begin
        { Common information }
        ModemType := StrPas(Header.ModemType);
        Compression := Header.Compression;
        Speed := Header.Speed;
        Bits := Header.Bits;
        DataSize := Stream.Size - SizeOf(TRMDHeader);
        { Driver specific information }
        if GetModemDriver(ModemType, DriverClass) then
        begin
          ModemName := DriverClass.ModemName;
          CompressionName := DriverClass.GetCompressionName(Compression);
          VoiceLen := DriverClass.GetVoiceLen(Compression, Speed, Bits, DataSize);
        end
        else
        begin
          ModemName := '?';
          CompressionName := '?';
          VoiceLen := 0;
        end;
      end;
    finally
      Stream.Free;
    end;
  finally
    FileClose(Handle);
  end;
  { Success }
  Result := OK;
end;

function GetCompression(const ModemName: string; const Name: string): Integer;
var
  DriverClass: TModemDriverClass;
begin
  { Get modem driver class }
  if GetModemDriver(ModemName, DriverClass) then
  begin
    Result := DriverClass.GetCompression(Name);
    Exit;
  end;
  { Try to return number from name }
  Result := StrToIntDef(Name, -1);
end;

function GetCompressionName(const ModemName: string; Compression: Integer): string;
var
  DriverClass: TModemDriverClass;
begin
  { Get modem driver class }
  if GetModemDriver(ModemName, DriverClass) then
  begin
    Result := DriverClass.GetCompressionName(Compression);
    Exit;
  end;
  { Return given number string }
  Result := IntToStr(Compression);
end;

function GetCompressionsList(const ModemName: string; Proc: TGetStrProc): Integer;
var
  DriverClass: TModemDriverClass;
begin
  { Get modem driver class }
  if GetModemDriver(ModemName, DriverClass) then
  begin
    Result := DriverClass.GetCompressionsList(Proc);
    Exit;
  end;
  { Compression list is empty }
  Result := 0;
end;

function GetCompressionsStringList(const ModemName: string; Strings: TStrings): Integer;
var
  DriverClass: TModemDriverClass;
begin
  { Get modem driver class }
  if GetModemDriver(ModemName, DriverClass) then
  begin
    Result := DriverClass.GetCompressionsStringList(Strings);
    Exit;
  end;
  { Compression list is empty }
  Result := 0;
end;

function GetDevice(const ModemName: string; const Name: string): Integer;
var
  DriverClass: TModemDriverClass;
begin
  { Get modem driver class }
  if GetModemDriver(ModemName, DriverClass) then
  begin
    Result := DriverClass.GetDevice(Name);
    Exit;
  end;
  { Try to return number from name }
  Result := StrToIntDef(Name, -1);
end;

function GetDeviceName(const ModemName: string; Device: Integer): string;
var
  DriverClass: TModemDriverClass;
begin
  { Get modem driver class }
  if GetModemDriver(ModemName, DriverClass) then
  begin
    Result := DriverClass.GetDeviceName(Device);
    Exit;
  end;
  { Return given number string }
  Result := IntToStr(Device);
end;

function GetDevicesList(const ModemName: string; Proc: TGetStrProc): Integer;
var
  DriverClass: TModemDriverClass;
begin
  { Get modem driver class }
  if GetModemDriver(ModemName, DriverClass) then
  begin
    Result := DriverClass.GetDevicesList(Proc);
    Exit;
  end;
  { Device list is empty }
  Result := 0;
end;

function GetDevicesStringList(const ModemName: string; Strings: TStrings): Integer;
var
  DriverClass: TModemDriverClass;
begin
  { Get modem driver class }
  if GetModemDriver(ModemName, DriverClass) then
  begin
    Result := DriverClass.GetDevicesStringList(Strings);
    Exit;
  end;
  { Device list is empty }
  Result := 0;
end;

function GetPlayDevicesList(const ModemName: string; Proc: TGetStrProc): Integer;
var
  DriverClass: TModemDriverClass;
begin
  { Get modem driver class }
  if GetModemDriver(ModemName, DriverClass) then
  begin
    Result := DriverClass.GetPlayDevicesList(Proc);
    Exit;
  end;
  { Device list is empty }
  Result := 0;
end;

function GetPlayDevicesStringList(const ModemName: string; Strings: TStrings): Integer;
var
  DriverClass: TModemDriverClass;
begin
  { Get modem driver class }
  if GetModemDriver(ModemName, DriverClass) then
  begin
    Result := DriverClass.GetPlayDevicesStringList(Strings);
    Exit;
  end;
  { Device list is empty }
  Result := 0;
end;

function GetRecDevicesList(const ModemName: string; Proc: TGetStrProc): Integer;
var
  DriverClass: TModemDriverClass;
begin
  { Get modem driver class }
  if GetModemDriver(ModemName, DriverClass) then
  begin
    Result := DriverClass.GetRecDevicesList(Proc);
    Exit;
  end;
  { Device list is empty }
  Result := 0;
end;

function GetRecDevicesStringList(const ModemName: string; Strings: TStrings): Integer;
var
  DriverClass: TModemDriverClass;
begin
  { Get modem driver class }
  if GetModemDriver(ModemName, DriverClass) then
  begin
    Result := DriverClass.GetRecDevicesStringList(Strings);
    Exit;
  end;
  { Device list is empty }
  Result := 0;
end;

end.
