{
  Voice Modem Library v1.0

  This file contains the ZyXEL 1496 specific hardware stuff.

  Roman M. Mochalov
  E-mail: roman@sar.nnov.ru
}

unit VML_ZyXEL1496;

interface

uses
  Windows, SysUtils, Classes, VML_Modem, VML_Consts, VML_IS101, VML_Idents;

const

{ ZyXEL 1496 available compression modes }

  COMPRESSION_ZYXEL1496_CELP     = 1;
  COMPRESSION_ZYXEL1496_ADPCM_2  = 2;
  COMPRESSION_ZYXEL1496_ADPCM_3  = 3;
  COMPRESSION_ZYXEL1496_ADPCM_30 = 30;
  COMPRESSION_ZYXEL1496_ADPCM_4  = 4;

type

{ TZyXEL1496 }

  TZyXEL1496 = class(TIS101)
  private
    FROMRelease: Integer;
  public
    constructor Create(AOwner: TComponent); override;
    class function GetCompressionMap(var Map: PIdentMap): Integer; override;
    class function GetDeviceMap(var Map: PIdentMap): Integer; override;
    function Init: Integer; override;
    function InitVoice: Integer; override;
    class function ModemName: string; override;
    class function ModemRMDName: string; override;
    function SetCompression(var Compression, Speed, Bits: Integer): Integer; override;
    function SetDevice(Device: Integer): Integer; override;
    property ROMRelease: Integer read FROMRelease;
  end;

implementation

{$I *.INC}

{ Some internal constants }

const

  CompressionMap: array[0..3] of TIdentMapEntry = (
    (Value: COMPRESSION_ZYXEL1496_CELP; Name: SCELP),
    (Value: COMPRESSION_ZYXEL1496_ADPCM_2; Name: SADPCM2),
    (Value: COMPRESSION_ZYXEL1496_ADPCM_3; Name: SADPCM3),
    (Value: COMPRESSION_ZYXEL1496_ADPCM_4; Name: SADPCM4));

  DeviceMap: array[0..2] of TIdentMapEntry = (
    (Value: DIALUP_LINE; Name: SDialupLine),
    (Value: EXTERNAL_MICROPHONE; Name: SExternalMicrophone),
    (Value: INTERNAL_SPEAKER; Name: SInternalSpeaker));

{ TZyXEL1496 }

constructor TZyXEL1496.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  PickPhoneAnsw := 'VCON';
  BeepTimeUnit := 100;
  IntrPlayCmd := DLE + DC4;
  IntrPlayAnsw := 'VCON';
  StopPlayAnsw := 'VCON';
  StopRecAnsw := 'VCON';
end;

class function TZyXEL1496.GetCompressionMap(var Map: PIdentMap): Integer;
begin
  Map := @CompressionMap;
  Result := High(CompressionMap) - Low(CompressionMap) + 1;
end;

class function TZyXEL1496.GetDeviceMap(var Map: PIdentMap): Integer;
begin
  Map := @DeviceMap;
  Result := High(DeviceMap) - Low(DeviceMap) + 1;
end;

function TZyXEL1496.Init: Integer;
var
  Buffer: string;
begin
  DebugMsgFmt(L_MESG, SInitModem, [ModemName]);
  { Check modem specified }
  if not ModemAssigned('Init', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Update modem state }
  Modem.State := INITIALIZING;
  try
    { Detect ROM release }
    if Modem.Command('ATI1', '') <> OK then
    begin
      Result := FAIL;
      Exit;
    end;
    if Modem.ReadStr(Buffer) = FAIL then
    begin
      Result := FAIL;
      Exit;
    end;
    if Modem.ReadStr(Buffer) = FAIL then
    begin
      Result := FAIL;
      Exit;
    end;
    if Modem.Command('', 'OK') <> VMA_USER_1 then
    begin
      Result := FAIL;
      Exit;
    end;
    if Length(Buffer) >= 13 then
    begin
      FROMRelease := 0; 
      FROMRelease := FROMRelease + (Ord(Buffer[11]) - Ord('0')) * 100;
      FROMRelease := FROMRelease + (Ord(Buffer[13]) - Ord('0')) * 10;
      FROMRelease := FROMRelease + (Ord(Buffer[14]) - Ord('0'));
    end
    else
      FROMRelease := 505;
    DebugMsgFmt(L_MESG, SROMRelease, [FROMRelease / 100]);
    { AT&H3 - Hardware CTS/RTS flow control.
      AT&R1 - Modem assumes RTS always ON. }
    if Modem.Command('AT&H3&R1', 'OK') <> VMA_USER_1 then
    begin
      DebugMsg(L_WARN, SFlowControlError);
      Result := FAIL;
      Exit;
    end;
    { Set extended verbal result messages }
    if Modem.Command('ATX5', 'OK') <> VMA_USER_1 then
    begin
      DebugMsg(L_WARN, SExResultCodesError);
      Result := FAIL;
      Exit;
    end;
    { ATS40.3=1 - Enable distincitve ring type 1 (RING)
      ATS40.4=1 - Enable distincitve ring type 2 (RING 1)
      ATS40.5=1 - Enable distincitve ring type 3 (RING 2)
      ATS40.6=1 - Enable distincitve ring type 4 (RING 3) }
    if Modem.Command('ATS40.3=1S40.4=1S40.5=1S40.6=1', 'OK') <> VMA_USER_1 then
      DebugMsg(L_WARN, SDistinctiveRingError);
  finally
    Modem.State := IDLE;
  end;
  { Init complete }
  Result := OK;
end;

function TZyXEL1496.InitVoice: Integer;
var
  TransmitGain, ReceiveGain: Integer;
begin
  DebugMsgFmt(L_MESG, SInitVoice, [ModemName]);
  { Check modem specified }
  if not ModemAssigned('InitVoice', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Update modem state }
  Modem.State := INITIALIZING;
  try
    { ATS39.6=1 - Enable DTMF detection after AT+VLS=2
      ATS39.7=0 - Don't include resynchronization information
                  in recorded voice data
      AT+VIT=60 - Set inactivity timer to 60 seconds }
    if Modem.Command('ATS39.6=1S39.7=0+VIT=60', 'OK|NO CARRIER') <> VMA_USER_1 then
      DebugMsg(L_WARN, SVoiceInitFailed);
    { AT+VSD=x,y - Set silence threshold and duration }
    if Modem.CommandFmt('AT+VSD=%d,%d', 'OK', [Profile.RecSilenceSens * 31 div 100, Profile.RecSilenceLen]) <> VMA_USER_1 then
      DebugMsg(L_WARN, SRecordingPreferencesError);
    { AT+VNH=1 - Disable autohangup }
    if Modem.Command('AT+VNH=1', 'OK') <> VMA_USER_1 then
      DebugMsg(L_WARN, SAutoHangUpError);
    { AT+VDH=x - Set the threshold for DTMF detection (0-32)
      AT+VDD=x - Set DTMF tone duration detection }
    if Modem.CommandFmt('AT+VDH=%d +VDD=%d', 'OK|NO CARRIER', [Profile.DTMFSens * 31 div 100, Profile.DTMFLen div 5]) <> VMA_USER_1 then
      DebugMsg(L_WARN, SDTMFPreferencesError);
    { AT+VGT - Set the transmit gain for voice samples. }
    TransmitGain := Profile.TransmitGain;
    if TransmitGain = -1 then TransmitGain := 75;
    if Modem.CommandFmt('AT+VGT=%d', 'OK', [TransmitGain * 255 div 100]) <> VMA_USER_1 then
      DebugMsg(L_WARN, STransmitGainError);
    { AT+VGR - Set receive gain for voice samples. }
    ReceiveGain := Profile.ReceiveGain;
    if ReceiveGain = -1 then ReceiveGain := 75;
    if Modem.CommandFmt('AT+VGR=%d', 'OK', [ReceiveGain * 255 div 100]) <> VMA_USER_1 then
      DebugMsg(L_WARN, SReceiveGainError);
    { Delay after ringback or before any ringback before modem assumes
      phone has been answered }
    if Modem.CommandFmt('AT+VRA=%d+VRN=%d', 'OK', [Profile.RingbackGoesAway, Profile.RingbackNeverCame]) <> VMA_USER_1 then
      DebugMsg(L_ERROR, SRingbackTimersError);
  finally
    Modem.State := IDLE;
  end;
  { Success }
  Result := OK;
end;

class function TZyXEL1496.ModemName: string;
begin
  Result := 'ZyXEL 1496';
end;

class function TZyXEL1496.ModemRMDName: string;
begin
  Result := 'ZyXEL 1496';
end;

function TZyXEL1496.SetCompression(var Compression, Speed, Bits: Integer): Integer;
begin
  { Check modem specified }
  if not ModemAssigned('SetCompression', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { The default mode for ZyXEL modems is ADPCM }
  if Compression = 0 then Compression := COMPRESSION_ZYXEL1496_ADPCM_2;
  { 9600 Hz is currently the only recording freq supported by ZyXEL }
  if Speed = 0 then Speed := 9600;
  if Speed <> 9600 then
  begin
    DebugMsgFmt(L_WARN, SIllegalSampleRate, [Speed]);
    Result := FAIL;
    Exit;
  end;
  { Setup compression }
  case Compression of
    COMPRESSION_ZYXEL1496_CELP:
      { This is the CELP 1 bits per sample mode }
      begin
        Bits := 1;
        { Set the mode to CELP }
        if Modem.Command('AT+VSM=1', 'OK') = VMA_FAIL then
        begin
          Result := VMA_FAIL;
          Exit;
        end;
      end;
    COMPRESSION_ZYXEL1496_ADPCM_2:
      { This is the ADPCM 2 bits per sample mode }
      begin
        Bits := 2;
        { Set the mode to ADPCM }
        if Modem.Command('AT+VSM=2', 'OK') = VMA_FAIL then
        begin
          Result := VMA_FAIL;
          Exit;
        end;
      end;
    COMPRESSION_ZYXEL1496_ADPCM_3:
      { This is the ADPCM 3 bits per sample mode }
      begin
        Bits := 3;
        { Set the mode to ADPCM }
        if Modem.Command('AT+VSM=3', 'OK') = VMA_FAIL then
        begin
          Result := VMA_FAIL;
          Exit;
        end;
      end;
    COMPRESSION_ZYXEL1496_ADPCM_30:
      { This is the ? mode }
      begin
        Bits := 3;
        { Set the mode to ADPCM }
        if Modem.Command('AT+VSM=30', 'OK') = VMA_FAIL then
        begin
          Result := VMA_FAIL;
          Exit;
        end;
      end;
    COMPRESSION_ZYXEL1496_ADPCM_4:
      { This is the ADPCM 4 bits per sample mode }
      begin
        Bits := 4;
        { Set the mode to ADPCM }
        if Modem.Command('AT+VSM=4', 'OK') = VMA_FAIL then
        begin
          Result := VMA_FAIL;
          Exit;
        end;
      end;
    else
      begin
        DebugMsgFmt(L_WARN, SIllegalCompressionMethod, [Compression]);
        Modem.LastError := VME_ILLEGAL_COMPRESSION;
        Result := FAIL;
        Exit;
      end;
  end;
  Result := OK;
  Exit;
end;

function TZyXEL1496.SetDevice(Device: Integer): Integer;
begin
  { Check modem specified }
  if not ModemAssigned('SetDevice', True) then
  begin
    Result := FAIL;
    Exit;
  end;
  { Set device }
  case Device of
    NO_DEVICE:
      begin
        if Modem.WriteStr('AT+VLS=0') <> OK then
        begin
          Result := FAIL;
          Exit;
        end;
        if Modem.Command('', 'AT+VLS=0|OK') and VMA_USER <> VMA_USER then
        begin
          Result := FAIL;
          Exit;
        end;
        Modem.Flush;
        Result := OK;
        Exit;
      end;
    DIALUP_LINE:
      begin
        Modem.Command('AT+VLS=2','VCON');
        Result := OK;
        Exit;
      end;
    EXTERNAL_MICROPHONE:
      begin
        Modem.Command('AT+VLS=8','OK|VCON');
        Result := OK;
        Exit;
      end;
    INTERNAL_SPEAKER:
      begin
        Modem.Command('AT+VLS=16','OK');
        Result := OK;
        Exit;
      end;
  end;
  { Given device unknown }
  DebugMsgFmt(L_WARN, SUnknownDevice, [Device]);
  Modem.LastError := VME_UNKNOWN_DEVICE;
  Result := FAIL;
end;

initialization
  RegisterModemDriver(TZyXEL1496.ModemName, TZyXEL1496);

end.
